﻿//******************************************************************************************************
//  CsvInputAdapter.cs - Gbtc
//
//  Copyright © 2010, Grid Protection Alliance.  All Rights Reserved.
//
//  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
//  the NOTICE file distributed with this work for additional information regarding copyright ownership.
//  The GPA licenses this file to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  04/06/2010 - Stephen C. Wills
//       Generated original version of source code.
//  07/03/2012 - J. Ritchie Carroll
//       Added high-resolution input timer, auto-repeat and transverse operational mode.
//  12/20/2012 - B.PrasannaKumari
//       Modified for Migrating adapters from the Time Series Framework to SIEGate Core.
//
//******************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;
using GSF;
using GSF.TimeSeries;
using GSF.TimeSeries.Adapters;
using SIEGate.Core;
using SIEGate.Core.Configurator;
using SIEGate.Core.Legacy;
using SIEGate.Core.Measurements.Values;

namespace CsvAdapters
{
    /// <summary>
    /// Represents an input adapter that reads measurements from a CSV file.
    /// </summary>
    [Description("CSV: reads measurements from a CSV file.")]
    public class CsvInputAdapter : InputAdapterBase
    {
        #region [ Members ]

        // Nested Types

        public class AdapterInfo : AdapterInfoBase
        {
            // Fields
            private string m_fileName;
            private double m_inputInterval;
            private int m_measurementsPerInterval;
            private int m_skipRows;
            private bool m_simulateTimestamp;
            private bool m_transverse;
            private bool m_autoRepeat;
            private bool m_useHighResolutionInputTimer;

            /// <summary>
            /// Gets or sets the name of the CSV file from which measurements will be read.
            /// </summary>
            [ConnectionStringParameter,
             Description("Define the name of the CSV file from which measurements will be read."),
             DefaultValue("measurements.csv")]
            public string FileName
            {
                get
                {
                    return m_fileName;
                }
                set
                {
                    m_fileName = value;
                }
            }

            /// <summary>
            /// Gets or sets the interval of time between sending frames into the concentrator.
            /// </summary>
            [ConnectionStringParameter,
             Description("Define the interval of time, in milliseconds, between sending frames into the concentrator."),
             DefaultValue(33.333333)]
            public double InputInterval
            {
                get
                {
                    return m_inputInterval;
                }
                set
                {
                    m_inputInterval = value;
                }
            }

            /// <summary>
            /// Gets or sets value that determines if the CSV input file data should be replayed repeatedly.
            /// </summary>
            [ConnectionStringParameter,
             Description("Define if the CSV input file data should be replayed repeatedly."),
             DefaultValue(false)]
            public bool AutoRepeat
            {
                get
                {
                    return m_autoRepeat;
                }
                set
                {
                    m_autoRepeat = value;
                }
            }

            /// <summary>
            /// Gets or sets number of lines to skip in the source file before the header line is encountered.
            /// </summary>
            [ConnectionStringParameter,
             Description("Define the number of lines to skip in the source file before the header line is encountered."),
             DefaultValue(0)]
            public int SkipRows
            {
                get
                {
                    return m_skipRows;
                }
                set
                {
                    m_skipRows = value;
                }
            }

            /// <summary>
            /// Gets or sets the number of measurements that are read from the CSV file in each frame.
            /// </summary>
            [ConnectionStringParameter,
             Description("Define the number of measurements measurements that are read from the CSV file in each frame."),
             DefaultValue(5)]
            public int MeasurementsPerInterval
            {
                get
                {
                    return m_measurementsPerInterval;
                }
                set
                {
                    m_measurementsPerInterval = value;
                }
            }

            /// <summary>
            /// Gets or sets a value that determines whether CSV file is in transverse mode for real-time concentration.
            /// </summary>
            [ConnectionStringParameter,
             Description("Indicate whether CSV file is in transverse mode for real-time concentration."),
             DefaultValue(false)]
            public bool TransverseMode
            {
                get
                {
                    return m_transverse;
                }
                set
                {
                    m_transverse = value;
                }
            }

            /// <summary>
            /// Gets or sets a value that determines whether timestamps are
            /// simulated for the purposes of real-time concentration.
            /// </summary>
            [ConnectionStringParameter,
             Description("Indicate whether timestamps are simulated for real-time concentration."),
             DefaultValue(false)]
            public bool SimulateTimestamp
            {
                get
                {
                    return m_simulateTimestamp;
                }
                set
                {
                    m_simulateTimestamp = value;
                }
            }

            /// <summary>
            /// Defines the column mappings must defined: e.g., 0=Timestamp; 1=PPA:12; 2=PPA13.
            /// </summary>
            [ConnectionStringParameter,
             Description("Defines the column mappings must defined: e.g., \"0=Timestamp; 1=PPA:12; 2=PPA13\"."),
             DefaultValue("")]
            public string ColumnMappings
            {
                get;
                set;
            }

            /// <summary>
            /// Gets or sets flag that determines if a high-resolution precision timer should be used for CSV file based input.
            /// </summary>
            /// <remarks>
            /// Useful when input frames need be accurately time-aligned to the local clock to better simulate
            /// an input device and calculate downstream latencies.<br/>
            /// This is only applicable when connection is made to a file for replay purposes.
            /// </remarks>
            [ConnectionStringParameter,
             Description("Determines if a high-resolution precision timer should be used for CSV file based input."),
             DefaultValue(false)]
            public bool UseHighResolutionInputTimer
            {
                get
                {
                    return m_useHighResolutionInputTimer;
                }
                set
                {
                    m_useHighResolutionInputTimer = value;
                }
            }

            private readonly Dictionary<string, string> m_defaultSettings = new Dictionary<string, string>()
            {
                { "fileName", "measurements.csv" },
                { "inputInterval", "33.333333" },
                { "autoRepeat", "false" },
                { "skipRows", "0" },
                { "measurementsPerInterval", "5" },
                { "transverse", "false" },
                { "simulateTimestamp", "false" },
                { "columnMappings", "" },
            };

            /// <summary>
            /// Set the member variables and properties for this AdapterInfo from a parsed configuration string
            /// </summary>
            /// <param name="configInfo">A Dictionary (string, string) that results from parsing a configuration string.</param>
            public override void Update(Dictionary<string, string> configInfo)
            {
                base.Update(configInfo);

                // Load optional parameters
                FileName = DefaultGet("fileName");
                InputInterval = double.Parse(DefaultGet("inputInterval"));
                AutoRepeat = DefaultGet("autoRepeat").ParseBoolean();
                SkipRows = int.Parse(DefaultGet("skipRows"));
                MeasurementsPerInterval = int.Parse(DefaultGet("measurementsPerInterval"));
                TransverseMode = DefaultGet("transverse").ParseBoolean();
                SimulateTimestamp = DefaultGet("simulateTimestamp").ParseBoolean();
                ColumnMappings = DefaultGet("columnMappings");
            }

            /// <summary>
            /// Set the DataSource for this AdapterInfo class and derive member variables and properties relevant to this instance from that source.
            /// </summary>
            /// <param name="dataSource"></param>
            public override void ExtractConfiguration(DataSet dataSource)
            {
                base.ExtractConfiguration(dataSource);
                ILegacyMeasurement measurement = new LegacyMeasurement();
                // Auto-assign output measurements based on column mappings
                if (this.DataSource != null && (object)DataSource.Tables != null && DataSource.Tables.Contains("ActiveMeasurements"))
                {
                    try
                    {
                        DataRow[] filteredRows = DataSource.Tables["ActiveMeasurements"].Select(string.Format("SignalID = '{0}'", measurement.ID));

                        if (filteredRows.Length > 0)
                        {
                            DataRow row = filteredRows[0];

                            // Assign other attributes
                            measurement.TagName = row["PointTag"].ToNonNullString();
                            measurement.Multiplier = double.Parse(row["Multiplier"].ToString());
                            measurement.Adder = double.Parse(row["Adder"].ToString());
                        }
                    }
                    catch
                    {
                        // Failure to lookup extra metadata is not catastrophic
                    }
                }
            }

            public AdapterInfo()
            {
                SetDefaults(m_defaultSettings);
            }
        }

        // Fields
        private StreamReader m_inStream;
        private string m_header;
        private readonly Dictionary<string, int> m_columns;
        private readonly Dictionary<int, ILegacyMeasurement> m_columnMappings;
        private System.Timers.Timer m_looseTimer;
        private PrecisionInputTimer m_precisionTimer;
        private bool m_disposed;

        private const string StrLogSource = "CsvInputAdapter";
        #endregion

        #region [ Constructors ]

        /// <summary>
        /// Initializes a new instance of the <see cref="CsvInputAdapter"/> class.
        /// </summary>
        public CsvInputAdapter(IAdapterInfo info)
            : base(info)
        {
            m_columns = new Dictionary<string, int>(StringComparer.CurrentCultureIgnoreCase);
            m_columnMappings = new Dictionary<int, ILegacyMeasurement>();

            m_statsCluster.RegisterCounter(info.Name, StrStatReceivedMeasurements);

            // Set minimum timer resolution to one millisecond to improve timer accuracy
            PrecisionTimer.SetMinimumTimerResolution(1);
        }

        #endregion

        #region [ Properties ]

        public new AdapterInfo Info
        {
            get
            {
                return ((AdapterInfo)base.Info);
            }
        }

        /// <summary>
        /// Gets or sets the name of the CSV file from which measurements will be read.
        /// </summary>
        [ConnectionStringParameter,
         Description("Define the name of the CSV file from which measurements will be read."),
         DefaultValue("measurements.csv")]
        public string FileName
        {
            get
            {
                return Info.FileName;
            }
            set
            {
                Info.FileName = value;
            }
        }

        /// <summary>
        /// Gets or sets the interval of time between sending frames into the concentrator.
        /// </summary>
        [ConnectionStringParameter,
         Description("Define the interval of time, in milliseconds, between sending frames into the concentrator."),
         DefaultValue(33.333333)]
        public double InputInterval
        {
            get
            {
                return Info.InputInterval;
            }
            set
            {
                Info.InputInterval = value;
            }
        }

        /// <summary>
        /// Gets or sets value that determines if the CSV input file data should be replayed repeatedly.
        /// </summary>
        [ConnectionStringParameter,
         Description("Define if the CSV input file data should be replayed repeatedly."),
         DefaultValue(false)]
        public bool AutoRepeat
        {
            get
            {
                return Info.AutoRepeat;
            }
            set
            {
                Info.AutoRepeat = value;
            }
        }

        /// <summary>
        /// Gets or sets number of lines to skip in the source file before the header line is encountered.
        /// </summary>
        [ConnectionStringParameter,
         Description("Define the number of lines to skip in the source file before the header line is encountered."),
         DefaultValue(0)]
        public int SkipRows
        {
            get
            {
                return Info.SkipRows;
            }
            set
            {
                Info.SkipRows = value;
            }
        }

        /// <summary>
        /// Gets or sets flag that determines if a high-resolution precision timer should be used for CSV file based input.
        /// </summary>
        /// <remarks>
        /// Useful when input frames need be accurately time-aligned to the local clock to better simulate
        /// an input device and calculate downstream latencies.<br/>
        /// This is only applicable when connection is made to a file for replay purposes.
        /// </remarks>
        [ConnectionStringParameter,
         Description("Determines if a high-resolution precision timer should be used for CSV file based input."),
         DefaultValue(false)]
        public bool UseHighResolutionInputTimer
        {
            get
            {
                return ((object)m_precisionTimer != null);
            }
            set
            {
                // Note that a 1-ms timer and debug mode don't mix, so the high-resolution timer is disabled while debugging
                if (value && (object)m_precisionTimer == null && !System.Diagnostics.Debugger.IsAttached)
                    m_precisionTimer = PrecisionInputTimer.Attach((int)(1000.0D / InputInterval), (Exception ex) => LogDispatcher.E(StrLogSource, "ProcessException", "{0}{1}{2}",
                                                                                                                                    ex.Message,
                                                                                                                                    Environment.NewLine,
                                                                                                                                    ex.StackTrace));
                else if (!value && m_precisionTimer != null)
                    PrecisionInputTimer.Detach(ref m_precisionTimer);
            }
        }

        /// <summary>
        /// Gets or sets the number of measurements that are read from the CSV file in each frame.
        /// </summary>
        [ConnectionStringParameter,
         Description("Define the number of measurements measurements that are read from the CSV file in each frame."),
         DefaultValue(5)]
        public int MeasurementsPerInterval
        {
            get
            {
                return Info.MeasurementsPerInterval;
            }
            set
            {
                Info.MeasurementsPerInterval = value;
            }
        }

        /// <summary>
        /// Gets or sets a value that determines whether CSV file is in transverse mode for real-time concentration.
        /// </summary>
        [ConnectionStringParameter,
         Description("Indicate whether CSV file is in transverse mode for real-time concentration."),
         DefaultValue(false)]
        public bool TransverseMode
        {
            get
            {
                return Info.TransverseMode;
            }
            set
            {
                Info.TransverseMode = value;
            }
        }

        /// <summary>
        /// Gets or sets a value that determines whether timestamps are
        /// simulated for the purposes of real-time concentration.
        /// </summary>
        [ConnectionStringParameter,
         Description("Indicate whether timestamps are simulated for real-time concentration."),
         DefaultValue(false)]
        public bool SimulateTimestamp
        {
            get
            {
                return Info.SimulateTimestamp;
            }
            set
            {
                Info.SimulateTimestamp = value;
            }
        }

        /// <summary>
        /// Defines the column mappings must defined: e.g., 0=Timestamp; 1=PPA:12; 2=PPA13.
        /// </summary>
        [ConnectionStringParameter,
         Description("Defines the column mappings must defined: e.g., \"0=Timestamp; 1=PPA:12; 2=PPA13\"."),
         DefaultValue("")]
        public string ColumnMappings
        {
            get
            {
                return Info.ColumnMappings;
            }
            set
            {
                Info.ColumnMappings = value;
            }
        }

        /// <summary>
        /// Gets a flag that determines if this <see cref="CsvInputAdapter"/>
        /// uses an asynchronous connection.
        /// </summary>
        protected override bool UseAsyncConnect
        {
            get
            {
                return false;
            }
        }

        /// <summary>
        /// Returns the detailed status of this <see cref="CsvInputAdapter"/>.
        /// </summary>
        public override string Status
        {
            get
            {
                StringBuilder status = new StringBuilder();

                status.Append(base.Status);
                status.AppendLine();
                status.AppendFormat("                 File name: {0}", FileName);
                status.AppendLine();
                status.AppendFormat("               File header: {0}", m_header);
                status.AppendLine();
                status.AppendFormat("            Input interval: {0}", InputInterval);
                status.AppendLine();
                status.AppendFormat(" Measurements per interval: {0}", MeasurementsPerInterval);
                status.AppendLine();
                status.AppendFormat("     Using traverse format: {0}", TransverseMode);
                status.AppendLine();
                status.AppendFormat("               Auto-repeat: {0}", AutoRepeat);
                status.AppendLine();
                status.AppendFormat("     Precision input timer: {0}", UseHighResolutionInputTimer ? "Enabled" : "Offline");
                status.AppendLine();
                status.AppendFormat("             Lines to skip: {0}", SkipRows);
                status.AppendLine();

                if (m_precisionTimer != null)
                {
                    status.AppendFormat("  Timer resynchronizations: {0}", m_precisionTimer.Resynchronizations);
                    status.AppendLine();
                }

                return status.ToString();
            }
        }

        /// <summary>
        /// Gets the flag indicating if this adapter supports temporal processing.
        /// </summary>
        public override bool SupportsTemporalProcessing
        {
            get
            {
                return true;
            }
        }

        #endregion

        #region [ Methods ]

        /// <summary>
        /// Releases the unmanaged resources used by the <see cref="CsvInputAdapter"/> object and optionally releases the managed resources.
        /// </summary>
        /// <param name="disposing">true to release both managed and unmanaged resources; false to release only unmanaged resources.</param>
        protected override void Dispose(bool disposing)
        {
            if (!m_disposed)
            {
                try
                {
                    if (disposing)
                    {
                        if (UseHighResolutionInputTimer)
                        {
                            PrecisionInputTimer.Detach(ref m_precisionTimer);
                        }
                        else if ((object)m_looseTimer != null)
                        {
                            m_looseTimer.Stop();
                            m_looseTimer.Dispose();
                        }
                        m_looseTimer = null;

                        if ((object)m_inStream != null)
                        {
                            m_inStream.Close();
                            m_inStream.Dispose();
                        }

                        m_inStream = null;

                        // Clear minimum timer resolution.
                        PrecisionTimer.ClearMinimumTimerResolution(1);
                    }
                }
                finally
                {
                    m_disposed = true; // Prevent duplicate dispose.
                    base.Dispose(disposing); // Call base class Dispose().
                }
            }
        }

        /// <summary>
        /// Initializes this <see cref="CsvInputAdapter"/>.
        /// </summary>
        public override void Initialize()
        {
            base.Initialize();

            UseHighResolutionInputTimer = UseHighResolutionInputTimer;

            if (!UseHighResolutionInputTimer)
                m_looseTimer = new System.Timers.Timer();

            if (TransverseMode)
            {
                Dictionary<int, string> columnMappings = new Dictionary<int, string>();
                int index;

                // Load column mappings:
                foreach (KeyValuePair<string, string> mapping in ColumnMappings.ParseKeyValuePairs())
                {
                    if (int.TryParse(mapping.Key, out index))
                        columnMappings[index] = mapping.Value;
                }

                if (!SimulateTimestamp && !columnMappings.Values.Contains("Timestamp", StringComparer.InvariantCultureIgnoreCase))
                    throw new InvalidOperationException("One of the column mappings must be defined as a \"Timestamp\": e.g., columnMappings={0=Timestamp; 1=PPA:12; 2=PPA13}.");

                // In transverse mode, maximum measurements per interval is set to maximum columns in input file
                MeasurementsPerInterval = columnMappings.Keys.Max() + 1;

                // Auto-assign output measurements based on column mappings
                OutputMeasurements = columnMappings.Where(kvp => string.Compare(kvp.Value, "Timestamp", true) != 0).Select(kvp =>
                {
                    string measurementID = kvp.Value;
                    ILegacyMeasurement measurement = new LegacyMeasurement(); //modified 12/20/2012
                    MeasurementKey key; //modified 12/20/2012
                    Guid id;

                    if (Guid.TryParse(measurementID, out id))
                    {
                        measurement.ID = id;
                        measurement.Key = MeasurementKey.LookupBySignalID(id); //modified 12/20/2012
                    }
                    else if (MeasurementKey.TryParse(measurementID, Guid.Empty, out key)) //modified 12/20/2012
                    {
                        measurement.Key = key;
                        measurement.ID = key.SignalID;
                    }

                    try
                    {
                        DataRow[] filteredRows = DataSource.Tables["ActiveMeasurements"].Select(string.Format("SignalID = '{0}'", measurement.ID));

                        if (filteredRows.Length > 0)
                        {
                            DataRow row = filteredRows[0];

                            // Assign other attributes
                            measurement.TagName = row["PointTag"].ToNonNullString();
                            measurement.Multiplier = double.Parse(row["Multiplier"].ToString());
                            measurement.Adder = double.Parse(row["Adder"].ToString());
                        }
                    }
                    catch
                    {
                        // Failure to lookup extra metadata is not catastrophic
                    }

                    // Associate measurement with column index
                    m_columnMappings[kvp.Key] = measurement;

                    return measurement;
                }).ToArray();

                int timestampColumn = columnMappings.First(kvp => string.Compare(kvp.Value, "Timestamp", true) == 0).Key;

                // Reserve a column mapping for timestamp value
                ILegacyMeasurement timestampMeasurement = new LegacyMeasurement()
                {
                    TagName = "Timestamp"
                };

                m_columnMappings[timestampColumn] = timestampMeasurement;
            }

            // Override input interval based on temporal processing interval if it's not set to default
            if (ProcessingInterval > -1)
                InputInterval = ProcessingInterval == 0 ? 1 : ProcessingInterval;

            if ((object)m_looseTimer != null)
            {
                m_looseTimer.Interval = InputInterval;
                m_looseTimer.AutoReset = true;
                m_looseTimer.Elapsed += m_looseTimer_Elapsed;
            }
        }

        /// <summary>
        /// Attempts to connect to this <see cref="CsvInputAdapter"/>.
        /// </summary>
        protected override void AttemptConnection()
        {
            string[] headings;

            m_inStream = new StreamReader(File.Open(FileName, FileMode.Open, FileAccess.Read, FileShare.ReadWrite));

            // Skip specified number of header lines that exist before column heading definitions
            for (int i = 0; i < SkipRows; i++)
            {
                m_inStream.ReadLine();
            }

            m_columns.Clear();
            m_header = m_inStream.ReadLine();
            headings = m_header.ToNonNullString().Split(',');

            for (int i = 0; i < headings.Length; i++)
            {
                m_columns.Add(headings[i], i);
            }

            if (UseHighResolutionInputTimer)
            {
                // Start a new thread to process measurements using precision timer
                (new Thread(ProcessMeasurements)).Start();
            }
            else
            {
                // Start common timer
                m_looseTimer.Start();
            }
        }

        /// <summary>
        /// Attempts to disconnect from this <see cref="CsvInputAdapter"/>.
        /// </summary>
        protected override void AttemptDisconnection()
        {
            if ((object)m_inStream != null)
            {
                m_inStream.Close();
                m_inStream.Dispose();
            }

            m_inStream = null;

            if (!UseHighResolutionInputTimer)
                m_looseTimer.Stop();
        }

        /// <summary>
        /// Gets a short one-line status of this <see cref="CsvInputAdapter"/>.
        /// </summary>
        /// <param name="maxLength">Maximum length of the status message.</param>
        /// <returns>Text of the status message.</returns>
        public override string GetShortStatus(int maxLength)
        {
            return string.Format("{0} measurements read from CSV file.", ProcessedMeasurements).CenterText(maxLength);
        }

        private void ProcessMeasurements()
        {
            while (Enabled && ReadNextRecord(m_precisionTimer.LastFrameTime))
            {
                // When high resolution input timing is requested, we only need to wait for the next signal...
                m_precisionTimer.FrameWaitHandle.Wait();
            }

            if (Enabled)
            {
                Stop();

                if (AutoRepeat)
                    Start();
            }
        }

        private void m_looseTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            if (!ReadNextRecord(DateTime.UtcNow.Ticks) && Enabled)
            {
                Stop();

                if (AutoRepeat)
                    Start();
            }
        }

        // Attempt to read the next record
        private bool ReadNextRecord(long currentTime)
        {
            try
            {
                List<ILegacyMeasurement> newMeasurements = new List<ILegacyMeasurement>();
                long fileTime = 0;
                int timestampColumn = 0;
                string[] fields = m_inStream.ReadLine().ToNonNullString().Split(',');

                if (m_inStream.EndOfStream || fields.Length < m_columns.Count)
                    return false;

                // Read time from Timestamp column in transverse mode
                if (TransverseMode)
                {
                    if (SimulateTimestamp)
                    {
                        fileTime = currentTime;
                    }
                    else
                    {
                        timestampColumn = m_columnMappings.First(kvp => string.Compare(kvp.Value.TagName, "Timestamp", true) == 0).Key;
                        fileTime = long.Parse(fields[timestampColumn]);
                    }
                }

                for (int i = 0; i < MeasurementsPerInterval; i++)
                {
                    ILegacyMeasurement measurement;

                    if (TransverseMode)
                    {
                        // No measurement will be defined for timestamp column
                        if (i == timestampColumn)
                            continue;

                        if (m_columnMappings.TryGetValue(i, out measurement))
                        {
                            measurement = LegacyMeasurement.Clone(measurement);
                            //measurement.Value = double.Parse(fields[i]);
                            ((LegacyValue)measurement.Value).SetRaw(double.Parse(fields[i])); //modified 12/20/2012

                        }
                        else
                        {
                            measurement = new LegacyMeasurement();
                            measurement.ID = Guid.Empty;
                            measurement.Key = MeasurementKey.Undefined;
                            //measurement.Value = double.NaN;
                            ((LegacyValue)measurement.Value).SetRaw(double.NaN); //modified 12/20/2012

                        }

                        if (SimulateTimestamp)
                            ((LegacyMeasurement)measurement).Timestamp = currentTime;
                        else if (m_columns.ContainsKey("Timestamp"))
                            ((LegacyMeasurement)measurement).Timestamp = fileTime;
                    }
                    else
                    {
                        measurement = new LegacyMeasurement();

                        if (m_columns.ContainsKey("Signal ID"))
                            measurement.ID = new Guid(fields[m_columns["Signal ID"]]);

                        if (m_columns.ContainsKey("Measurement Key"))
                            measurement.Key = MeasurementKey.Parse(fields[m_columns["Measurement Key"]], measurement.ID);

                        if (SimulateTimestamp)
                            ((LegacyMeasurement)measurement).Timestamp = currentTime;
                        else if (m_columns.ContainsKey("Timestamp"))
                            ((LegacyMeasurement)measurement).Timestamp = long.Parse(fields[m_columns["Timestamp"]]);

                        if (m_columns.ContainsKey("Value"))
                            //measurement.Value = double.Parse(fields[m_columns["Value"]]);
                            ((LegacyValue)measurement.Value).SetRaw(double.Parse(fields[m_columns["Value"]]));
                    }

                    newMeasurements.Add(measurement);
                }

                m_statsCluster.AddToCounter(StrStatReceivedMeasurements, newMeasurements.Count);
                OnNewMeasurements(newMeasurements);
            }
            catch (Exception ex)
            {
                LogDispatcher.E(StrLogSource, "ProcessException", "{0}{1}{2}",
                                ex.Message,
                                Environment.NewLine,
                                ex.StackTrace);
            }

            return true;
        }

        #endregion
    }
}
