﻿//******************************************************************************************************
//  Dnp3InputAdapter.cs - Gbtc
//
//  Copyright © 2010, Grid Protection Alliance.  All Rights Reserved.
//
//  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
//  the NOTICE file distributed with this work for additional information regarding copyright ownership.
//  The GPA licenses this file to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  10/05/2012 - Adam Crain
//       Generated original version of source code.
//  10/8/2012 - Danyelle Gilliam
//        Modified Header
//  12/26/2012 - Prasanna Kumari
//       Applied changes for Migrating adapters from the Time Series Framework to SIEGate Core
//
//******************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.IO;
using DNP3.Adapter;
using DNP3.Interface;
using GSF.TimeSeries.Adapters;
using SIEGate.Core;
using SIEGate.Core.Configurator;
using SIEGate.Core.Legacy;

namespace Dnp3Adapters
{
    [Description("DNP3: Reads measurements from a remote dnp3 endpoint")]
    public class Dnp3InputAdapter : InputAdapterBase
    {
        #region [ Members ]

        // Nested Types

        public class AdapterInfo : AdapterInfoBase
        {
            // Fields
            /// <summary>
            /// The filename for the communication configuration file
            /// </summary>
            private String m_commsFileName;

            /// <summary>
            /// The filename for the measurement mapping configuration file
            /// </summary>
            private String m_mappingFileName;
            
            /// <summary>
            /// Gets or sets the name of the xml file from which the communication parameters will be read
            /// </summary>
            [ConnectionStringParameter,
            Description("Define the name of the XML file from which the communication configuration will be read"),
            DefaultValue("comms1.xml")]
            public string CommsFileName
            {
                get
                {
                    return m_commsFileName;
                }
                set
                {
                    m_commsFileName = value;
                }
            }

            /// <summary>
            /// Gets or sets the name of the xml file from which the measurement mapping is read
            /// </summary>
            [ConnectionStringParameter,
            Description("Define the name of the XML file from which the communication configuration will be read"),
            DefaultValue("device1.xml")]
            public string MappingFileName
            {
                get
                {
                    return m_mappingFileName;
                }
                set
                {
                    m_mappingFileName = value;
                }
            }

            private readonly Dictionary<string, string> m_defaultSettings = new Dictionary<string, string>()
            {
                {"commsFile", "comms1.xml"},
                {"mappingFile", "device1.xml"}
            };

            public override void Update(Dictionary<string, string> configInfo)
            {
                base.Update(configInfo);
                CommsFileName = DefaultGet("commsFile");
                MappingFileName = DefaultGet("mappingFile"); 
            }

            public override void ExtractConfiguration(DataSet configurationTables)
            {
                base.ExtractConfiguration(configurationTables);
            }

            public AdapterInfo()
            {
                SetDefaults(m_defaultSettings);
            }

         }
        /// <summary>
        /// DNP3 manager shared across all of the DNP3 input adapters
        /// </summary>
        private static StackManager m_Manager = new StackManager();

        /// <summary>
        /// Configuration for the master. Set during the Initialize call.
        /// </summary>
        private MasterConfiguration m_MasterConfig;

        /// <summary>
        /// Configuration for the measurement map. Set during the Initialize call.
        /// </summary>
        private MeasurementMap m_MeasMap;

        /// <summary>
        /// Gets set during the AttemptConnection() call and using during AttemptDisconnect()
        /// </summary>
        private String m_portName;

        /// <summary>
        /// Flag that records wether or not the port/master have been added so that the resource can be cleaned
        /// up in the Dispose/AttemptDisconnect methods
        /// </summary>
        private bool m_active;

        private const string StrLogSource = "Dnp3InputAdapter";
        #endregion

        #region [ Constructors ]

        /// <summary>
        /// Creates a new instance of the <see cref="Dnp3InputAdapter"/> class.
        /// </summary>
        /// <param name="info">Adapter configuration information.</param>
        public Dnp3InputAdapter(IAdapterInfo info)
            : base(info)
        {
            m_statsCluster.RegisterCounter(info.Name, StrStatProcessedMeasurements);
        }

        #endregion

        #region [ Properties ]

        public new AdapterInfo Info
        {
            get
            {
                return ((AdapterInfo)base.Info);
            }
        }

        /// <summary>
        /// Gets or sets the name of the xml file from which the communication parameters will be read
        /// </summary>
        [ConnectionStringParameter,
        Description("Define the name of the XML file from which the communication configuration will be read"),
        DefaultValue("comms1.xml")]
        public string CommsFileName
        {
            get
            {
                return Info.CommsFileName;
            }
            set
            {
                Info.CommsFileName = value;
            }
        }

        /// <summary>
        /// Gets or sets the name of the xml file from which the measurement mapping is read
        /// </summary>
        [ConnectionStringParameter,
        Description("Define the name of the XML file from which the communication configuration will be read"),
        DefaultValue("device1.xml")]
        public string MappingFileName
        {
            get
            {
                return Info.MappingFileName;
            }
            set
            {
                Info.MappingFileName = value;
            }
        }

        #endregion

        #region [ Methods ]

        public override void Initialize()
        {
            base.Initialize();
            string commsFileName = Info.CommsFileName;
            string mappingFileName = Info.MappingFileName;
            Settings.TryGetValue("commsFile", out commsFileName);
            Settings.TryGetValue("mappingFile", out mappingFileName);                                  

            try
            {
                if (Info.CommsFileName == null) throw new ArgumentException("The required commsFile parameter was not specified");
                if (Info.MappingFileName == null) throw new ArgumentException("The required mappingFile parameter was not specified");

                m_MasterConfig = ReadConfig<MasterConfiguration>(CommsFileName);
                m_MeasMap = ReadConfig<MeasurementMap>(MappingFileName);                
            }
            catch (Exception ex)
            {
                LogDispatcher.E(StrLogSource, "ProcessException", "{0}{1}{2}",
                                ex.Message,
                                Environment.NewLine,
                                ex.StackTrace);
            }            
        }

        protected override void Dispose(bool disposing)
        {
            if (m_active)
            {
                m_Manager.RemovePort(m_portName);
                m_active = false;
            }
            base.Dispose(disposing);
        }

        private T ReadConfig<T>(string path)
        {
            var ser = new System.Xml.Serialization.XmlSerializer(typeof(T));
            var stream = new StreamReader(path);
            try
            {
                return (T)ser.Deserialize(stream);
            }
            finally
            {
                stream.Close();
            }
        }
        

        protected override bool UseAsyncConnect
        {
            get { return false; }
        }

        protected override void AttemptConnection()
        {            
            var tcp = m_MasterConfig.client;
            var master = m_MasterConfig.master;
            m_portName = tcp.address + ":" + tcp.port;
            m_Manager.AddTCPClient(m_portName, tcp.level, tcp.retryMs, tcp.address, tcp.port);            
            var adapter = new TimeSeriesDataObserver(new MeasurementLookup(m_MeasMap));
            adapter.NewMeasurements += new TimeSeriesDataObserver.OnNewMeasurements(adapter_NewMeasurements);
            adapter.NewMeasurements += new TimeSeriesDataObserver.OnNewMeasurements(OnNewMeasurements);           
            var acceptor = m_Manager.AddMaster(m_portName, Name, FilterLevel.LEV_WARNING, adapter, m_MasterConfig.master);
            m_active = true;
        }

        void adapter_NewMeasurements(ICollection<ILegacyMeasurement> measurements)
        {
            m_statsCluster.AddToCounter(StrStatReceivedMeasurements, measurements.Count);
        }        

        protected override void AttemptDisconnection()
        {
            if (m_active)
            {                
                m_Manager.RemovePort(m_portName);
                m_active = false;
            }
        }

        public override bool SupportsTemporalProcessing
        {
            get { return false; }
        }

        public override string GetShortStatus(int maxLength)
        {
            return "The adapter has received " + m_statsCluster.GetCounter(StrStatReceivedMeasurements).LastIntervalValue + " measurements";
        }

        #endregion
    }    
    
}
