﻿//******************************************************************************************************
//  DynamicCalculator.cs - Gbtc
//
//  Copyright © 2010, Grid Protection Alliance.  All Rights Reserved.
//
//  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
//  the NOTICE file distributed with this work for additional information regarding copyright ownership.
//  The GPA licenses this file to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  02/06/2012 - Stephen C. Wills
//       Generated original version of source code.
//  12/26/2012 - Prasanna Kumari
//       Applied changes for Migrating adapters from the Time Series Framework to SIEGate Core
//  2/2013 - University of Illinois, Information Trust Institute, Coordinated Science Laboratory
//       Updated code to work for the SIEGate system and architecture
//
//
//******************************************************************************************************

using System.Data;
using Ciloci.Flee;
using GSF;
using GSF.TimeSeries;
using GSF.TimeSeries.Adapters;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;
using System.Text;
using SIEGate.Core;
using SIEGate.Core.Legacy;
using SIEGate.Core.Configurator;
using SIEGate.Core.Measurements.Values;

namespace DynamicCalculator
{
    /// <summary>
    /// The DynamicCalculator is an action adapter which takes multiple
    /// input measurements and performs a calculation on those measurements
    /// to generate its own calculated measurement.
    /// </summary>
    [Description("Dynamic Calculator: Performs arithmetic operations on multiple input signals.")]
    public class DynamicCalculator : ActionAdapterBase
    {
        public class AdapterInfo : ActionAdapterBase.AdapterInfoBase
        {
            public override void Update(Dictionary<string, string> configInfo)
            {
                base.Update(configInfo);
            }

            public override void ExtractConfiguration(DataSet dataSource)
            {
                base.ExtractConfiguration(dataSource);
            }
        }

        #region [ Members ]

        // Fields
        private string m_expressionText;
        private string m_variableList;
        private string m_imports;
        private bool m_supportsTemporalProcessing;
        private string m_waitHandleReleaseName;

        private HashSet<string> m_variableNames;
        private Dictionary<SIEGate.Core.Legacy.MeasurementKey, string> m_keyMapping;
        private SortedDictionary<int, string> m_nonAliasedTokens;

        private string m_aliasedExpressionText;
        private ExpressionContext m_expressionContext;
        private IDynamicExpression m_expression;

        private const string StrLogSource = "DynamicCalculator";
        #endregion

        #region [ Constructors ]

        /// <summary>
        /// Creates a new instance of the <see cref="DynamicCalculator"/>.
        /// </summary>
        public DynamicCalculator(IAdapterInfo info)
            : base(info)
        {
            m_variableNames = new HashSet<string>();
            m_keyMapping = new Dictionary<SIEGate.Core.Legacy.MeasurementKey, string>();
            m_nonAliasedTokens = new SortedDictionary<int, string>();
            m_expressionContext = new ExpressionContext();
        }

        #endregion

        #region [ Properties ]

        /// <summary>
        /// Gets or sets the textual representation of the expression.
        /// </summary>
        [SIEGate.Core.Configurator.ConnectionStringParameter,
        Description("Define the arithmetic expression used to perform calculations.")]
        public string ExpressionText
        {
            get
            {
                return m_expressionText;
            }
            set
            {
                m_expressionText = value;
                PerformAliasReplacement();
            }
        }

        /// <summary>
        /// Gets or sets the list of variables used in the expression.
        /// </summary>
        [SIEGate.Core.Configurator.ConnectionStringParameter,
        Description("Define the list of variables used in the expression.")]
        public string VariableList
        {
            get
            {
                return m_variableList;
            }
            set
            {
                string keyList;

                if (m_variableList != value)
                {
                    // Store the variable list for get operations
                    m_variableList = value;

                    // Empty the collection of variable names
                    m_variableNames.Clear();
                    m_keyMapping.Clear();
                    m_nonAliasedTokens.Clear();

                    // If the value is null, do not attempt to process it
                    if ((object)value == null)
                        return;

                    // Build the collection of variable names with the new value
                    foreach (string token in value.Split(';'))
                        AddVariable(token);

                    // Perform alias replacement on tokens that were not explicitly aliased
                    PerformAliasReplacement();

                    // Build the key list which will define the input measurements for this adapter
                    keyList = m_keyMapping.Keys.Select(key => key.ToString())
                        .Aggregate((runningKeyList, nextKey) => runningKeyList + ";" + nextKey);

                    // Set the input measurements for this adapter
                    InputMeasurementKeys = LegacyAdapterBase.ParseInputMeasurementKeys(DataSource, keyList);
                }
            }
        }

        /// <summary>
        /// Gets or sets the list of types which define
        /// methods to be imported into the expression parser.
        /// </summary>
        [SIEGate.Core.Configurator.ConnectionStringParameter,
        Description("Define the list of types which define methods to be imported into the expression parser."),
        DefaultValue("AssemblyName={mscorlib, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089}, TypeName=System.Math")]
        public string Imports
        {
            get
            {
                return m_imports;
            }
            set
            {
                Dictionary<string, string> parsedTypeDef;
                string assemblyName, typeName;
                Assembly asm;
                Type t;

                foreach (string typeDef in value.Split(';'))
                {
                    try
                    {
                        parsedTypeDef = typeDef.ParseKeyValuePairs(',');
                        assemblyName = parsedTypeDef["assemblyName"];
                        typeName = parsedTypeDef["typeName"];
                        asm = Assembly.Load(new AssemblyName(assemblyName));
                        t = asm.GetType(typeName);

                        m_expressionContext.Imports.AddType(t);
                    }
                    catch(Exception ex)
                    {
                        LogDispatcher.E(StrLogSource,
                                        "ProcessException",
                                        "Unable to load type from assembly: {0}{1}{2}{3}",
                                        typeDef,
                                        ex.Message,
                                        Environment.NewLine,
                                        ex.StackTrace);
                    }
                }

                m_imports = value;
            }
        }

        /// <summary>
        /// Gets the flag indicating if this adapter supports temporal processing.
        /// </summary>
        [SIEGate.Core.Configurator.ConnectionStringParameter,
        Description("Define the flag indicating if this adapter supports temporal processing."),
        DefaultValue(false)]
        public override bool SupportsTemporalProcessing
        {
            get
            {
                return m_supportsTemporalProcessing;
            }
        }

        /// <summary>
        /// Gets or sets the name of the wait handle to be released when
        /// the calculation is finished. The default value of null indicates
        /// that there will be no wait handle sent and therefore no adapter
        /// synchronization.
        /// </summary>
        [SIEGate.Core.Configurator.ConnectionStringParameter,
        Description("Define the name of the wait handle to be released for adapter synchronization."),
        DefaultValue(null)]
        public string WaitHandleReleaseName
        {
            get
            {
                return m_waitHandleReleaseName;
            }
            set
            {
                m_waitHandleReleaseName = value;
            }
        }

        #endregion

        #region [ Methods ]

        /// <summary>
        /// Initializes <see cref="DynamicCalculator"/>.
        /// </summary>
        public override void Initialize()
        {
            base.Initialize();

            const string errorMessage = "{0} is missing from Settings - Example: expressionText=x+y; variableList={x = PPA:1; y = PPA:2}";

            Dictionary<string, string> settings;
            string setting;
            settings = Settings;

            if (OutputMeasurements.Length != 1)
                throw new ArgumentException(string.Format("Exactly one output measurement must be defined. Amount defined: {0}", OutputMeasurements.Length));

            // Load required parameters

            if (!settings.TryGetValue("expressionText", out setting))
                throw new ArgumentException(string.Format(errorMessage, "expressionText"));
            ExpressionText = settings["expressionText"];

            if (!settings.TryGetValue("variableList", out setting))
                throw new ArgumentException(string.Format(errorMessage, "variableList"));
            VariableList = settings["variableList"];

            // Load optional parameters

            if (settings.TryGetValue("imports", out setting))
                Imports = setting;
            else
                Imports = "AssemblyName={mscorlib, Version=2.0.0.0, Culture=neutral, PublicKeyToken=b77a5c561934e089}, TypeName=System.Math";

            if (settings.TryGetValue("supportsTemporalProcessing", out setting))
                m_supportsTemporalProcessing = setting.ParseBoolean();
            else
                m_supportsTemporalProcessing = false;

            if (settings.TryGetValue("waitHandleReleaseName", out setting))
                m_waitHandleReleaseName = setting;
            else
                m_waitHandleReleaseName = null;
        }

        /// <summary>
        /// Publish <see cref="IFrame"/> of time-aligned collection of <see cref="IMeasurement"/> values that arrived within the
        /// concentrator's defined <see cref="ConcentratorBase.LagTime"/>.
        /// </summary>
        /// <param name="frame"><see cref="IFrame"/> of measurements with the same timestamp that arrived within <see cref="ConcentratorBase.LagTime"/> that are ready for processing.</param>
        /// <param name="index">Index of <see cref="IFrame"/> within a second ranging from zero to <c><see cref="ConcentratorBase.FramesPerSecond"/> - 1</c>.</param>
        protected override void PublishFrame(GSF.TimeSeries.IFrame frame, int index)
        {
            ConcurrentDictionary<SIEGate.Core.Legacy.MeasurementKey, ILegacyMeasurement> measurements;
            ILegacyMeasurement measurement;
            string name;

            measurements = frame.Measurements;
            m_expressionContext.Variables.Clear();

            // Set the values of variables in the expression
            foreach (SIEGate.Core.Legacy.MeasurementKey key in m_keyMapping.Keys)
            {
                name = m_keyMapping[key];

                if (measurements.TryGetValue(key, out measurement))
                    m_expressionContext.Variables[name] = measurement.AdjustedValue;
            }

            // Compile the expression if it has not been compiled already
            if ((object)m_expression == null)
                m_expression = m_expressionContext.CompileDynamic(m_expressionText);

            // Evaluate the expression and generate the measurement
            GenerateCalculatedMeasurement(m_expression.Evaluate() as IConvertible);

            // Release wait handle for adapter synchronization
            if ((object)m_waitHandleReleaseName != null)
                GetExternalEventHandle(m_waitHandleReleaseName).Set();
        }

        // Adds a variable to the key-variable map.
        private void AddVariable(string token)
        {
            // This determines whether the variable has been
            // explicitly aliased or not and then delegates
            // the work to the appropriate helper method
            if (token.Contains('='))
                AddAliasedVariable(token);
            else
                AddNotAliasedVariable(token);
        }

        // Adds an explicitly aliased variable to the key-variable map.
        private void AddAliasedVariable(string token)
        {
            string[] splitToken = token.Split('=');
            SIEGate.Core.Legacy.MeasurementKey key;
            string alias;

            if (splitToken.Length > 2)
                throw new FormatException(string.Format("Too many equals signs: {0}", token));

            key = GetKey(splitToken[1].Trim());
            alias = splitToken[0].Trim();
            AddMapping(key, alias);
        }

        // Adds a variable to the key-variable map which has not been explicitly aliased.
        private void AddNotAliasedVariable(string token)
        {
            string alias;
            SIEGate.Core.Legacy.MeasurementKey key;

            token = token.Trim();
            m_nonAliasedTokens.Add(-token.Length, token);

            key = GetKey(token);
            alias = token.ReplaceCharacters('_', c => !char.IsLetterOrDigit(c));

            // Ensure that the generated alias is unique
            while (m_variableNames.Contains(alias))
                alias += "_";

            AddMapping(key, alias);
        }

        // Adds the given mapping to the key-variable map.
        private void AddMapping(SIEGate.Core.Legacy.MeasurementKey key, string alias)
        {
            if (m_variableNames.Contains(alias))
                throw new ArgumentException(string.Format("Variable name is not unique: {0}", alias));

            m_variableNames.Add(alias);
            m_keyMapping.Add(key, alias);
        }

        // Performs alias replacement on tokens that were not explicitly aliased.
        private void PerformAliasReplacement()
        {
            StringBuilder aliasedExpressionTextBuilder = new StringBuilder(m_expressionText);
            SIEGate.Core.Legacy.MeasurementKey key;
            string alias;

            foreach (string token in m_nonAliasedTokens.Values)
            {
                key = GetKey(token);
                alias = m_keyMapping[key];
                aliasedExpressionTextBuilder.Replace(token, alias);
            }

            m_aliasedExpressionText = aliasedExpressionTextBuilder.ToString();
        }

        // Gets a measurement key based on a token which
        // may be either a signal ID or measurement key.
        private SIEGate.Core.Legacy.MeasurementKey GetKey(string token)
        {
            SIEGate.Core.Legacy.MeasurementKey undefined;
            Guid signalId;

            SIEGate.Core.Legacy.MeasurementKey key;

            if (Guid.TryParse(token, out signalId))
            {
                // Defined using the measurement's GUID
                undefined = SIEGate.Core.Legacy.MeasurementKey.Undefined;
                key = new SIEGate.Core.Legacy.MeasurementKey(signalId, undefined.ID, undefined.Source);
            }
            else
            {
                // Defined using the measurement's key
                key = SIEGate.Core.Legacy.MeasurementKey.Parse(token);
            }

            return key;
        }

        // Generates a measurement with the given value and sends it into the system
        private void GenerateCalculatedMeasurement(IConvertible value)
        {
            ILegacyMeasurement calculatedMeasurement;

            if ((object)value == null)
                throw new InvalidOperationException("Calculation must not return a type that does not convert to double.");

            calculatedMeasurement = LegacyMeasurement.Clone(OutputMeasurements[0], Convert.ToDouble(value), DateTime.UtcNow.Ticks);
            OnNewMeasurement(calculatedMeasurement);
        }

        // Helper method to raise the NewMeasurements event
        // when only a single measurement is to be provided.
        private void OnNewMeasurement(ILegacyMeasurement measurement)
        {
            OnNewMeasurements(new ILegacyMeasurement[] { measurement });
        }

        #endregion
    }
}
