﻿//******************************************************************************************************
//  FileBlockReader.cs - Gbtc
//
//  Copyright © 2013, Grid Protection Alliance.  All Rights Reserved.
//
//  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
//  the NOTICE file distributed with this work for additional information regarding copyright ownership.
//  The GPA licenses this file to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  04/25/2013 - Stephen C. Wills
//       Generated original version of source code.
//
//******************************************************************************************************

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.IO;
using System.Linq;
using System.Text;
using System.Timers;
using GSF;
using GSF.IO;
using SIEGate.Core;
using SIEGate.Core.Adapters;
using SIEGate.Core.Configurator;
using SIEGate.Core.Legacy;
using SIEGate.Core.Measurements;
using SIEGate.Core.Measurements.Values;

namespace FileAdapters
{
    /// <summary>
    /// Action adapter that reads blocks of data into buffers and publishes them as measurements.
    /// </summary>
    [Description("FileBlockReader: Reads blocks of data from files and publishes them as buffer block measurements.")]
    public class FileBlockReader : ActionAdapter
    {
        #region [ Members ]

        // Nested Types

        public class AdapterInfo : AdapterInfoBase
        {
            #region [ Members ]

            // Constants
            private readonly Dictionary<string, string> m_defaults = new Dictionary<string, string>(StringComparer.CurrentCultureIgnoreCase)
            {
                { "filePattern", "*" },
                { "blockSize", "16384" },
                { "watchInterval", "5.0" },
                { "processInterval", "1.0" },
                { "retransmissionStat", "" },
                { "retransmissionThreshold", "25.0" },
                { "blockSizeAdjustment", "5.0" },
                { "processIntervalAdjustment", "5.0" }
            };

            // Fields
            private string m_watchDirectory;
            private string m_filePattern;
            private int m_blockSize;
            private double m_watchInterval;
            private double m_processInterval;
            private double m_retransmissionThreshold;
            private double m_blockSizeAdjustment;
            private double m_processIntervalAdjustment;

            private DataSet m_configurationTables;

            #endregion

            #region [ Constructors ]

            public AdapterInfo()
            {
                SetDefaults(m_defaults);
            }

            #endregion

            #region [ Properties ]

            /// <summary>
            /// Gets or sets the path to the directory to be watched for files.
            /// </summary>
            [ConnectionStringParameter,
            Description("Defines the path to the directory to be watched for files.")]
            public string WatchDirectory
            {
                get
                {
                    return m_watchDirectory;
                }
                set
                {
                    m_watchDirectory = value;
                }
            }

            /// <summary>
            /// Gets or sets the pattern used to match file that appear in the watch folder.
            /// </summary>
            [ConnectionStringParameter,
            DefaultValue(DefaultFilePattern),
            Description("Determines which files are to processed when they appear in the watch folder.")]
            public string FilePattern
            {
                get
                {
                    return m_filePattern;
                }
                set
                {
                    m_filePattern = value;
                }
            }

            /// <summary>
            /// Gets or sets the statistic that defines the number of buffer block retransmissions in the system.
            /// </summary>
            [ConnectionStringParameter,
            DefaultValue(""),
            Description("Determines the statistic that defines the number of buffer block retransmissions in the system.")]
            public string RetransmissionStat
            {
                get
                {
                    return IncomingStreamIDs.First().ToString();
                }
                set
                {
                    IncomingStreamIDs = new HashSet<Guid>(GsfHandler.ParseStreamIDs(m_configurationTables, value).Take(1));
                }
            }

            /// <summary>
            /// Gets or sets the size of each block of data that is read from the file.
            /// </summary>
            [ConnectionStringParameter,
            DefaultValue(DefaultBlockSize),
            Description("Determines the size of each block of data that is read from the file.")]
            public int BlockSize
            {
                get
                {
                    return m_blockSize;
                }
                set
                {
                    m_blockSize = value;
                }
            }

            /// <summary>
            /// Gets or sets the amount of time, in seconds, between each scan of the watch folder.
            /// </summary>
            [ConnectionStringParameter,
            DefaultValue(DefaultWatchInterval),
            Description("Determines the amount of time, in seconds, between each scan of the watch folder.")]
            public double WatchInterval
            {
                get
                {
                    return m_watchInterval;
                }
                set
                {
                    m_watchInterval = value;
                }
            }

            /// <summary>
            /// Gets or sets the amount of time, in seconds, between each publication of a block of data from the active file.
            /// </summary>
            [ConnectionStringParameter,
            DefaultValue(DefaultProcessInterval),
            Description("Determines the amount of time, in seconds, between each publication of a block of data from the active file.")]
            public double ProcessInterval
            {
                get
                {
                    return m_processInterval;
                }
                set
                {
                    m_processInterval = value;
                }
            }

            /// <summary>
            /// Gets or sets the percentage of buffer blocks that can be retransmitted
            /// during a reporting interval before the <see cref="FileBlockReader"/>
            /// begins throttling its buffer blocks.
            /// </summary>
            [ConnectionStringParameter,
            DefaultValue(DefaultRetransmissionThreshold),
            Description("Defines the percentage of buffer blocks that can be retransmitted before throttling begins.")]
            public double RetransmissionThreshold
            {
                get
                {
                    return m_retransmissionThreshold;
                }
                set
                {
                    m_retransmissionThreshold = value;
                }
            }

            /// <summary>
            /// Gets or sets the percentage of adjustment to be applied
            /// to the buffer size when throttling buffer blocks.
            /// </summary>
            [ConnectionStringParameter,
            DefaultValue(DefaultBlockSizeAdjustment),
            Description("Defines the percentage by which to adjust the buffer size when throttling buffer blocks.")]
            public double BlockSizeAdjustment
            {
                get
                {
                    return m_blockSizeAdjustment;
                }
                set
                {
                    m_blockSizeAdjustment = value;
                }
            }

            /// <summary>
            /// Gets or sets the percentage of adjustment to be applied
            /// to the process interval when throttling buffer blocks.
            /// </summary>
            [ConnectionStringParameter,
            DefaultValue(DefaultBlockSizeAdjustment),
            Description("Defines the percentage by which to adjust the process interval when throttling buffer blocks.")]
            public double ProcessIntervalAdjustment
            {
                get
                {
                    return m_processIntervalAdjustment;
                }
                set
                {
                    m_processIntervalAdjustment = value;
                }
            }

            #endregion

            #region [ Methods ]

            public override void ExtractConfiguration(DataSet configurationTables)
            {
                const string errorMessage = @"{0} is missing from Settings - Example: watchDirectory=C:\Files; outputMeasurements=FILES:20";

                Dictionary<string, string> settings;
                string setting;
                int blockSize;
                double watchInterval;
                double processInterval;
                double retransmissionThreshold;
                double blockSizeAdjustment;
                double processIntervalAdjustment;

                base.ExtractConfiguration(configurationTables);
                m_configurationTables = configurationTables;
                settings = Settings;

                // Required parameters

                if (!settings.TryGetValue("watchDirectory", out setting))
                    throw new ArgumentException(string.Format(errorMessage, "watchDirectory"));

                m_watchDirectory = FilePath.GetAbsolutePath(setting);

                if (OutgoingStreamIDs.Count <= 0)
                    throw new ArgumentException(string.Format(errorMessage, "outputMeasurements"));

                // Optional parameters

                m_filePattern = DefaultGet("filePattern");
                RetransmissionStat = DefaultGet("retransmissionStat");

                if (int.TryParse(DefaultGet("blockSize"), out blockSize))
                    BlockSize = blockSize;

                if (double.TryParse(DefaultGet("watchInterval"), out watchInterval))
                    m_watchInterval = watchInterval;

                if (double.TryParse(DefaultGet("processInterval"), out processInterval))
                    m_processInterval = processInterval;

                if (double.TryParse(DefaultGet("retransmissionThreshold"), out retransmissionThreshold))
                    m_retransmissionThreshold = retransmissionThreshold;

                if (double.TryParse(DefaultGet("blockSizeAdjustment"), out blockSizeAdjustment))
                    m_blockSizeAdjustment = blockSizeAdjustment;

                if (double.TryParse(DefaultGet("processIntervalAdjustment"), out processIntervalAdjustment))
                    m_processIntervalAdjustment = processIntervalAdjustment;

                if (!Directory.Exists(m_watchDirectory))
                    Directory.CreateDirectory(m_watchDirectory);
            }

            #endregion
        }

        // Constants

        /// <summary>
        /// Default value for <see cref="FilePattern"/>.
        /// </summary>
        public const string DefaultFilePattern = "*";

        /// <summary>
        /// Default value for <see cref="BlockSize"/>.
        /// </summary>
        public const int DefaultBlockSize = 16384;

        /// <summary>
        /// Default value for <see cref="WatchInterval"/>.
        /// </summary>
        public const double DefaultWatchInterval = 5.0D;

        /// <summary>
        /// Default value for <see cref="ProcessInterval"/>.
        /// </summary>
        public const double DefaultProcessInterval = 1.0D;

        /// <summary>
        /// Default value for <see cref="RetransmissionThreshold"/>.
        /// </summary>
        public const double DefaultRetransmissionThreshold = 25.0D;

        /// <summary>
        /// Default value for <see cref="BlockSizeAdjustment"/>.
        /// </summary>
        public const double DefaultBlockSizeAdjustment = 5.0D;

        /// <summary>
        /// Default value for <see cref="ProcessIntervalAdjustment"/>.
        /// </summary>
        public const double DefaultProcessIntervalAdjustment = 5.0D;

        // Fields
        private readonly List<string> m_processedFiles;
        private readonly Queue<string> m_unprocessedFiles;
        private FileStream m_activeFileStream;
        private Timer m_watchTimer;
        private Timer m_processTimer;
        private byte[] m_buffer;
        [Obsolete("Statistic")]
        private long m_bufferBlocksSent;
        private long m_bufferBlocksSentLastAdjustment;
        private int m_throttleMultiplier;

        private bool m_enabled;
        private bool m_disposed;
        private const string StrStatProcessedFiles = "Processed Files";
        private const string StrStatQueuedFiles = "Queued Files";
        #endregion

        #region [ Constructors ]

        /// <summary>
        /// Creates a new instance of the <see cref="FileBlockReader"/> class.
        /// </summary>
        public FileBlockReader(AdapterInfo info)
            : base(info)
        {
            m_processedFiles = new List<string>();
            m_unprocessedFiles = new Queue<string>();
            m_statsCluster.RegisterCounter(info.Name, StrStatProcessedFiles);
            m_statsCluster.RegisterDelegateCounter(info.Name, StrStatQueuedFiles, () => m_unprocessedFiles.Count);
            Initialize();
            Start();
        }

        /// <summary>
        /// Releases the unmanaged resources before the <see cref="FileBlockReader"/> object is reclaimed by <see cref="GC"/>.
        /// </summary>
        ~FileBlockReader()
        {
            Dispose(false);
        }

        #endregion

        #region [ Properties ]

        /// <summary>
        /// Gets or sets the adapter info for this adapter.
        /// </summary>
        public new AdapterInfo Info
        {
            get
            {
                return (AdapterInfo)base.Info;
            }
            set
            {
                base.Info = value;
                m_buffer = new byte[AdjustedBlockSize];

                if ((object)m_watchTimer != null)
                    m_watchTimer.Interval = WatchInterval * 1000.0D;

                if ((object)m_processTimer != null)
                    m_processTimer.Interval = ProcessInterval * 1000.0D;
            }
        }

        /// <summary>
        /// Gets or sets the path to the directory to be watched for files.
        /// </summary>
        [ConnectionStringParameter,
        Description("Defines the path to the directory to be watched for files.")]
        public string WatchDirectory
        {
            get
            {
                return Info.WatchDirectory;
            }
            set
            {
                Info.WatchDirectory = value;
            }
        }

        /// <summary>
        /// Gets or sets the pattern used to match file that appear in the watch folder.
        /// </summary>
        [ConnectionStringParameter,
        DefaultValue(DefaultFilePattern),
        Description("Determines which files are to processed when they appear in the watch folder.")]
        public string FilePattern
        {
            get
            {
                return Info.FilePattern;
            }
            set
            {
                Info.FilePattern = value;
            }
        }

        /// <summary>
        /// Gets or sets the statistic that defines the number of buffer block retransmissions in the system.
        /// </summary>
        [ConnectionStringParameter,
        DefaultValue(null),
        Description("Determines the statistic that defines the number of buffer block retransmissions in the system.")]
        public string RetransmissionStat
        {
            get
            {
                return Info.RetransmissionStat;
            }
            set
            {
                Info.RetransmissionStat = value;
            }
        }

        /// <summary>
        /// Gets or sets the size of each block of data that is read from the file.
        /// </summary>
        [ConnectionStringParameter,
        DefaultValue(DefaultBlockSize),
        Description("Determines the size of each block of data that is read from the file.")]
        public int BlockSize
        {
            get
            {
                return Info.BlockSize;
            }
            set
            {
                Info.BlockSize = value;
                m_buffer = new byte[AdjustedBlockSize];
            }
        }

        /// <summary>
        /// Gets or sets the amount of time, in seconds, between each scan of the watch folder.
        /// </summary>
        [ConnectionStringParameter,
        DefaultValue(DefaultWatchInterval),
        Description("Determines the amount of time, in seconds, between each scan of the watch folder.")]
        public double WatchInterval
        {
            get
            {
                return Info.WatchInterval;
            }
            set
            {
                Info.WatchInterval = value;

                if ((object)m_watchTimer != null)
                    m_watchTimer.Interval = value * 1000.0D;
            }
        }

        /// <summary>
        /// Gets or sets the amount of time, in seconds, between each publication of a block of data from the active file.
        /// </summary>
        [ConnectionStringParameter,
        DefaultValue(DefaultProcessInterval),
        Description("Determines the amount of time, in seconds, between each publication of a block of data from the active file.")]
        public double ProcessInterval
        {
            get
            {
                return Info.ProcessInterval;
            }
            set
            {
                Info.ProcessInterval = value;

                if ((object)m_processTimer != null)
                    m_processTimer.Interval = AdjustedProcessInterval * 1000.0D;
            }
        }

        /// <summary>
        /// Gets or sets the percentage of buffer blocks that can be retransmitted
        /// during a reporting interval before the <see cref="FileBlockReader"/>
        /// begins throttling its buffer blocks.
        /// </summary>
        [ConnectionStringParameter,
        DefaultValue(DefaultRetransmissionThreshold),
        Description("Defines the percentage of buffer blocks that can be retransmitted before throttling begins.")]
        public double RetransmissionThreshold
        {
            get
            {
                return Info.RetransmissionThreshold;
            }
            set
            {
                Info.RetransmissionThreshold = value;
            }
        }

        /// <summary>
        /// Gets or sets the percentage of adjustment to be applied
        /// to the buffer size when throttling buffer blocks.
        /// </summary>
        [ConnectionStringParameter,
        DefaultValue(DefaultBlockSizeAdjustment),
        Description("Defines the percentage by which to adjust the buffer size when throttling buffer blocks.")]
        public double BlockSizeAdjustment
        {
            get
            {
                return Info.BlockSizeAdjustment;
            }
            set
            {
                Info.BlockSizeAdjustment = value;
            }
        }

        /// <summary>
        /// Gets or sets the percentage of adjustment to be applied
        /// to the process interval when throttling buffer blocks.
        /// </summary>
        [ConnectionStringParameter,
        DefaultValue(DefaultProcessIntervalAdjustment),
        Description("Defines the percentage by which to adjust the process interval when throttling buffer blocks.")]
        public double ProcessIntervalAdjustment
        {
            get
            {
                return Info.ProcessIntervalAdjustment;
            }
            set
            {
                Info.ProcessIntervalAdjustment = value;
            }
        }

        /// <summary>
        /// Gets the block size after making adjustments for throttling.
        /// </summary>
        private int AdjustedBlockSize
        {
            get
            {
                double blockSizeAdjustment = BlockSize * (BlockSizeAdjustment * 0.01);
                return BlockSize - (int)(m_throttleMultiplier * blockSizeAdjustment);
            }
        }

        /// <summary>
        /// Gets the process interval after making adjustments for throttling.
        /// </summary>
        private double AdjustedProcessInterval
        {
            get
            {
                double processIntervalAdjustment = ProcessInterval * (ProcessIntervalAdjustment * 0.01);
                return ProcessInterval + (m_throttleMultiplier * processIntervalAdjustment);
            }
        }

        #endregion

        #region [ Methods ]

        /// <summary>
        /// Initializes <see cref="FileBlockReader"/>.
        /// </summary>
        public void Initialize()
        {
            m_buffer = new byte[AdjustedBlockSize];

            m_watchTimer = new Timer();
            m_watchTimer.AutoReset = false;
            m_watchTimer.Interval = WatchInterval * 1000.0D;
            m_watchTimer.Elapsed += WatchTimer_Elapsed;

            m_processTimer = new Timer();
            m_processTimer.AutoReset = false;
            m_processTimer.Interval = ProcessInterval * 1000.0D;
            m_processTimer.Elapsed += ProcessTimer_Elapsed;
        }

        /// <summary>
        /// Gets a short one-line status of this <see cref="FileBlockReader"/>.
        /// </summary>
        /// <param name="maxLength">Maximum number of available characters for display.</param>
        /// <returns>A short one-line summary of the current status of this <see cref="FileBlockReader"/>.</returns>
        public string GetShortStatus(int maxLength)
        {
            if ((object)m_activeFileStream != null)
                return string.Format("Currently reading from file {0}", Path.GetFileName(m_unprocessedFiles.Peek())).CenterText(maxLength);

            return string.Format("{0} files processed by {1}", m_statsCluster.GetCounterValue(StrStatProcessedFiles), Info.Name).CenterText(maxLength);
        }

        /// <summary>
        /// Empties the processed file list.
        /// </summary>
        [AdapterCommand("Empties the processed file list so that newly dropped files with the same name as old processed files will not be deleted.")]
        public void ClearProcessedFileList()
        {
            m_processedFiles.Clear();
        }

        /// <summary>
        /// Starts the <see cref="FileBlockReader"/> or restarts it if it is already running.
        /// </summary>
        public void Start()
        {
            m_enabled = true;
            m_watchTimer.Start();
            m_processTimer.Start();
        }

        /// <summary>
        /// Stops the <see cref="FileBlockReader"/>.
        /// </summary>		
        public void Stop()
        {
            m_enabled = false;
            m_watchTimer.Stop();
            m_processTimer.Stop();

            if ((object)m_activeFileStream != null)
            {
                m_activeFileStream.Dispose();
                m_activeFileStream = null;
            }
        }

        protected override void HandleMeasurement(IMeasurement m)
        {
            ManageThrottleAdjustments((int)m.Value.GetRaw());
        }

        // Determines how to adjust the process interval and block size based on the
        // number of retransmissions during the last statistic reporting interval
        private void ManageThrottleAdjustments(int retransmissions)
        {
            long bufferBlocksSent = m_bufferBlocksSent;
            int throttleMultiplier = m_throttleMultiplier;
            int bufferBlocksSentSinceLastAdjustment = (int)(bufferBlocksSent - m_bufferBlocksSentLastAdjustment);

            // Adjust the throttle multiplier based on the number of retransmissions
            if (retransmissions >= bufferBlocksSentSinceLastAdjustment * (RetransmissionThreshold * 0.01))
                m_throttleMultiplier++;
            else if (retransmissions < bufferBlocksSentSinceLastAdjustment * (RetransmissionThreshold * 0.005))
                m_throttleMultiplier--;

            // Throttle multiplier cannot be less than 0
            if (m_throttleMultiplier < 0)
                m_throttleMultiplier = 0;

            // If the throttle multiplier changed, update the block size and process interval
            if (throttleMultiplier != m_throttleMultiplier)
            {
                m_buffer = new byte[AdjustedBlockSize];
                m_processTimer.Interval = AdjustedProcessInterval * 1000.0D;
            }

            // Keep track of the buffer blocks sent for the next time we make adjustments
            m_bufferBlocksSentLastAdjustment = bufferBlocksSent;
        }

        // Scans the watch folder for new files to transfer
        private void WatchTimer_Elapsed(object sender, ElapsedEventArgs elapsedEventArgs)
        {
            if (m_enabled)
            {
                // Scan each file and add them to the unprocessed files lists
                foreach (string file in FilePath.GetFileList(Path.Combine(WatchDirectory, FilePattern)))
                {
                    if (!m_unprocessedFiles.Contains(file))
                        m_unprocessedFiles.Enqueue(file);
                }

                // Done scanning, so start the timer for another scan
                if ((object)m_watchTimer != null)
                    m_watchTimer.Start();
            }
        }

        // Reads the next block from the active file
        private void ProcessTimer_Elapsed(object sender, ElapsedEventArgs elapsedEventArgs)
        {
            byte[] buffer = m_buffer;
            int bytesRead = 0;
            int fileNameByteLength;
            FileInfo fileInfo;

            Guid streamID;
            Ticks timestamp;
            MeasurementStateFlags flags;
            IMeasurementValue bufferBlock;

            if (m_enabled)
            {
                if ((object)m_activeFileStream != null)
                {
                    // Set first byte to 0 to indicate that
                    // this is not the beginning of a file
                    buffer[0] = 0;

                    // Read a block into memory
                    bytesRead = 1 + m_activeFileStream.Read(buffer, 1, buffer.Length - 1);

                    if (bytesRead == 1)
                    {
                        // Notify that processing is done for the current file
                        LogDispatcher.I("FileBlockReader", "", "Done processing file {0}.", Path.GetFileName(m_unprocessedFiles.Peek()));

                        // Delete the now-processed file
                        m_activeFileStream.Dispose();
                        m_activeFileStream = null;
                        File.Delete(m_unprocessedFiles.Peek());

                        // Move unprocessed file into processed file list
                        m_processedFiles.Add(m_unprocessedFiles.Peek());
                        m_unprocessedFiles.Dequeue();

                        m_statsCluster.AddToCounter(StrStatProcessedFiles, 1);
                    }
                }

                if ((object)m_activeFileStream == null && m_unprocessedFiles.Count > 0)
                {
                    // Notify that processing has started for a new file
                    LogDispatcher.I("FileBlockReader", "", "Now processing file {0}...", Path.GetFileName(m_unprocessedFiles.Peek()));

                    // Get info about the next file to process
                    fileInfo = new FileInfo(m_unprocessedFiles.Peek());

                    // Set first byte to 1 to indicate that
                    // this is the beginning of a file
                    buffer[0] = 1;

                    // Copy file info into the buffer
                    fileNameByteLength = Encoding.Unicode.GetByteCount(fileInfo.Name);
                    EndianOrder.BigEndian.CopyBytes(fileNameByteLength, buffer, 1);
                    Encoding.Unicode.GetBytes(fileInfo.Name, 0, fileInfo.Name.Length, buffer, 5);
                    EndianOrder.BigEndian.CopyBytes(fileInfo.Length, buffer, 5 + fileNameByteLength);
                    bytesRead = 1 + 4 + fileNameByteLength + 8;

                    // Wait for lock to open the file
                    FilePath.WaitForReadLock(fileInfo.FullName);

                    // Open file and read
                    m_activeFileStream = fileInfo.OpenRead();
                    bytesRead += m_activeFileStream.Read(buffer, bytesRead, buffer.Length - bytesRead);
                }

                if ((object)m_activeFileStream != null)
                {
                    streamID = OutgoingStreamIDs.First();
                    timestamp = DateTime.UtcNow.Ticks;
                    flags = MeasurementStateFlags.Normal;
                    bufferBlock = new BufferBlockValue(buffer, 0, bytesRead);

                    // Publish next block of file data
                    OnEmit(new Measurement(streamID, timestamp, flags, bufferBlock));
                    m_bufferBlocksSent++;
                }

                // Done reading, so start the timer for another read
                if ((object)m_processTimer != null)
                    m_processTimer.Start();
            }
        }

        /// <summary>
        /// Releases all the resources used by the <see cref="FileBlockReader"/> object.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Releases the unmanaged resources used by the <see cref="FileBlockReader"/> object and optionally releases the managed resources.
        /// </summary>
        /// <param name="disposing">true to release both managed and unmanaged resources; false to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (!m_disposed)
            {
                try
                {
                    if (disposing)
                    {
                        if ((object)m_activeFileStream != null)
                        {
                            m_activeFileStream.Dispose();
                            m_activeFileStream = null;
                        }

                        if ((object)m_watchTimer != null)
                        {
                            m_watchTimer.Stop();
                            m_watchTimer.Dispose();
                            m_watchTimer = null;
                        }

                        if ((object)m_processTimer != null)
                        {
                            m_processTimer.Stop();
                            m_processTimer.Dispose();
                            m_processTimer = null;
                        }
                    }
                }
                finally
                {
                    m_disposed = true;  // Prevent duplicate dispose.
                }
            }
        }

        #endregion
    }
}
