﻿//******************************************************************************************************
//  CalculatedMeasurementBase.cs - Gbtc
//
//  Copyright © 2012, Grid Protection Alliance.  All Rights Reserved.
//
//  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
//  the NOTICE file distributed with this work for additional information regarding copyright ownership.
//  The GPA licenses this file to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  10/19/2009 - J. Ritchie Carroll
//       Generated original version of source code.
//  04/21/2010 - J. Ritchie Carroll
//       Added signal type summary to the calculated measurement status.
//  12/04/2012 - J. Ritchie Carroll
//       Migrated to PhasorProtocolAdapters project.
//  12/26/2012 - Prasanna Kumari
//       Applied changes for Migrating adapters from the Time Series Framework to SIEGate Core
//  2/2013 - University of Illinois, Information Trust Institute, Coordinated Science Laboratory
//       Updated code to work for the SIEGate system and architecture
//
//
//******************************************************************************************************

using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Text;
using GSF;
using GSF.PhasorProtocols;
using GSF.TimeSeries;
using GSF.TimeSeries.Adapters;
using SIEGate.Core;
using SIEGate.Core.Legacy;
using SIEGate.Core.Configurator;
using SIEGate.Core.Measurements.Values;
using SIEGate.Core.Stats;


namespace PhasorProtocolAdapters
{
    #region [ Enumerations ]

    /// <summary>
    /// Signal type enumeration.
    /// </summary>
    /// <remarks>
    /// <para>
    /// The signal type represents the explicit type of a signal that a value represents.
    /// </para>
    /// <para>
    /// Contrast the <see cref="SignalType"/> enumeration with the <see cref="SignalKind"/>
    /// enumeration which defines an abstract type for a signal (e.g., simply phase or angle).
    /// </para>
    /// </remarks>
    [Serializable()]
    public enum SignalType
    {
        /// <summary>
        /// Current phase magnitude.
        /// </summary>
        IPHM = 1,
        /// <summary>
        /// Current phase angle.
        /// </summary>
        IPHA = 2,
        /// <summary>
        /// Voltage phase magnitude.
        /// </summary>
        VPHM = 3,
        /// <summary>
        /// Voltage phase angle.
        /// </summary>
        VPHA = 4,
        /// <summary>
        /// Frequency.
        /// </summary>
        FREQ = 5,
        /// <summary>
        /// Frequency delta (dF/dt).
        /// </summary>
        DFDT = 6,
        /// <summary>
        /// Analog value.
        /// </summary>
        ALOG = 7,
        /// <summary>
        /// Status flags.
        /// </summary>
        FLAG = 8,
        /// <summary>
        /// Digital value.
        /// </summary>
        DIGI = 9,
        /// <summary>
        /// Calculated value.
        /// </summary>
        CALC = 10,
        /// <summary>
        /// Statistical value.
        /// </summary>
        STAT = 11,
        /// <summary>
        /// Alarm value.
        /// </summary>
        ALRM = 12,
        /// <summary>
        /// Undefined signal.
        /// </summary>
        NONE = -1
    }

    #endregion

    /// <summary>
    /// Represents the base class for calculated measurements that use phasor data.
    /// </summary>
    /// <remarks>
    /// This base class extends <see cref="ActionAdapterBase"/> by automatically looking up the
    /// <see cref="SignalType"/> for each of the input and output measurements.
    /// </remarks>
    public abstract class CalculatedMeasurementBase : ActionAdapterBase
    {
        #region [ Nested Classes ]
        public class AdapterInfo : ActionAdapterBase.AdapterInfoBase
        {
            private readonly Dictionary<string, string> m_defaultSettings = new Dictionary<string, string>()
            {
                {"supportsTemporalProcessing", "false"},
                {"configurationSection", null}
            };

            private string m_configurationSection = null;
            private SignalType[] m_inputMeasurementKeyTypes;
            private SignalType[] m_outputMeasurementTypes;

            public bool SupportsTemporalProcessing { get; set; }

            public string ConfigurationSection
            {
                get
                {
                    // Default value is Name, but since it's a variable we
                    // can't put it into m_defaultSettings, so we'll return
                    // Name if m_configurationSection hasn't been
                    // meaningfully set
                    if (m_configurationSection == null)
                    {
                        return Name;
                    }
                    else
                    {
                        return m_configurationSection;
                    }
                }
                set
                {
                    m_configurationSection = value;
                }
            }

            /// <summary>
            /// Gets or sets primary keys of input measurements the calculated measurement expects.
            /// </summary>
            public override SIEGate.Core.Legacy.MeasurementKey[] InputMeasurementKeys
            {
                get
                {
                    return base.InputMeasurementKeys;
                }
                set
                {
                    base.InputMeasurementKeys = value;

                    m_inputMeasurementKeyTypes = new SignalType[value.Length];

                    for (int i = 0; i < m_inputMeasurementKeyTypes.Length; i++)
                    {
                        m_inputMeasurementKeyTypes[i] = LookupSignalType(value[i]);
                    }
                }
            }

            /// <summary>
            /// Gets or sets output measurements that the calculated measurement will produce, if any.
            /// </summary>
            public override ILegacyMeasurement[] OutputMeasurements
            {
                get
                {
                    return base.OutputMeasurements;
                }
                set
                {
                    base.OutputMeasurements = value;

                    m_outputMeasurementTypes = new SignalType[value.Length];

                    for (int i = 0; i < m_outputMeasurementTypes.Length; i++)
                    {
                        m_outputMeasurementTypes[i] = LookupSignalType(value[i].Key);
                    }
                }
            }

            /// <summary>
            /// Gets or sets input measurement <see cref="SignalType"/>'s for each of the <see cref="ActionAdapterBase.InputMeasurementKeys"/>, if any.
            /// </summary>
            public virtual SignalType[] InputMeasurementKeyTypes
            {
                get
                {
                    return m_inputMeasurementKeyTypes;
                }
            }

            /// <summary>
            /// Gets or sets output measurement <see cref="SignalType"/>'s for each of the <see cref="ActionAdapterBase.OutputMeasurements"/>, if any.
            /// </summary>
            public virtual SignalType[] OutputMeasurementTypes
            {
                get
                {
                    return m_outputMeasurementTypes;
                }
            }

            public AdapterInfo()
                : base()
            {
                SetDefaults(m_defaultSettings);
            }

            /// <summary>
            /// Set the DataSource for this AdapterInfo class and derive member variables and properties relevant to this instance from that source.
            /// </summary>
            /// <param name="dataSource"></param>
            public override void ExtractConfiguration(DataSet dataSource)
            {
                base.ExtractConfiguration(dataSource);
            }

            // Lookup signal type for given measurement key
            private SignalType LookupSignalType(MeasurementKey key)
            {
                try
                {
                    DataRow[] filteredRows = DataSource.Tables["ActiveMeasurements"].Select(string.Format("ID = '{0}'", key.ToString()));

                    if (filteredRows.Length > 0)
                        return (SignalType)Enum.Parse(typeof(SignalType), filteredRows[0]["SignalType"].ToString(), true);
                }
                catch (Exception ex)
                {
                    LogDispatcher.E(StrLogSource, "ProcessException", "Failed to lookup signal type for measurement {0}: {1}{2}{3}", key.ToString(), ex.Message, Environment.NewLine, ex.StackTrace);
                }

                return SignalType.NONE;
            }


            /// <summary>
            /// Set the member variables and properties for this AdapterInfo from a parsed configuration string
            /// </summary>
            /// <param name="configInfo">A Dictionary (string, string) that results from parsing a configuration string.</param>
            public override void Update(Dictionary<string, string> configInfo)
            {
                base.Update(configInfo);

                SupportsTemporalProcessing = DefaultGet("supportsTemporalProcessing").ParseBoolean();

                ConfigurationSection = DefaultGet("configurationSection");
            }
        }

        #endregion


        #region [ Members ]
        private const string StrLogSource = "CalculatedMeasurementBase";
        public const string StrStatLastPublicationTime = "Last Publication Time";

        protected Cluster m_statsCluster;
        #endregion

        #region [ Constructors ]

        /// <summary>
        /// Create a CalculatedMeasurementBase instance
        /// </summary>
        /// <param name="info"></param>
        public CalculatedMeasurementBase(IAdapterInfo info)
            : base(info)
        {
            m_statsCluster = Stats.StatDispatcher.RegisterCluster(info.ID);
            m_statsCluster.RegisterDuration(Info.Name, StrStatLastPublicationTime);
        }

        #endregion

        #region [ Properties ]

        /// <summary>
        /// Gets or sets primary keys of input measurements the calculated measurement expects.
        /// </summary>
        public override SIEGate.Core.Legacy.MeasurementKey[] InputMeasurementKeys
        {
            get
            {
                return ((AdapterInfo) Info).InputMeasurementKeys;
            }
            set
            {
                ((AdapterInfo) Info).InputMeasurementKeys = value;
            }
        }

        /// <summary>
        /// Gets or sets output measurements that the calculated measurement will produce, if any.
        /// </summary>
        public override ILegacyMeasurement[] OutputMeasurements
        {
            get
            {
                return ((AdapterInfo) Info).OutputMeasurements;
            }
            set
            {
                ((AdapterInfo) Info).OutputMeasurements = value;
            }
        }

        /// <summary>
        /// Gets or sets input measurement <see cref="SignalType"/>'s for each of the <see cref="ActionAdapterBase.InputMeasurementKeys"/>, if any.
        /// </summary>
        public virtual SignalType[] InputMeasurementKeyTypes
        {
            get
            {
                return ((AdapterInfo) Info).InputMeasurementKeyTypes;
            }
        }

        /// <summary>
        /// Gets or sets output measurement <see cref="SignalType"/>'s for each of the <see cref="ActionAdapterBase.OutputMeasurements"/>, if any.
        /// </summary>
        public virtual SignalType[] OutputMeasurementTypes
        {
            get
            {
                return ((AdapterInfo) Info).OutputMeasurementTypes;
            }
        }

        /// <summary>
        /// Gets or sets the configuration section to use for this <see cref="CalculatedMeasurementBase"/>.
        /// </summary>
        public virtual string ConfigurationSection
        {
            get
            {
                return ((AdapterInfo) Info).ConfigurationSection;
            }
            set
            {
                ((AdapterInfo)Info).ConfigurationSection = value;
            }
        }

        /// <summary>
        /// Gets the flag indicating if this adapter supports temporal processing.
        /// </summary>
        public override bool SupportsTemporalProcessing
        {
            get
            {
                return ((AdapterInfo) Info).SupportsTemporalProcessing;
            }
        }

        /// <summary>
        /// Returns the detailed status of the calculated measurement.
        /// </summary>
        /// <remarks>
        /// Derived classes should extend status with implementation specific information.
        /// </remarks>
        public override string Status
        {
            get
            {
                StringBuilder status = new StringBuilder();
                int count;

                status.AppendFormat("     Configuration section: {0}", ConfigurationSection);
                status.AppendLine();
                status.Append(base.Status);

                if (OutputMeasurements != null && OutputMeasurements.Length > 0)
                {
                    status.AppendLine();
                    status.AppendLine("Output measurements signal type summary:");
                    status.AppendLine();

                    foreach (SignalType signalType in Enum.GetValues(typeof(SignalType)))
                    {
                        count = OutputMeasurements.Where((key, index) => OutputMeasurementTypes[index] == signalType).Count();

                        if (count > 0)
                        {
                            status.AppendFormat("{0} {1} signal{2}", count.ToString().PadLeft(15), signalType.GetFormattedSignalTypeName(), count > 1 ? "s" : "");
                            status.AppendLine();
                        }
                    }
                }

                if (InputMeasurementKeys != null && InputMeasurementKeys.Length > 0)
                {
                    status.AppendLine();
                    status.AppendLine("Input measurement keys signal type summary:");
                    status.AppendLine();

                    foreach (SignalType signalType in Enum.GetValues(typeof(SignalType)))
                    {
                        count = InputMeasurementKeys.Where((key, index) => InputMeasurementKeyTypes[index] == signalType).Count();

                        if (count > 0)
                        {
                            status.AppendFormat("{0} {1} signal{2}", count.ToString().PadLeft(15), signalType.GetFormattedSignalTypeName(), count > 1 ? "s" : "");
                            status.AppendLine();
                        }
                    }
                }

                return status.ToString();
            }
        }

        #endregion

        #region [ Methods ]

        /// <summary>
        /// Intializes <see cref="CalculatedMeasurementBase"/>.
        /// </summary>
        public override void Initialize()
        {
            base.Initialize();
        }

        #endregion
    }
}