﻿//******************************************************************************************************
//  WavInputAdapter.cs - Gbtc
//
//  Copyright © 2010, Grid Protection Alliance.  All Rights Reserved.
//
//  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
//  the NOTICE file distributed with this work for additional information regarding copyright ownership.
//  The GPA licenses this file to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  06/29/2011 - Stephen C. Wills
//       Generated original version of source code.
//  07/01/2011 - Stephen C. Wills
//       Modified WaveData dependency to refer to GSF.Media and moved the library
//       into TimeSeriesFramework.
//  12/27/2012 - Prasanna Kumari
//       Applied changes for Migrating adapters from the Time Series Framework to SIEGate Core
//
//******************************************************************************************************

using System.Data;
using GSF;
using GSF.Media;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Threading;
using SIEGate.Core;
using SIEGate.Core.Adapters;
using SIEGate.Core.Legacy;
using SIEGate.Core.Configurator;
using SIEGate.Core.Measurements;
using SIEGate.Core.Measurements.Values;

namespace WavInputAdapter
{
    /// <summary>
    /// Represents an input adapter that reads measurements from a WAV file.
    /// </summary>
    [Description("WAV: Reads measurements from a WAV file.")]
    [EditorBrowsable(EditorBrowsableState.Advanced)] // Normally defined as an input device protocol
    public class WavInputAdapter : InputAdapter, IDisposable
    {
        #region [ Members ]

        // Nested Types

        public class AdapterInfo : AdapterInfoBase
        {
            public override void ExtractConfiguration(DataSet configurationTables)
            {
                Dictionary<string, string> settings = Settings;
                string setting;

                if (settings.TryGetValue("outputMeasurements", out setting))
                    settings["outgoingStreamIDs"] = setting;

                base.ExtractConfiguration(configurationTables);

                if (settings.TryGetValue("outputSourceIDs", out setting))
                {
                    foreach (string sourceID in setting.Split(','))
                    {
                        DataRow[] rows = configurationTables.Tables["ActiveMeasurements"].Select(string.Format("ID LIKE '{0}:*'", sourceID));
                        Guid signalID;

                        foreach (DataRow row in rows)
                        {
                            if (Guid.TryParse(row["SignalID"].ToNonNullString(), out signalID))
                                OutgoingStreamIDs.Add(signalID);
                        }
                    }
                }
            }
        }

        // Constants
        private const double DefaultRecoveryDelay = 5.0D;
        private const long GapThreshold = Ticks.PerSecond;

        // Fields
        private WaveDataReader m_data;
        private int m_dataIndex;
        private int m_channels;
        private int m_sampleRate;
        private int m_numSamples;
        private TimeSpan m_audioLength;
        private long m_startTime;
        private bool m_enabled;
        private bool m_disposed;

        private const string StrLogSource = "WavInputAdapter";
        #endregion

        #region [ Constructors ]

        public WavInputAdapter(AdapterInfo info)
            : base(info)
        {
            Initialize();
            Start();
        }

        /// <summary>
        /// Releases the unmanaged resources before the <see cref="WavInputAdapter"/> object is reclaimed by <see cref="GC"/>.
        /// </summary>
        ~WavInputAdapter()
        {
            Dispose(false);
        }

        #endregion

        #region [ Properties ]

        public new AdapterInfo Info
        {
            get
            {
                return (AdapterInfo)base.Info;
            }
            set
            {
                base.Info = value;
            }
        }

        /// <summary>
        /// Gets or sets the name of the file from which to read measurements.
        /// </summary>
        [ConnectionStringParameter,
        Description("The name of the file from which to read measurements.")]
        public string WavFileName
        {
            get;
            set;
        }

        /// <summary>
        /// Gets or sets the amount of time, in seconds, needed to recover from a back log.
        /// </summary>
        [ConnectionStringParameter,
        Description("The amount of time, in seconds, needed to recover from a back log."),
        DefaultValue(DefaultRecoveryDelay)]
        public double RecoveryDelay
        {
            get;
            set;
        }

        #endregion

        #region [ Methods ]

        /// <summary>
        /// Initializes the connection string parameters of this <see cref="WavInputAdapter"/>.
        /// </summary>
        public void Initialize()
        {
            Dictionary<string, string> settings = Info.Settings;
            string setting;

            if (settings.TryGetValue("recoveryDelay", out setting))
                RecoveryDelay = double.Parse(setting);
            else
                RecoveryDelay = DefaultRecoveryDelay;

            if (!settings.TryGetValue("wavFileName", out setting))
                throw new ArgumentException("wavFileName is missing from settings - Example: wavFileName=Bohemian Rhapsody.wav");

            WavFileName = setting;
        }

        public void Start()
        {
            m_enabled = true;
            AttemptConnection();
        }

        public void Stop()
        {
            m_enabled = false;
            AttemptDisconnection();
        }

        /// <summary>
        /// Attempts to open the file to start getting wave data.
        /// </summary>
        protected void AttemptConnection()
        {
            WaveFile fileInfo = WaveFile.Load(WavFileName, false);

            m_channels = fileInfo.Channels;
            m_sampleRate = fileInfo.SampleRate;
            m_numSamples = fileInfo.DataChunk.ChunkSize / fileInfo.BlockAlignment;
            m_audioLength = fileInfo.AudioLength;

            m_data = WaveDataReader.FromFile(WavFileName);
            m_dataIndex = 0;

            //if (file.Channels != OutputMeasurements.Length)
            //    throw new ArgumentException(string.Format("The number of channels in the WAV file must match the number of output measurements. Channels: {0}, Measurements: {1}", file.Channels, OutputMeasurements.Length));

            m_startTime = DateTime.UtcNow.Ticks;

            Thread t = new Thread(ProcessMeasurements);
            t.IsBackground = true;
            t.Start();
        }

        /// <summary>
        /// Attempts to close the file and release resources held by the adapter.
        /// </summary>
        protected void AttemptDisconnection()
        {
            if (m_data != null)
            {
                m_data.Close();
                m_data.Dispose();
            }
            m_data = null;
        }

        /// <summary>
        /// Releases all the resources used by the <see cref="WavInputAdapter"/> object.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Releases the unmanaged resources used by the <see cref="WavInputAdapter"/> object and optionally releases the managed resources.
        /// </summary>
        /// <param name="disposing">true to release both managed and unmanaged resources; false to release only unmanaged resources.</param>
        protected virtual void Dispose(bool disposing)
        {
            if (!m_disposed)
            {
                try
                {
                    Stop();

                    if (disposing)
                    {
                        if (m_data != null)
                        {
                            m_data.Close();
                            m_data.Dispose();
                        }
                        m_data = null;
                    }
                }
                finally
                {
                    m_disposed = true;  // Prevent duplicate dispose.
                }
            }
        }

        /// <summary>
        /// Gets a short one-line status of this <see cref="LegacyAdapterBase"/>.
        /// </summary>
        /// <param name="maxLength">Maximum number of available characters for display.</param>
        /// <returns>A short one-line summary of the current status of this <see cref="LegacyAdapterBase"/>.</returns>
        public string GetShortStatus(int maxLength)
        {
            TimeSpan time = Ticks.FromSeconds(m_dataIndex / (double)m_sampleRate);
            return string.Format("Streaming {0} at time {1} / {2} - {3:0.00%}.", Path.GetFileName(WavFileName), time.ToString(@"m\:ss"), m_audioLength.ToString(@"m\:ss"), time.TotalSeconds / m_audioLength.TotalSeconds);
        }

        // Generates new measurements since the last time this was called.
        private void ProcessMeasurements()
        {
            // Declare the variables use in this method.
            List<IMeasurement> measurements = new List<IMeasurement>((int)(Ticks.ToSeconds(GapThreshold) * m_sampleRate * m_channels * 1.1D));
            Guid[] streamIDs = Info.OutgoingStreamIDs.ToArray();
            LittleBinaryValue[] sample;

            while (m_enabled)
            {
                try
                {
                    SpinWait spinner = new SpinWait();

                    // Determine what time it is now.
                    long now = DateTime.UtcNow.Ticks;

                    // Assign a timestamp to the next sample based on its location
                    // in the file relative to the other samples in the file.
                    long timestamp = m_startTime + (m_dataIndex * Ticks.PerSecond / m_sampleRate);

                    if (now - timestamp > GapThreshold)
                    {
                        m_startTime = now - (m_dataIndex * Ticks.PerSecond / m_sampleRate) + Ticks.FromSeconds(RecoveryDelay);
                        timestamp = now;
                        LogDispatcher.I(StrLogSource, "StatusMessage", "Start time reset.");
                    }

                    // Keep generating measurements until
                    // we catch up to the current time.
                    while (timestamp < now)
                    {
                        sample = m_data.GetNextSample();

                        // If the sample is null, we've reached the end of the file.
                        // Close and reopen it, resetting the data index and start time.
                        if (sample == null)
                        {
                            m_data.Close();
                            m_data.Dispose();

                            m_data = WaveDataReader.FromFile(WavFileName);
                            m_dataIndex = 0;

                            m_startTime = timestamp;
                            sample = m_data.GetNextSample();
                        }

                        // Create new measurements, one for each channel,
                        // and add them to the measurements list.
                        for (int i = 0; i < m_channels; i++)
                        {
                            measurements.Add(new Measurement(streamIDs[i], timestamp, MeasurementStateFlags.Normal, new DoubleValue(sample[i].ConvertToType(TypeCode.Double))));
                        }

                        // Update the data index and recalculate
                        // the assigned timestamp for the next sample.
                        m_dataIndex++;
                        timestamp = m_startTime + (m_dataIndex * Ticks.PerSecond / m_sampleRate);
                    }

                    OnEmit(measurements.ToArray());
                    measurements.Clear();

                    while (DateTime.UtcNow.Ticks - timestamp <= GapThreshold / 100)
                    {
                        // Ahead of schedule -- pause for a moment
                        spinner.SpinOnce();
                    }
                }
                catch (Exception ex)
                {
                    LogDispatcher.E(StrLogSource, "ProcessException", "{0}{1}{2}", ex.Message, Environment.NewLine, ex.StackTrace);
                }
            }
        }

        protected override void HandleMeasurement(IMeasurement m)
        {
        }

        #endregion
    }
}
