﻿//******************************************************************************************************
//  DataPublisher.cs - Gbtc
//
//  Copyright © 2010, Grid Protection Alliance.  All Rights Reserved.
//
//  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
//  the NOTICE file distributed with this work for additional information regarding copyright ownership.
//  The GPA licenses this file to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  2/2013 - University of Illinois, Information Trust Institute, Coordinated Science Laboratory
//       Updated code to work for the SIEGate system and architecture
//
//  08/20/2010 - J. Ritchie Carroll
//       Generated original version of source code.
//  11/15/2010 - Mehulbhai P Thakkar
//       Fixed bug when DataSubscriber tries to resubscribe by setting subscriber. Initialized manually 
//       in ReceiveClientDataComplete event handler.
//  12/02/2010 - J. Ritchie Carroll
//       Fixed an issue for when DataSubcriber dynamically resubscribes with a different
//       synchronization method (e.g., going from unsynchronized to synchronized)
//  05/26/2011 - J. Ritchie Carroll
//       Implemented subscriber authentication model.
//  02/07/2012 - Mehulbhai Thakkar
//       Modified m_metadataTables to include filter expression and made the list ";" seperated.
//
//******************************************************************************************************

using System.Collections.ObjectModel;
using System.Security;
using GSF.Communication;
using GSF.Configuration;
using GSF.Data;
using GSF.IO;
using GSF.Net.Security;
using GSF.Security.Cryptography;
using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Globalization;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Net;
using System.Net.Security;
using System.Net.Sockets;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Text.RegularExpressions;
using System.Xml;
using SIEGate.Core;
using SIEGate.Core.Configurator;
using SIEGate.Core.Legacy;

namespace GSF.TimeSeries.Transport
{
    #region [ Enumerations ]

    /// <summary>
    /// Server commands received by <see cref="DataPublisher"/> and sent by <see cref="DataSubscriber"/>.
    /// </summary>
    /// <remarks>
    /// Solicited server commands will receive a <see cref="ServerResponse.Succeeded"/> or <see cref="ServerResponse.Failed"/>
    /// response code along with an associated success or failure message. Message type for successful responses will be based
    /// on server command - for example, server response for a successful MetaDataRefresh command will return a serialized
    /// <see cref="DataSet"/> of the available server metadata. Message type for failed responses will always be a string of
    /// text respresenting the error message.
    /// </remarks>
    public enum ServerCommand : byte
    {
        /// <summary>
        /// Authenticate command.
        /// </summary>
        /// <remarks>
        /// Requests that server authenticate client with the following encrypted authentication packet.
        /// Successful return message type will be string indicating client ID was validated.
        /// Client should not attempt further steps if authentication fails, user will need to take action.
        /// Client should send MetaDataRefresh and Subscribe commands upon successful authentication.
        /// If server is setup to not require authentication, authentication step can be skipped - however
        /// it will be up to the implementation to know if authentication is required - the server cannot be
        /// queried to determine if authentication is required.
        /// It is expected that use of this protocol between gateways over a shared network (i.e., the lower
        /// security zone) will require authentication.
        /// It is expected that use of this protocol between gateway and other systems (e.g., a PDC) in a
        /// private network (i.e., the higher security zone) can be setup to not require authentication.
        /// </remarks>
        Authenticate = 0x00,
        /// <summary>
        /// Meta data refresh command.
        /// </summary>
        /// <remarks>
        /// Requests that server send an updated set of metadata so client can refresh its point list.
        /// Successful return message type will be a <see cref="DataSet"/> containing server device and measurement metadata.
        /// Received device list should be defined as children of the "parent" server device connection similar to the way
        /// PMUs are defined as children of a parent PDC device connection.
        /// Devices and measurements contain unique Guids that should be used to key metadata updates in local repository.
        /// </remarks>
        MetaDataRefresh = 0x01,
        /// <summary>
        /// Subscribe command.
        /// </summary>
        /// <remarks>
        /// Requests a subscription of streaming data from server based on connection string that follows.
        /// It will not be necessary to stop an existing subscription before requesting a new one.
        /// Successful return message type will be string indicating total number of allowed points.
        /// Client should wait for UpdateSignalIndexCache and UpdateBaseTime response codes before attempting
        /// to parse data when using the compact measurement format.
        /// </remarks>
        Subscribe = 0x02,
        /// <summary>
        /// Unsubscribe command.
        /// </summary>
        /// <remarks>
        /// Requests that server stop sending streaming data to the client and cancel the current subscription.
        /// </remarks>
        Unsubscribe = 0x03,
        /// <summary>
        /// Rotate cipher keys.
        /// </summary>
        /// <remarks>
        /// Manually requests that server send a new set of cipher keys for data packet encryption.
        /// </remarks>
        RotateCipherKeys = 0x04,
        /// <summary>
        /// Update processing interval.
        /// </summary>
        /// <remarks>
        /// Manually requests server to update the processing interval with the following specified value.
        /// </remarks>
        UpdateProcessingInterval = 0x05,
        /// <summary>
        /// Define operational modes for subscriber connection.
        /// </summary>
        /// <remarks>
        /// As soon as connection is established, requests that server set operational
        /// modes that affect how the subscriber and publisher will communicate.
        /// </remarks>
        DefineOperationalModes = 0x06,
        /// <summary>
        /// Confirm receipt of a notification.
        /// </summary>
        /// <remarks>
        /// This message is sent in response to <see cref="ServerResponse.Notify"/>.
        /// </remarks>
        ConfirmNotification = 0x07,
        /// <summary>
        /// Confirm receipt of a buffer block measurement.
        /// </summary>
        /// <remarks>
        /// This message is sent in response to <see cref="ServerResponse.BufferBlock"/>.
        /// </remarks>
        ConfirmBufferBlock = 0x08
    }

    /// <summary>
    /// Server responses sent by <see cref="DataPublisher"/> and received by <see cref="DataSubscriber"/>.
    /// </summary>
    public enum ServerResponse : byte
    {
        // Although the server commands and responses will be on two different paths, the response enumeration values
        // are defined as distinct from the command values to make it easier to identify codes from a wire analysis.

        /// <summary>
        /// Command succeeded response.
        /// </summary>
        /// <remarks>
        /// Informs client that its solicited server command succeeded, original command and success message follow.
        /// </remarks>
        Succeeded = 0x80,
        /// <summary>
        /// Command failed response.
        /// </summary>
        /// <remarks>
        /// Informs client that its solicited server command failed, original command and failure message follow.
        /// </remarks>
        Failed = 0x81,
        /// <summary>
        /// Data packet response.
        /// </summary>
        /// <remarks>
        /// Unsolicited response informs client that a data packet follows.
        /// </remarks>
        DataPacket = 0x82,
        /// <summary>
        /// Update signal index cache response.
        /// </summary>
        /// <remarks>
        /// Unsolicited response requests that client update its runtime signal index cache with the one that follows.
        /// </remarks>
        UpdateSignalIndexCache = 0x83,
        /// <summary>
        /// Update runtime base-timestamp offsets response.
        /// </summary>
        /// <remarks>
        /// Unsolicited response requests that client update its runtime base-timestamp offsets with those that follow.
        /// </remarks>
        UpdateBaseTimes = 0x84,
        /// <summary>
        /// Update runtime cipher keys response.
        /// </summary>
        /// <remarks>
        /// Response, solicited or unsolicited, requests that client update its runtime data cipher keys with those that follow.
        /// </remarks>
        UpdateCipherKeys = 0x85,
        /// <summary>
        /// Data start time response packet.
        /// </summary>
        /// <remarks>
        /// Unsolicited response provides the start time of data being processed from the first measurement.
        /// </remarks>
        DataStartTime = 0x86,
        /// <summary>
        /// Processing complete notification.
        /// </summary>
        /// <remarks>
        /// Unsolicited response provides notification that input processing has completed, typically via temporal constraint.
        /// </remarks>
        ProcessingComplete = 0x87,
        /// <summary>
        /// Buffer block response.
        /// </summary>
        /// <remarks>
        /// Unsolicited response informs client that a raw buffer block follows.
        /// </remarks>
        BufferBlock = 0x88,
        /// <summary>
        /// Notify response.
        /// </summary>
        /// <remarks>
        /// Unsolicited response provides a notification message to the client.
        /// </remarks>
        Notify = 0x89,
        /// <summary>
        /// No operation keep-alive ping.
        /// </summary>
        /// <remarks>
        /// The command channel can remain quiet for some time, this command allows a period test of client connectivity.
        /// </remarks>
        NoOP = 0xFF
    }

    /// <summary>
    /// <see cref="DataPublisher"/> data packet flags.
    /// </summary>
    [Flags()]
    public enum DataPacketFlags : byte
    {
        /// <summary>
        /// Determines if data packet is synchronized.
        /// </summary>
        /// <remarks>
        /// Bit set = synchronized, bit clear = unsynchronized.
        /// </remarks>
        Synchronized = (byte)Bits.Bit00,
        /// <summary>
        /// Determines if serialized measurement is compact.
        /// </summary>
        /// <remarks>
        /// Bit set = compact, bit clear = full fidelity.
        /// </remarks>
        Compact = (byte)Bits.Bit01,
        /// <summary>
        /// Detemines which cipher index to use when encrypting data packet.
        /// </summary>
        /// <remarks>
        /// Bit set = use odd cipher index (i.e., 1), bit clear = use even cipher index (i.e., 0).
        /// </remarks>
        CipherIndex = (byte)Bits.Bit02,
        /// <summary>
        /// No flags set.
        /// </summary>
        /// <remarks>
        /// This would represent unsynchronized, full fidelity measurement data packets.
        /// </remarks>
        NoFlags = (Byte)Bits.Nil
    }

    /// <summary>
    /// Operational modes that affect how <see cref="DataPublisher"/> and <see cref="DataSubscriber"/> communicate.
    /// </summary>
    /// <remarks>
    /// Operational modes only apply to fundamental protocol control.
    /// </remarks>
    [Flags()]
    public enum OperationalModes : uint
    {
        /// <summary>
        /// Mask to get version number of protocol.
        /// </summary>
        /// <remarks>
        /// Version number is currently set to 0.
        /// </remarks>
        VersionMask = (uint)(Bits.Bit04 | Bits.Bit03 | Bits.Bit02 | Bits.Bit01 | Bits.Bit00),
        /// <summary>
        /// Mask to get mode of compression.
        /// </summary>
        /// <remarks>
        /// Compression mode is currently not supported.
        /// These bits are simply reserved for different compression modes.
        /// </remarks>
        CompressionModeMask = (uint)(Bits.Bit07 | Bits.Bit06 | Bits.Bit05),
        /// <summary>
        /// Mask to get character encoding used when exchanging messages between publisher and subscriber.
        /// </summary>
        /// <remarks>
        /// 00 = UTF-16 little endian<br/>
        /// 01 = UTF-16 big endian<br/>
        /// 10 = UTF-8<br/>
        /// 11 = ANSI
        /// </remarks>
        EncodingMask = (uint)(Bits.Bit09 | Bits.Bit08),
        /// <summary>
        /// Determines type of serialization to use when exchanging signal index cache and metadata.
        /// </summary>
        /// <remarks>
        /// Bit set = common serialization format, bit clear = .NET serialization format
        /// </remarks>
        UseCommonSerializationFormat = (uint)Bits.Bit24,
        /// <summary>
        /// Determines whether the signal index cache is compressed when exchanging between publisher and subscriber.
        /// </summary>
        /// <remarks>
        /// Bit set = compress, bit clear = no compression
        /// </remarks>
        CompressSignalIndexCache = (uint)Bits.Bit30,
        /// <summary>
        /// Determines whether is compressed when exchanging between publisher and subscriber.
        /// </summary>
        /// <remarks>
        /// Bit set = compress, bit clear = no compression
        /// </remarks>
        CompressMetadata = (uint)Bits.Bit31,
        /// <summary>
        /// No flags set.
        /// </summary>
        /// <remarks>
        /// This would represent protocol version 0,
        /// UTF-16 little endian character encoding,
        /// .NET serialization and no compression.
        /// </remarks>
        NoFlags = (uint)Bits.Nil
    }

    /// <summary>
    /// Enumeration for character encodings supported by the Gateway Exchange Protocol.
    /// </summary>
    public enum OperationalEncoding : uint
    {
        /// <summary>
        /// UTF-16 little endian
        /// </summary>
        Unicode = (uint)Bits.Nil,
        /// <summary>
        /// UTF-16 bit endian
        /// </summary>
        BigEndianUnicode = (uint)Bits.Bit08,
        /// <summary>
        /// UTF-8
        /// </summary>
        UTF8 = (uint)Bits.Bit09,
        /// <summary>
        /// ANSI
        /// </summary>
        ANSI = (uint)(Bits.Bit09 | Bits.Bit08)
    }

    /// <summary>
    /// Enumeration for compression modes supported by the Gateway Exchange Protocol.
    /// </summary>
    public enum GatewayCompressionMode : uint
    {
        /// <summary>
        /// GZip compression
        /// </summary>
        GZip = (uint)Bits.Bit05,
        /// <summary>
        /// No compression
        /// </summary>
        None = (uint)Bits.Nil
    }

    #endregion

    /// <summary>
    /// Represents a data publishing server that allows multiple connections for data subscriptions.
    /// </summary>
    [Description("DataPublisher: server component that allows gateway-style subscription connections.")]
    [EditorBrowsable(EditorBrowsableState.Always)]
    public class DataPublisher : NetworkCommunicator, IDisposable
    {
        #region [ Members ]

        // Nested Types
        public class NetworkCommunicatorInfo : NetworkCommunicatorInfoBase
        {
            private DataSet m_dataSource;

            public override void ExtractConfiguration(DataSet configurationTables)
            {
                base.ExtractConfiguration(configurationTables);
                m_dataSource = configurationTables;
            }

            internal DataSet DataSource
            {
                get
                {
                    return m_dataSource;
                }
            }
        }

        // Events

        /// <summary>
        /// Indicates that a new client has connected to the publisher.
        /// </summary>
        /// <remarks>
        /// <see cref="EventArgs{T1, T2, T3}.Argument1"/> is the <see cref="Guid"/> based subscriber ID.<br/>
        /// <see cref="EventArgs{T1, T2, T3}.Argument2"/> is the connection identification (e.g., IP and DNS name, if available).<br/>
        /// <see cref="EventArgs{T1, T2, T3}.Argument3"/> is the subscriber information as reported by the client.
        /// </remarks>
        public event EventHandler<EventArgs<Guid, string, string>> ClientConnected;

        // Constants

        /// <summary>
        /// Default value for <see cref="AllowSynchronizedSubscription"/>.
        /// </summary>
        public const bool DefaultAllowSynchronizedSubscription = false;

        /// <summary>
        /// Default value for <see cref="UseBaseTimeOffsets"/>.
        /// </summary>
        public const bool DefaultUseBaseTimeOffsets = false;

        /// <summary>
        /// Default value for <see cref="CipherKeyRotationPeriod"/>.
        /// </summary>
        public const double DefaultCipherKeyRotationPeriod = 60000.0D;

        // Maximum packet size before software fragmentation of UDP payload
        internal const int MaxPacketSize = ushort.MaxValue / 2;

        // Length of random salt prefix
        internal const int CipherSaltLength = 8;

        // Fields
        private TlsServer m_commandChannel;
        private CertificatePolicyChecker m_certificateChecker;
        private Dictionary<X509Certificate, DataRow> m_subscriberIdentities;
        private Dictionary<Guid, List<ClientConnection>> m_subscriberConnections;
        private ConcurrentDictionary<Guid, ClientConnection> m_clientConnections;
        private ConcurrentDictionary<MeasurementKey, Guid> m_signalIDCache;
        private Dictionary<Guid, Dictionary<int, string>> m_clientNotifications;
        private object m_clientNotificationsLock;
        private System.Timers.Timer m_commandChannelRestartTimer;
        private System.Timers.Timer m_cipherKeyRotationTimer;
        private INotificationReceiver m_signalProxy;
        private DataSet m_dataSource;
        private string m_metadataTables;
        private long m_processedMeasurements;
        private bool m_encryptPayload;
        private bool m_sharedDatabase;
        private bool m_useBaseTimeOffsets;
        private object m_subscriberConnectionsLock;
        private bool m_disposed;
        private volatile bool m_running = false;


        #endregion

        #region [ Constructors ]

        public DataPublisher(IMessageBusProxy busProxy, NetworkCommunicatorInfoBase info)
            : base(busProxy, info)
        {
            m_subscriberConnections = new Dictionary<Guid, List<ClientConnection>>();
            m_clientConnections = new ConcurrentDictionary<Guid, ClientConnection>();
            m_signalIDCache = new ConcurrentDictionary<MeasurementKey, Guid>();
            m_clientNotifications = new Dictionary<Guid, Dictionary<int, string>>();
            m_clientNotificationsLock = new object();

            m_metadataTables =
                "SELECT NodeID, UniqueID, OriginalSource, IsConcentrator, Acronym, Name, ParentAcronym, ProtocolName, FramesPerSecond, Enabled FROM DeviceDetail WHERE OriginalSource IS NULL AND IsConcentrator = 0;" +
                "SELECT Internal, DeviceAcronym, DeviceName, SignalAcronym, ID, SignalID, PointTag, SignalReference, Description, Enabled FROM MeasurementDetail WHERE Internal <> 0;" +
                "SELECT DeviceAcronym, Label, Type, Phase, SourceIndex FROM PhasorDetail";

            // Setup a timer for restarting the command channel if it fails
            m_commandChannelRestartTimer = new System.Timers.Timer(2000.0D);
            m_commandChannelRestartTimer.AutoReset = false;
            m_commandChannelRestartTimer.Enabled = false;
            m_commandChannelRestartTimer.Elapsed += m_commandChannelRestartTimer_Elapsed;

            // Setup a timer for rotating cipher keys
            m_cipherKeyRotationTimer = new System.Timers.Timer(DefaultCipherKeyRotationPeriod);
            m_cipherKeyRotationTimer.AutoReset = true;
            m_cipherKeyRotationTimer.Enabled = false;
            m_cipherKeyRotationTimer.Elapsed += m_cipherKeyRotationTimer_Elapsed;

            m_subscriberConnectionsLock = new object();

            DataSource = Info.DataSource;

            Initialize();
        }

        /// <summary>
        /// Releases the unmanaged resources before the <see cref="DataPublisher"/> object is reclaimed by <see cref="GC"/>.
        /// </summary>
        ~DataPublisher()
        {
            Dispose(false);
        }

        #endregion

        #region [ Properties ]

        public new NetworkCommunicatorInfo Info
        {
            get
            {
                return (NetworkCommunicatorInfo)base.Info;
            }
            set
            {
                base.Info = value;
            }
        }

        /// <summary>
        /// Gets or sets flag that determines whether data sent over the data channel should be encrypted.
        /// </summary>
        [ConnectionStringParameter,
        Description("Define the flag that determines whether data sent over the data channel should be encrypted. This value is only relevant when requireAuthentication is true.")]
        public bool EncryptPayload
        {
            get
            {
                return m_encryptPayload;
            }
            set
            {
                m_encryptPayload = value;

                // Start cipher key rotation timer when encrypting payload
                if (m_cipherKeyRotationTimer != null)
                    m_cipherKeyRotationTimer.Enabled = value;
            }
        }

        /// <summary>
        /// Gets or sets flag that indicates whether this publisher is publishing
        /// data that this node subscribed to from another node in a shared database.
        /// </summary>
        [ConnectionStringParameter,
        Description("Define the flag that indicates whether this publisher is publishing data that this node subscribed to from another node in a shared database.")]
        public bool SharedDatabase
        {
            get
            {
                return m_sharedDatabase;
            }
            set
            {
                m_sharedDatabase = value;
            }
        }

        /// <summary>
        /// Gets or sets the cipher key rotation period.
        /// </summary>
        [ConnectionStringParameter,
        Description("Define the period, in milliseconds, over which new cipher keys will be provided to subscribers when EncryptPayload is true.")]
        public double CipherKeyRotationPeriod
        {
            get
            {
                if (m_cipherKeyRotationTimer != null)
                    return m_cipherKeyRotationTimer.Interval;

                return double.NaN;
            }
            set
            {
                if (value < 1000.0D)
                {
                    value = 1000.0D;
                    LogDispatcher.Log(LogLevel.Warn, Name, Tag, "Cipher key rotation period should not be set to less than 1000 milliseconds.");
                    // throw new ArgumentOutOfRangeException("value", "Cipher key rotation period should not be set to less than 1000 milliseconds.");
                }
                if (m_cipherKeyRotationTimer != null)
                    m_cipherKeyRotationTimer.Interval = value;
                else
                    throw new ArgumentException("Cannot assign new cipher rotation period, timer is not defined.");
            }
        }

        /// <summary>
        /// Gets or sets <see cref="DataSet"/> based data source used to load each <see cref="ILegacyAdapter"/>.
        /// Updates to this property will cascade to all items in this <see cref="LegacyAdapterCollectionBase{T}"/>.
        /// </summary>
        public DataSet DataSource
        {
            get
            {
                return m_dataSource;
            }
            set
            {
                m_dataSource = value;

                UpdateRights();
                UpdateClientNotifications();
                UpdateCertificateChecker();
            }
        }

        /// <summary>
        /// Gets the status of this <see cref="DataPublisher"/>.
        /// </summary>
        /// <remarks>
        /// Derived classes should provide current status information about the adapter for display purposes.
        /// </remarks>
        public string Status
        {
            get
            {
                StringBuilder status = new StringBuilder();

                if (m_commandChannel != null)
                    status.Append(m_commandChannel.Status);

                return status.ToString();
            }
        }

        /// <summary>
        /// Gets or sets the name of this <see cref="DataPublisher"/>.
        /// </summary>
        /// <remarks>
        /// The assigned name is used as the settings category when persisting the TCP server settings.
        /// </remarks>
        public string Name
        {
            get
            {
                return Info.Name ?? "DATAPUBLISHER";
            }
        }

        /// <summary>
        /// Gets or sets the logging tag name <see cref="DataPublisher"/>
        /// </summary>
        /// <remarks>
        /// The assigned tag is used as the logging tagname for this module
        /// </remarks>
        public string Tag
        {
            get { return "general"; }
        }


        /// <summary>
        /// Gets or sets comma seperated list of tables to include in meta-data exchange.
        /// </summary>
        public string MetadataTables
        {
            get
            {
                return m_metadataTables;
            }
            set
            {
                m_metadataTables = value;
            }
        }

        /// <summary>
        /// Gets dictionary of connected clients.
        /// </summary>
        internal protected ConcurrentDictionary<Guid, ClientConnection> ClientConnections
        {
            get
            {
                return m_clientConnections;
            }
        }

        /// <summary>
        /// Gets or sets reference to <see cref="TcpServer"/> command channel, attaching and/or detaching to events as needed.
        /// </summary>
        protected TlsServer CommandChannel
        {
            get
            {
                return m_commandChannel;
            }
            set
            {
                if (m_commandChannel != null)
                {
                    // Detach from events on existing command channel reference
                    m_commandChannel.ClientConnected -= m_commandChannel_ClientConnected;
                    m_commandChannel.ClientDisconnected -= m_commandChannel_ClientDisconnected;
                    m_commandChannel.ClientConnectingException -= m_commandChannel_ClientConnectingException;
                    m_commandChannel.ReceiveClientDataComplete -= m_commandChannel_ReceiveClientDataComplete;
                    m_commandChannel.ReceiveClientDataException -= m_commandChannel_ReceiveClientDataException;
                    m_commandChannel.SendClientDataException -= m_commandChannel_SendClientDataException;
                    m_commandChannel.ServerStarted -= m_commandChannel_ServerStarted;
                    m_commandChannel.ServerStopped -= m_commandChannel_ServerStopped;

                    if (m_commandChannel != value)
                        m_commandChannel.Dispose();
                }

                // Assign new command channel reference
                m_commandChannel = value;

                if (m_commandChannel != null)
                {
                    // Attach to desired events on new command channel reference
                    m_commandChannel.ClientConnected += m_commandChannel_ClientConnected;
                    m_commandChannel.ClientDisconnected += m_commandChannel_ClientDisconnected;
                    m_commandChannel.ClientConnectingException += m_commandChannel_ClientConnectingException;
                    m_commandChannel.ReceiveClientDataComplete += m_commandChannel_ReceiveClientDataComplete;
                    m_commandChannel.ReceiveClientDataException += m_commandChannel_ReceiveClientDataException;
                    m_commandChannel.SendClientDataException += m_commandChannel_SendClientDataException;
                    m_commandChannel.ServerStarted += m_commandChannel_ServerStarted;
                    m_commandChannel.ServerStopped += m_commandChannel_ServerStopped;
                }
            }
        }

        /// <summary>
        /// Gets the number of clients currently connected to the publisher.
        /// </summary>
        public int Count
        {
            get
            {
                return m_commandChannel.ClientIDs.Length;
            }
        }

        public long ProcessedMeasurements
        {
            get
            {
                return m_processedMeasurements;
            }
        }

        /// <summary>
        /// Gets flag indicating if publisher is connected and listening.
        /// </summary>
        public bool IsConnected
        {
            get
            {
                return m_commandChannel.Enabled;
            }
        }

        #endregion

        #region [ Methods ]

        /// <summary>
        /// Releases all the resources used by the <see cref="DataPublisher"/> object.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        /// <summary>
        /// Releases the unmanaged resources used by the <see cref="DataPublisher"/> object and optionally releases the managed resources.
        /// </summary>
        /// <param name="disposing">true to release both managed and unmanaged resources; false to release only unmanaged resources.</param>
        protected void Dispose(bool disposing)
        {
            if (!m_disposed)
            {
                try
                {
                    if (disposing)
                    {
                        CommandChannel = null;

                        if (m_clientConnections != null)
                            m_clientConnections.Values.AsParallel().ForAll(cc => cc.Dispose());

                        m_clientConnections = null;

                        // Dispose command channel restart timer
                        if (m_commandChannelRestartTimer != null)
                        {
                            m_commandChannelRestartTimer.Elapsed -= m_commandChannelRestartTimer_Elapsed;
                            m_commandChannelRestartTimer.Dispose();
                        }
                        m_commandChannelRestartTimer = null;

                        // Dispose the cipher key rotation timer
                        if (m_cipherKeyRotationTimer != null)
                        {
                            m_cipherKeyRotationTimer.Elapsed -= m_cipherKeyRotationTimer_Elapsed;
                            m_cipherKeyRotationTimer.Dispose();
                        }
                        m_cipherKeyRotationTimer = null;
                    }
                }
                finally
                {
                    m_disposed = true;          // Prevent duplicate dispose.
                }
            }
        }

        /// <summary>
        /// Intializes <see cref="DataPublisher"/>.
        /// </summary>
        public void Initialize()
        {
            // Create a new TLS server
            TlsServer commandChannel = new TlsServer();

            // Create certificate checker
            m_certificateChecker = new CertificatePolicyChecker();
            m_subscriberIdentities = new Dictionary<X509Certificate, DataRow>();

            // Initialize default settings
            // SettingsCategory maps to the type of DataPublisher, e.g., internal vs external
            commandChannel.SettingsCategory = Name.Replace("!", "").ToLower();
            commandChannel.ConfigurationString = "port=6165";
            commandChannel.PayloadAware = true;
            commandChannel.PersistSettings = true;
            commandChannel.RequireClientCertificate = true;
            commandChannel.CertificateChecker = m_certificateChecker;

            // Assign command channel client reference and attach to needed events
            CommandChannel = commandChannel;

            // Initialize TLS server (loads config file settings)
            m_commandChannel.Initialize();

            // Load trusted clients and their configured policies
            UpdateCertificateChecker();
        }

        public override void Start()
        {
            // Start cipher key rotation timer when encrypting payload
            if (EncryptPayload && m_cipherKeyRotationTimer != null)
                m_cipherKeyRotationTimer.Start();
            // Start the command channel
            m_commandChannel.Start();
            m_running = true;
            while (m_running)
            {
                doit();
            }
        }

        public override void Stop()
        {
            m_running = false;
            m_cipherKeyRotationTimer.Stop();
            m_commandChannel.Stop();
        }

        /// <summary>
        /// Gets a short one-line status of this <see cref="DataPublisher"/>.
        /// </summary>
        /// <param name="maxLength">Maximum number of available characters for display.</param>
        /// <returns>A short one-line summary of the current status of the <see cref="DataPublisher"/>.</returns>
        public string GetShortStatus(int maxLength)
        {
            if ((object)m_commandChannel != null)
                return string.Format("Publishing data to {0} clients.", m_commandChannel.ClientIDs.Length).CenterText(maxLength);

            return "Currently not connected".CenterText(maxLength);
        }

        /// <summary>
        /// Enumerates connected clients.
        /// </summary>
        [AdapterCommand("Enumerates connected clients.")]
        public virtual void EnumerateClients()
        {
            StringBuilder clientEnumeration = new StringBuilder();
            Guid[] clientIDs = (Guid[])m_commandChannel.ClientIDs.Clone();
            ClientConnection connection;
            string timestampFormat;

            clientEnumeration.AppendFormat("\r\nIndices for {0} connected clients:\r\n\r\n", clientIDs.Length);

            for (int i = 0; i < clientIDs.Length; i++)
            {
                if (m_clientConnections.TryGetValue(clientIDs[i], out connection) && (object)connection != null)
                {
                    timestampFormat = string.Format("{0} Format, {1}-byte timestamps", "Compact", GetTimestampSize(connection));
                    clientEnumeration.AppendFormat("  {0} - {1}\r\n          {2}\r\n          {3}\r\n          {4}\r\n\r\n", i.ToString().PadLeft(3), connection.ConnectionID, connection.SubscriberInfo, timestampFormat, connection.OperationalModes);
                }
            }

            // Display enumeration
            LogDispatcher.I(StrLogSource, "StatusMessage", clientEnumeration.ToString());
        }

        /// <summary>
        /// Rotates cipher keys for specified client connection.
        /// </summary>
        /// <param name="clientIndex">Enumerated index for client connection.</param>
        [AdapterCommand("Rotates cipher keys for client connection using its enumerated index.")]
        public virtual void RotateCipherKeys(int clientIndex)
        {
            Guid clientID = Guid.Empty;
            bool success = true;

            try
            {
                clientID = m_commandChannel.ClientIDs[clientIndex];
            }
            catch
            {
                success = false;
                LogDispatcher.E(StrLogSource, "StatusMessage", "ERROR: Failed to find connected client with enumerated index " + clientIndex);
            }

            if (success)
            {
                ClientConnection connection;

                if (m_clientConnections.TryGetValue(clientID, out connection))
                    connection.RotateCipherKeys();
                else
                    LogDispatcher.E(StrLogSource, "StatusMessage", "ERROR: Failed to find connected client " + clientID);
            }
        }

        /// <summary>
        /// Gets subscriber information for specified client connection.
        /// </summary>
        /// <param name="clientIndex">Enumerated index for client connection.</param>
        [AdapterCommand("Gets subscriber information for client connection using its enumerated index.")]
        public virtual string GetSubscriberInfo(int clientIndex)
        {
            Guid clientID = Guid.Empty;
            bool success = true;

            try
            {
                clientID = m_commandChannel.ClientIDs[clientIndex];
            }
            catch
            {
                success = false;
                LogDispatcher.E(StrLogSource, "StatusMessage", "ERROR: Failed to find connected client with enumerated index " + clientIndex);
            }

            if (success)
            {
                ClientConnection connection;

                if (m_clientConnections.TryGetValue(clientID, out connection))
                    return connection.SubscriberInfo;

                LogDispatcher.E(StrLogSource, "StatusMessage", "ERROR: Failed to find connected client " + clientID);
            }

            return "";
        }

        /// <summary>
        /// Gets subscriber status for specified subscriber ID.
        /// </summary>
        /// <param name="subscriberID">Guid based subscriber ID for client connection.</param>
        [AdapterCommand("Gets subscriber status for client connection using its subscriber ID.")]
        public virtual Tuple<Guid, bool, string> GetSubscriberStatus(Guid subscriberID)
        {
            return new Tuple<Guid, bool, string>(subscriberID, GetConnectionProperty(subscriberID, cc => cc.IsConnected), GetConnectionProperty(subscriberID, cc => cc.SubscriberInfo));
        }

        /// <summary>
        /// Sends a notification to all subscribers.
        /// </summary>
        /// <param name="message">The message to be sent.</param>
        [AdapterCommand("Sends a notification to all subscribers.")]
        public virtual void SendNotification(string message)
        {
            string notification = string.Format("[{0}] {1}", DateTime.UtcNow.ToString("yyyy-MM-dd HH:mm:ss.fff"), message);

            lock (m_clientNotificationsLock)
            {
                foreach (Dictionary<int, string> notifications in m_clientNotifications.Values)
                    notifications.Add(notification.GetHashCode(), notification);

                SerializeClientNotifications();
                SendAllNotifications();
            }

            LogDispatcher.I(StrLogSource, "StatusMessage", "Sent notification: {0}", message);
        }

        /// <summary>
        /// Updates signal index cache based on input measurement keys.
        /// </summary>
        /// <param name="clientID">Client ID of connection over which to update signal index cache.</param>
        /// <param name="signalIndexCache">New signal index cache.</param>
        /// <param name="inputMeasurementKeys">Subscribed measurement keys.</param>
        public void UpdateSignalIndexCache(Guid clientID, SignalIndexCache signalIndexCache, MeasurementKey[] inputMeasurementKeys)
        {
            ConcurrentDictionary<ushort, Tuple<Guid, string, uint>> reference = new ConcurrentDictionary<ushort, Tuple<Guid, string, uint>>();
            List<Guid> unauthorizedKeys = new List<Guid>();
            ushort index = 0;
            Guid signalID;

            byte[] serializedSignalIndexCache;
            ClientConnection connection;

            if (inputMeasurementKeys != null)
            {
                // We will now go through the client's requested keys and see which ones are authorized for subscription,
                // this information will be available through the returned signal index cache which will also define
                // a runtime index optimization for the allowed measurements.
                foreach (MeasurementKey key in inputMeasurementKeys)
                {
                    // Validate that subscriber has rights to this signal
                    if (SubscriberHasRights(signalIndexCache.SubscriberID, key, out signalID))
                        reference.TryAdd(index++, new Tuple<Guid, string, uint>(signalID, key.Source, key.ID));
                    else
                        unauthorizedKeys.Add(key.SignalID);
                }
            }

            signalIndexCache.Reference = reference;
            signalIndexCache.UnauthorizedSignalIDs = unauthorizedKeys.ToArray();
            serializedSignalIndexCache = SerializeSignalIndexCache(clientID, signalIndexCache);

            // Send client updated signal index cache
            if (m_clientConnections.TryGetValue(clientID, out connection) && connection.IsSubscribed)
                SendClientResponse(clientID, ServerResponse.UpdateSignalIndexCache, ServerCommand.Subscribe, serializedSignalIndexCache);
        }

        /// <summary>
        /// Updates each subscription's inputs based on
        /// possible updates to that subscriber's rights.
        /// </summary>
        protected void UpdateRights()
        {
            foreach (ClientConnection connection in m_clientConnections.Values)
                UpdateRights(connection);
        }

        /// <summary>
        /// Determines if subscriber has rights to specified <paramref name="signalID"/>.
        /// </summary>
        /// <param name="subscriberID"><see cref="Guid"/> based subscriber ID.</param>
        /// <param name="signalID"><see cref="Guid"/> signal ID to lookup.</param>
        /// <returns><c>true</c> if subscriber has rights to specified <see cref="MeasurementKey"/>; otherwise <c>false</c>.</returns>
        protected bool SubscriberHasRights(Guid subscriberID, Guid signalID)
        {
            try
            {
                bool allowed = false;

                // Lookup explicitly defined individual measurements
                DataRow[] explicitMeasurement = DataSource.Tables["SubscriberMeasurements"].Select(string.Format("SubscriberID='{0}' AND SignalID='{1}'", subscriberID, signalID));

                if (explicitMeasurement.Length > 0)
                    return explicitMeasurement.All(row => row["Allowed"].ToNonNullString("0").ParseBoolean());

                // Lookup implicitly defined group based measurements
                DataRow[] implicitMeasurements;

                foreach (DataRow subscriberMeasurementGroup in DataSource.Tables["SubscriberMeasurementGroups"].Select(string.Format("SubscriberID='{0}'", subscriberID)))
                {
                    implicitMeasurements = DataSource.Tables["MeasurementGroupMeasurements"].Select(string.Format("SignalID='{0}' AND MeasurementGroupID={1}", signalID, int.Parse(subscriberMeasurementGroup["MeasurementGroupID"].ToNonNullString("0"))));

                    if (implicitMeasurements.Length > 0)
                    {
                        // If all implicit rules allow access, we can return true;
                        // if even one rule denies access, return false
                        if (subscriberMeasurementGroup["Allowed"].ToNonNullString("0").ParseBoolean())
                            allowed = true;
                        else
                            return false;
                    }
                }

                return allowed;
            }
            catch (Exception ex)
            {
                LogDispatcher.E(StrLogSource, "ProcessException", "Failed to determine subscriber rights for {0} due to exception: {1}{2}{3}", GetConnectionProperty(subscriberID, cc => cc.SubscriberAcronym), ex.Message, Environment.NewLine, ex.StackTrace);
                LogDispatcher.E(StrLogSource, "ProcessException", ex);
            }

            return false;
        }

        /// <summary>
        /// Determines if subscriber has rights to specified <see cref="MeasurementKey"/>.
        /// </summary>
        /// <param name="subscriberID"><see cref="Guid"/> based subscriber ID.</param>
        /// <param name="key"><see cref="MeasurementKey"/> to lookup.</param>
        /// <param name="signalID"><see cref="Guid"/> signal ID if found; otherwise an empty Guid.</param>
        /// <returns><c>true</c> if subscriber has rights to specified <see cref="MeasurementKey"/>; otherwise <c>false</c>.</returns>
        protected bool SubscriberHasRights(Guid subscriberID, MeasurementKey key, out Guid signalID)
        {
            signalID = LookupSignalID(key);

            if (signalID != Guid.Empty)
                return SubscriberHasRights(subscriberID, signalID);

            return false;
        }

        private void UpdateClientNotifications()
        {
            Guid subscriberID;

            lock (m_clientNotificationsLock)
            {
                m_clientNotifications.Clear();

                foreach (DataRow row in DataSource.Tables["Subscribers"].Rows)
                {
                    if (Guid.TryParse(row["ID"].ToNonNullString(), out subscriberID))
                        m_clientNotifications.Add(subscriberID, new Dictionary<int, string>());
                }

                DeserializeClientNotifications();
            }
        }

        private void SerializeClientNotifications()
        {
            string notificationsFileName = FilePath.GetAbsolutePath(string.Format("{0}Notifications.txt", Name));

            // Delete existing file for reserialization
            if (File.Exists(notificationsFileName))
                File.Delete(notificationsFileName);

            using (FileStream fileStream = File.OpenWrite(notificationsFileName))
            using (TextWriter writer = new StreamWriter(fileStream))
            {
                foreach (KeyValuePair<Guid, Dictionary<int, string>> pair in m_clientNotifications)
                {
                    foreach (string notification in pair.Value.Values)
                        writer.WriteLine("{0},{1}", pair.Key, notification);
                }
            }
        }

        private void DeserializeClientNotifications()
        {
            string notificationsFileName = FilePath.GetAbsolutePath(string.Format("{0}Notifications.txt", Name));
            Dictionary<int, string> notifications;
            string notification;
            Guid subscriberID;

            if (File.Exists(notificationsFileName))
            {
                using (FileStream fileStream = File.OpenRead(notificationsFileName))
                using (TextReader reader = new StreamReader(fileStream))
                {
                    string line = reader.ReadLine();

                    while ((object)line != null)
                    {
                        int separatorIndex = line.IndexOf(',');

                        if (Guid.TryParse(line.Substring(0, separatorIndex), out subscriberID))
                        {
                            if (m_clientNotifications.TryGetValue(subscriberID, out notifications))
                            {
                                notification = line.Substring(separatorIndex + 1);
                                notifications.Add(notification.GetHashCode(), notification);
                            }
                        }

                        line = reader.ReadLine();
                    }
                }
            }
        }

        private void SendAllNotifications()
        {
            foreach (ClientConnection connection in m_clientConnections.Values)
                SendNotifications(connection);
        }

        private void SendNotifications(ClientConnection connection)
        {
            Dictionary<int, string> notifications;
            byte[] hash;
            byte[] message;

            using (MemoryStream buffer = new MemoryStream())
            {
                if (m_clientNotifications.TryGetValue(connection.SubscriberID, out notifications))
                {
                    foreach (KeyValuePair<int, string> pair in notifications)
                    {
                        hash = EndianOrder.BigEndian.GetBytes(pair.Key);
                        message = connection.Encoding.GetBytes(pair.Value);

                        buffer.Write(hash, 0, hash.Length);
                        buffer.Write(message, 0, message.Length);
                        SendClientResponse(connection.ClientID, ServerResponse.Notify, ServerCommand.Subscribe, buffer.ToArray());
                        buffer.Position = 0;
                    }
                }
            }
        }

        /// <summary>
        /// Looks up <see cref="Guid"/> signal ID for given <see cref="MeasurementKey"/>.
        /// </summary>
        /// <param name="key"><see cref="MeasurementKey"/> to lookup.</param>
        /// <returns><see cref="Guid"/> signal ID if found; otherwise an empty Guid.</returns>
        protected Guid LookupSignalID(MeasurementKey key)
        {
            // Attempt to lookup measurement key's signal ID that may have already been cached
            return m_signalIDCache.GetOrAdd(key, keyParam =>
            {
                Guid signalID = Guid.Empty;

                if (DataSource != null && DataSource.Tables.Contains("ActiveMeasurements"))
                {
                    // Attempt to lookup input measurement keys for given source IDs from default measurement table, if defined
                    try
                    {
                        DataRow[] filteredRows = DataSource.Tables["ActiveMeasurements"].Select("ID='" + keyParam.ToString() + "'");

                        if (filteredRows.Length > 0)
                            Guid.TryParse(filteredRows[0]["SignalID"].ToString(), out signalID);
                    }
                    catch
                    {
                        // Errors here are not catastrophic, this simply limits the auto-assignment of input measurement keys based on specified source ID's
                    }
                }

                if (signalID == Guid.Empty)
                    signalID = key.SignalID;

                return signalID;
            });
        }

        /// <summary>
        /// Gets the text encoding associated with a particular client.
        /// </summary>
        /// <param name="clientID">ID of client.</param>
        /// <returns>Text encoding associated with a particular client.</returns>
        protected internal Encoding GetClientEncoding(Guid clientID)
        {
            ClientConnection connection;

            if (m_clientConnections.TryGetValue(clientID, out connection))
            {
                Encoding clientEncoding = connection.Encoding;

                if ((object)clientEncoding != null)
                    return clientEncoding;
            }

            // Default to unicode
            return Encoding.Unicode;
        }

        /// <summary>
        /// Sends the start time of the first measurement in a connection transmission.
        /// </summary>
        /// <param name="clientID">ID of client to send response.</param>
        /// <param name="startTime">Start time, in <see cref="Ticks"/>, of first measurement transmitted.</param>
        internal protected virtual bool SendDataStartTime(Guid clientID, Ticks startTime)
        {
            bool result = SendClientResponse(clientID, ServerResponse.DataStartTime, ServerCommand.Subscribe, EndianOrder.BigEndian.GetBytes((long)startTime));

            ClientConnection connection;

            if (m_clientConnections.TryGetValue(clientID, out connection))
                LogDispatcher.I(StrLogSource, "StatusMessage", "Start time sent to {0}.", connection.ConnectionID);

            return result;
        }

        // Handle input processing complete notifications

        /// <summary>
        /// Sends response back to specified client.
        /// </summary>
        /// <param name="clientID">ID of client to send response.</param>
        /// <param name="response">Server response.</param>
        /// <param name="command">In response to command.</param>
        /// <returns><c>true</c> if send was successful; otherwise <c>false</c>.</returns>
        internal protected virtual bool SendClientResponse(Guid clientID, ServerResponse response, ServerCommand command)
        {
            return SendClientResponse(clientID, response, command, (byte[])null);
        }

        /// <summary>
        /// Sends response back to specified client with a message.
        /// </summary>
        /// <param name="clientID">ID of client to send response.</param>
        /// <param name="response">Server response.</param>
        /// <param name="command">In response to command.</param>
        /// <param name="status">Status message to return.</param>
        /// <returns><c>true</c> if send was successful; otherwise <c>false</c>.</returns>
        internal protected virtual bool SendClientResponse(Guid clientID, ServerResponse response, ServerCommand command, string status)
        {
            if (status != null)
                return SendClientResponse(clientID, response, command, GetClientEncoding(clientID).GetBytes(status));

            return SendClientResponse(clientID, response, command);
        }

        /// <summary>
        /// Sends response back to specified client with a formatted message.
        /// </summary>
        /// <param name="clientID">ID of client to send response.</param>
        /// <param name="response">Server response.</param>
        /// <param name="command">In response to command.</param>
        /// <param name="formattedStatus">Formatted status message to return.</param>
        /// <param name="args">Arguments for <paramref name="formattedStatus"/>.</param>
        /// <returns><c>true</c> if send was successful; otherwise <c>false</c>.</returns>
        internal protected virtual bool SendClientResponse(Guid clientID, ServerResponse response, ServerCommand command, string formattedStatus, params object[] args)
        {
            if (!string.IsNullOrWhiteSpace(formattedStatus))
                return SendClientResponse(clientID, response, command, GetClientEncoding(clientID).GetBytes(string.Format(formattedStatus, args)));

            return SendClientResponse(clientID, response, command);
        }

        /// <summary>
        /// Sends response back to specified client with attached data.
        /// </summary>
        /// <param name="clientID">ID of client to send response.</param>
        /// <param name="response">Server response.</param>
        /// <param name="command">In response to command.</param>
        /// <param name="data">Data to return to client; null if none.</param>
        /// <returns><c>true</c> if send was successful; otherwise <c>false</c>.</returns>
        internal protected virtual bool SendClientResponse(Guid clientID, ServerResponse response, ServerCommand command, byte[] data)
        {
            return SendClientResponse(clientID, (byte)response, (byte)command, data);
        }

        // Attempts to get the subscriber for the given client based on that client's X.509 certificate.
        private void TryFindClientDetails(ClientConnection connection)
        {
            TlsServer commandChannel = m_commandChannel as TlsServer;
            TransportProvider<TlsServer.TlsSocket> client;
            X509Certificate remoteCertificate;
            X509Certificate trustedCertificate;
            DataRow subscriber;

            // If connection is not TLS, there is no X.509 certificate
            if ((object)commandChannel == null)
                return;

            // If connection is not found, cannot get X.509 certificate
            if (!commandChannel.TryGetClient(connection.ClientID, out client))
                return;

            // Get remote certificate and corresponding trusted certificate
            remoteCertificate = client.Provider.SslStream.RemoteCertificate;
            trustedCertificate = m_certificateChecker.GetTrustedCertificate(remoteCertificate);

            if ((object)trustedCertificate != null)
            {
                if (m_subscriberIdentities.TryGetValue(trustedCertificate, out subscriber))
                {
                    // Load client details from subscriber identity
                    connection.SubscriberID = Guid.Parse(subscriber["ID"].ToNonNullString(Guid.Empty.ToString()).Trim());
                    connection.SubscriberAcronym = subscriber["Acronym"].ToNonNullString().Trim();
                    connection.SubscriberName = subscriber["Name"].ToNonNullString().Trim();
                    connection.ValidIPAddresses = ParseAddressList(subscriber["ValidIPAddresses"].ToNonNullString());
                }
            }
        }

        // Parses a list of IP addresses.
        private List<IPAddress> ParseAddressList(string addressList)
        {
            string[] splitList = addressList.Split(';', ',');
            List<IPAddress> ipAddressList = new List<IPAddress>();
            IPAddress ipAddress;

            foreach (string address in splitList)
            {
                if (IPAddress.TryParse(address.Trim(), out ipAddress))
                    ipAddressList.Add(ipAddress);
            }

            return ipAddressList;
        }

        // Update certificate validation routine.
        private void UpdateCertificateChecker()
        {
            CertificatePolicy policy;
            SslPolicyErrors validPolicyErrors;
            X509ChainStatusFlags validChainFlags;

            string trustedCertificatesPath;
            string certificateFile;
            X509Certificate certificate;

            if ((object)m_certificateChecker == null)
                return;

            if ((object)m_commandChannel == null)
                return;

            m_certificateChecker.DistrustAll();
            m_subscriberIdentities.Clear();
            trustedCertificatesPath = FilePath.GetAbsolutePath(m_commandChannel.TrustedCertificatesPath);

            foreach (DataRow subscriber in DataSource.Tables["Subscribers"].Select("Enabled <> 0"))
            {
                certificateFile = Path.Combine(trustedCertificatesPath, subscriber["CertificateFile"].ToNonNullString());

                if (File.Exists(certificateFile))
                {
                    policy = new CertificatePolicy();

                    if (Enum.TryParse(subscriber["ValidPolicyErrors"].ToNonNullString(), out validPolicyErrors))
                        policy.ValidPolicyErrors = validPolicyErrors;

                    if (Enum.TryParse(subscriber["ValidChainFlags"].ToNonNullString(), out validChainFlags))
                        policy.ValidChainFlags = validChainFlags;

                    certificate = new X509Certificate2(certificateFile);
                    m_certificateChecker.Trust(certificate, policy);
                    m_subscriberIdentities.Add(certificate, subscriber);
                }
            }
        }

        // Update rights for the given subscription.
        private void UpdateRights(ClientConnection connection)
        {
            Guid subscriberID = connection.SubscriberID;
            UnsynchronizedClientSubscription subscription;

            // TODO: Signal OutputAdapter for this subscription
            //if (s_clientSubscriptions.TryGetValue(subscriberID, out subscription))
            //{
            //    MeasurementKey[] requestedInputs = LegacyAdapterBase.ParseInputMeasurementKeys(dataSource, connection.RequestedInputFilter);
            //    HashSet<Guid> authorizedSignals = new HashSet<Guid>();
            //    string message;

            //    foreach (MeasurementKey input in requestedInputs)
            //    {
            //        if (SubscriberHasRights(subscriberID, input.SignalID))
            //            authorizedSignals.Add(input.SignalID);
            //    }

            //    if (!authorizedSignals.SetEquals(subscription.RequestedInputs))
            //    {
            //        message = string.Format("Update to authorized signals caused subscription to change. Now subscribed to {0} signals.", authorizedSignals.Count);
            //        subscription.RequestedInputs = authorizedSignals;
            //        SendClientResponse(connection.ClientID, ServerResponse.Succeeded, ServerCommand.Subscribe, message);
            //    }
            //}
        }

        // Send binary response packet to client
        private bool SendClientResponse(Guid clientID, byte responseCode, byte commandCode, byte[] data)
        {
            ClientConnection connection;
            bool success = false;

            // Attempt to lookup associated client connection
            if (m_clientConnections.TryGetValue(clientID, out connection) && (object)connection != null)
            {
                try
                {
                    MemoryStream responsePacket = new MemoryStream();
                    bool dataPacketResponse = responseCode == (byte)ServerResponse.DataPacket;

                    // Add response code
                    responsePacket.WriteByte(responseCode);

                    // Add original in response to command code
                    responsePacket.WriteByte(commandCode);

                    if ((object)data == null || data.Length == 0)
                    {
                        // Add zero sized data buffer to response packet
                        responsePacket.Write(s_zeroLengthBytes, 0, 4);
                    }
                    else
                    {
                        // If response is for a data packet and a connection key is defined, encrypt the data packet payload
                        if (dataPacketResponse && (object)connection.KeyIVs != null)
                        {
                            // Get a local copy of volatile keyIVs and cipher index since these can change at any time
                            byte[][][] keyIVs = connection.KeyIVs;
                            int cipherIndex = connection.CipherIndex;

                            // Reserve space for size of data buffer to go into response packet
                            responsePacket.Write(s_zeroLengthBytes, 0, 4);

                            // Get data packet flags
                            DataPacketFlags flags = (DataPacketFlags)data[0];

                            // Encode current cipher index into data packet flags
                            if (cipherIndex > 0)
                                flags |= DataPacketFlags.CipherIndex;

                            // Write data packet flags into response packet
                            responsePacket.WriteByte((byte)flags);

                            // Copy source data payload into a memory stream
                            MemoryStream sourceData = new MemoryStream(data, 1, data.Length - 1);

                            // Encrypt payload portion of data packet and copy into the response packet
                            Common.SymmetricAlgorithm.Encrypt(sourceData, responsePacket, keyIVs[cipherIndex][0], keyIVs[cipherIndex][1]);

                            // Calculate length of encrypted data payload
                            int payloadLength = (int)responsePacket.Length - 6;

                            // Move the response packet position back to the packet size reservation
                            responsePacket.Seek(2, SeekOrigin.Begin);

                            // Add the actual size of payload length to response packet
                            responsePacket.Write(EndianOrder.BigEndian.GetBytes(payloadLength), 0, 4);
                        }
                        else
                        {
                            // Add size of data buffer to response packet
                            responsePacket.Write(EndianOrder.BigEndian.GetBytes(data.Length), 0, 4);

                            // Add data buffer
                            responsePacket.Write(data, 0, data.Length);
                        }
                    }

                    // Data packets can be published on a UDP data channel, so check for this...
                    IServer publishChannel = dataPacketResponse ? connection.PublishChannel : m_commandChannel;

                    // Send response packet
                    if ((object)publishChannel != null && publishChannel.CurrentState == ServerState.Running)
                    {
                        byte[] responseData = responsePacket.ToArray();

                        if (publishChannel is UdpServer)
                            publishChannel.MulticastAsync(responseData, 0, responseData.Length);
                        else
                            publishChannel.SendToAsync(clientID, responseData, 0, responseData.Length);

                        success = true;
                    }
                }
                catch (ObjectDisposedException)
                {
                    // This happens when there is still data to be sent to a disconnected client - we can safely ignore this exception
                }
                catch (NullReferenceException)
                {
                    // This happens when there is still data to be sent to a disconnected client - we can safely ignore this exception
                }
                catch (SocketException ex)
                {
                    if (!HandleSocketException(clientID, ex))
                        LogDispatcher.E(StrLogSource, "ProcessException", new InvalidOperationException(String.Format("Failed to send response packet to client due to exception:{0}", ex.Message), ex));
                }
                catch (InvalidOperationException ex)
                {
                    // Could still be processing threads with client data after client has been disconnected, this can be safely ignored
                    if (!ex.Message.StartsWith("No client found"))
                        LogDispatcher.E(StrLogSource, "ProcessException", new InvalidOperationException(String.Format("Failed to send response packet to client due to exception:{0}", ex.Message), ex));
                       
                }
                catch (Exception ex)
                {
                    LogDispatcher.E(StrLogSource, "ProcessException", new InvalidOperationException(String.Format("Failed to send response packet to client due to exception:{0}", ex.Message), ex));
                 
                }
            }

            return success;
        }

        private int GetTimestampSize(ClientConnection connection)
        {
            return !connection.UseMillisecondResolution ? 4 : 2;
        }

        // Socket exception handler
        private bool HandleSocketException(Guid clientID, Exception ex)
        {
            SocketException socketException = ex as SocketException;

            if ((object)socketException != null)
            {
                // WSAECONNABORTED and WSAECONNRESET are common errors after a client disconnect,
                // if they happen for other reasons, make sure disconnect procedure is handled
                if (socketException.ErrorCode == 10053 || socketException.ErrorCode == 10054)
                {
                    DisconnectClient(clientID);
                    return true;
                }
            }

            if ((object)ex != null)
                HandleSocketException(clientID, ex.InnerException);

            return false;
        }

        // Disconnect client - this should be called from non-blocking thread (e.g., thread pool)
        private void DisconnectClient(Guid clientID)
        {
            try
            {
                List<ClientConnection> subscriberConnections;
                ClientConnection clientConnection;
                uint subscriptionID;

                if (m_clientConnections.TryRemove(clientID, out clientConnection))
                {
                    lock (m_subscriberConnectionsLock)
                    {
                        if (m_subscriberConnections.TryGetValue(clientConnection.SubscriberID, out subscriberConnections))
                        {
                            subscriberConnections.Remove(clientConnection);

                            if (ConfigGraph.GetProviderFor(clientConnection.SubscriberID, out subscriptionID))
                                Receiver.SendNotification(subscriptionID, "sendThese", new HashSet<Guid>(subscriberConnections.SelectMany(connection => connection.SignalIndexCache.Reference.Values).Select(value => value.Item1)));

                            if (subscriberConnections.Count == 0)
                            {
                                m_subscriberConnections.Remove(clientConnection.SubscriberID);
                                MessageBusProxy.UnSubscribe(new HashSet<Guid>() { clientConnection.SubscriberID });
                            }
                        }
                    }

                    clientConnection.Dispose();
                    LogDispatcher.I(StrLogSource, "StatusMessage", "Client disconnected from command channel.");
                }
            }
            catch (Exception ex)
            {
                LogDispatcher.E(StrLogSource, "ProcessException",
                                new InvalidOperationException(
                                    String.Format("Encountered an exception while processing client disconnect: {0}", ex.Message),
                                    ex));
            }
        }

        // Gets specfied property from client connection based on subscriber ID
        private TResult GetConnectionProperty<TResult>(Guid subscriberID, Func<ClientConnection, TResult> predicate)
        {
            TResult result = default(TResult);

            // Lookup client connection by subscriber ID
            ClientConnection connection = m_clientConnections.Values.FirstOrDefault(cc => cc.SubscriberID == subscriberID);

            // Extract desired property from client connection using given predicate function
            if ((object)connection != null)
                result = predicate(connection);

            return result;
        }

        /// <summary>
        /// Raises the <see cref="ClientConnected"/> event.
        /// </summary>
        /// <param name="subscriberID">Subscriber <see cref="Guid"/> (normally <see cref="ClientConnection.SubscriberID"/>).</param>
        /// <param name="connectionID">Connection identification (normally <see cref="ClientConnection.ConnectionID"/>).</param>
        /// <param name="subscriberInfo">Subscriber information (normally <see cref="ClientConnection.SubscriberInfo"/>).</param>
        protected virtual void OnClientConnected(Guid subscriberID, string connectionID, string subscriberInfo)
        {
            try
            {
                if ((object)ClientConnected != null)
                    ClientConnected(this, new EventArgs<Guid, string, string>(subscriberID, connectionID, subscriberInfo));
            }
            catch (Exception ex)
            {
                // We protect our code from consumer thrown exceptions
                LogDispatcher.E(StrLogSource,
                                "ProcessException",
                                new InvalidOperationException(String.Format("Exception in consumer handler for ClientConnected event: {0}", ex.Message), ex)
                             );
            }
        }

        // Cipher key rotation timer handler
        private void m_cipherKeyRotationTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            if ((object)m_clientConnections != null)
            {
                foreach (ClientConnection connection in m_clientConnections.Values)
                {
                    if ((object)connection != null && connection.Authenticated)
                        connection.RotateCipherKeys();
                }
            }
        }

        // Command channel restart timer handler
        private void m_commandChannelRestartTimer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            if ((object)m_commandChannel != null)
            {
                try
                {
                    // After a short delay, we try to restart the command channel
                    m_commandChannel.Start();
                }
                catch (Exception ex)
                {
                    LogDispatcher.E(StrLogSource,"ProcessException", new InvalidOperationException(
                                        String.Format("Failed to restart data publisher command channel: {0}", ex.Message),
                                        ex)
                                    );
                }
            }
        }

        public void doit()
        {
            IMeasurement[] rawValues = MessageBusProxy.Recv();
            ICollection<CompactMeasurement> compactMeasurements;
            ICollection<BufferBlockMeasurement> bufferBlockMeasurements;

            Queue<CompactMeasurement> measurementQueue;
            ICollection<CompactMeasurement> packet;
            CompactMeasurement measurement;
            Ticks epoch;
            Ticks diffLimit;

            byte[] bufferBlock;
            ushort bufferBlockSignalIndex;

            List<ClientConnection> connections;

            foreach (IMeasurement raw in rawValues)
            {
                lock (m_subscriberConnectionsLock)
                {
                    if (m_subscriberConnections.TryGetValue(raw.StreamID, out connections))
                    {
                        compactMeasurements = raw.Value.GetRaw().Item1;
                        bufferBlockMeasurements = raw.Value.GetRaw().Item2;

                        foreach (ClientConnection connection in connections)
                        {
                            if ((object)compactMeasurements != null)
                            {
                                ClientConnection connection1 = connection;
                                IEnumerable<CompactMeasurement> filteredMeasurements = compactMeasurements.Where(m => connection1.SignalIndexCache.GetSignalIndex(m.SignalID) != ushort.MaxValue).ToArray();

                                if (filteredMeasurements.Any())
                                {
                                    measurementQueue = new Queue<CompactMeasurement>(filteredMeasurements.OrderBy(m => m.Timestamp));
                                    packet = new Collection<CompactMeasurement>();
                                    epoch = measurementQueue.Peek().Timestamp;
                                    diffLimit = !connection.UseMillisecondResolution ? uint.MaxValue : ushort.MaxValue;

                                    while (measurementQueue.Count > 0)
                                    {
                                        measurement = measurementQueue.Dequeue();

                                        if (measurement.Timestamp - epoch > diffLimit)
                                        {
                                            SendPacket(connection, epoch, packet);
                                            m_processedMeasurements += packet.Count;
                                            epoch = measurement.Timestamp;
                                            packet.Clear();
                                        }

                                        packet.Add(measurement);
                                    }

                                    if (packet.Count > 0)
                                    {
                                        SendPacket(connection, epoch, packet);
                                        m_processedMeasurements += packet.Count;
                                    }
                                }
                            }

                            if ((object)bufferBlockMeasurements != null)
                            {
                                ClientConnection connection1 = connection;
                                foreach (BufferBlockMeasurement bufferBlockMeasurement in bufferBlockMeasurements.Where(m => connection1.SignalIndexCache.GetSignalIndex(m.StreamID) != ushort.MaxValue))
                                {
                                    // Handle buffer block measurements as a special case - this can be any kind of data,
                                    // measurement subscriber will need to know how to interpret buffer
                                    bufferBlock = new byte[6 + bufferBlockMeasurement.Buffer.Length];

                                    // Prepend sequence number
                                    EndianOrder.BigEndian.CopyBytes(connection.BufferBlockSequenceNumber, bufferBlock, 0);
                                    connection.BufferBlockSequenceNumber++;

                                    // Copy signal index into buffer
                                    bufferBlockSignalIndex = connection.SignalIndexCache.GetSignalIndex(bufferBlockMeasurement.StreamID);
                                    EndianOrder.BigEndian.CopyBytes(bufferBlockSignalIndex, bufferBlock, 4);

                                    // Append measurement data and send
                                    Buffer.BlockCopy(bufferBlockMeasurement.Buffer, 0, bufferBlock, 6, bufferBlockMeasurement.Length);
                                    SendClientResponse(connection.ClientID, ServerResponse.BufferBlock, ServerCommand.Subscribe, bufferBlock);
                                }

                                m_processedMeasurements += bufferBlockMeasurements.Count;
                            }
                        }
                    }
                }
            }
        }

        private void SendPacket(ClientConnection connection, Ticks epoch, ICollection<CompactMeasurement> measurements)
        {
            MemoryStream packet = new MemoryStream();

            packet.WriteByte((byte)DataPacketFlags.Compact);
            packet.Write(EndianOrder.BigEndian.GetBytes((long)epoch), 0, 8);
            packet.Write(EndianOrder.BigEndian.GetBytes(measurements.Count), 0, 4);

            foreach (CompactMeasurement measurement in measurements)
            {
                ushort runtimeID = connection.SignalIndexCache.GetSignalIndex(measurement.SignalID);
                long timeDiff = measurement.Timestamp - epoch;

                packet.WriteByte((byte)measurement.StateFlags);
                packet.Write(EndianOrder.BigEndian.GetBytes(runtimeID), 0, 2);
                packet.Write(EndianOrder.BigEndian.GetBytes(measurement.Value), 0, 4);

                if (!connection.UseMillisecondResolution)
                    packet.Write(EndianOrder.BigEndian.GetBytes((uint)timeDiff), 0, 4);
                else
                    packet.Write(EndianOrder.BigEndian.GetBytes((ushort)(timeDiff / Ticks.PerMillisecond)), 0, 2);
            }

            SendClientResponse(connection.ClientID, ServerResponse.DataPacket, ServerCommand.Subscribe, packet.ToArray());
        }

        #region [ Server Command Request Handlers ]

        // Handles authentication request
        private void HandleAuthenticationRequest(ClientConnection connection)
        {
            string message = "WARNING: Received authentication request from client while running in TLS mode.";
            SendClientResponse(connection.ClientID, ServerResponse.Failed, ServerCommand.Authenticate, message);
            LogDispatcher.W(StrLogSource, "StatusMessage", message);
        }

        // Handles subscribe request
        private void HandleSubscribeRequest(ClientConnection connection, byte[] buffer, int startIndex, int length)
        {
            Guid clientID = connection.ClientID;
            string message, setting;

            // Handle subscribe
            try
            {
                // Make sure there is enough buffer for flags and connection string length
                if (length >= 6)
                {
                    // Next byte is the data packet flags
                    DataPacketFlags flags = (DataPacketFlags)buffer[startIndex];
                    startIndex++;

                    bool useSynchronizedSubscription = ((byte)(flags & DataPacketFlags.Synchronized) > 0);

                    if (useSynchronizedSubscription)
                    {
                        // Remotely synchronized subscriptions are currently disallowed by data publisher
                        message = "Client request for remotely synchronized data subscription was denied. Data publisher is currently configured to disallow synchronized subscriptions.";
                        SendClientResponse(clientID, ServerResponse.Failed, ServerCommand.Subscribe, message);
                        LogDispatcher.E(StrLogSource, "ProcessException", message);
                    }
                    else
                    {
                        bool useCompactMeasurementFormat = ((byte)(flags & DataPacketFlags.Compact) > 0);

                        // Non-compact format not supported by SIEGate
                        if (!useCompactMeasurementFormat)
                            throw new InvalidOperationException("Non-compact measurement format not supported by SIEGate.");

                        // Next 4 bytes are an integer representing the length of the connection string that follows
                        int byteLength = EndianOrder.BigEndian.ToInt32(buffer, startIndex);
                        startIndex += 4;

                        if (byteLength > 0 && length >= 6 + byteLength)
                        {
                            string connectionString = GetClientEncoding(clientID).GetString(buffer, startIndex, byteLength);
                            Dictionary<string, string> subscriptionSettings = connectionString.ParseKeyValuePairs();
                            MeasurementKey[] inputMeasurementKeys = null;

                            // Pass subscriber assembly information to connection, if defined
                            if (subscriptionSettings.TryGetValue("assemblyInfo", out setting))
                                connection.SubscriberInfo = setting;

                            // Set up UDP data channel if client has requested this
                            connection.DataChannel = null;

                            if (subscriptionSettings.TryGetValue("dataChannel", out setting))
                            {
                                Socket clientSocket = connection.GetCommandChannelSocket();
                                Dictionary<string, string> dataChannelSettings = setting.ParseKeyValuePairs();
                                IPEndPoint localEndPoint = null;
                                string networkInterface = "::0";

                                // Make sure return interface matches incoming client connection
                                if ((object)clientSocket != null)
                                    localEndPoint = clientSocket.LocalEndPoint as IPEndPoint;

                                if ((object)localEndPoint != null)
                                {
                                    networkInterface = localEndPoint.Address.ToString();

                                    // Remove dual-stack prefix
                                    if (networkInterface.StartsWith("::ffff:", true, CultureInfo.InvariantCulture))
                                        networkInterface = networkInterface.Substring(7);
                                }

                                if (dataChannelSettings.TryGetValue("port", out setting) || dataChannelSettings.TryGetValue("localport", out setting))
                                {
                                    connection.DataChannel = new UdpServer(string.Format("Port=-1; Clients={0}:{1}; interface={2}", connection.IPAddress, int.Parse(setting), networkInterface));
                                    connection.DataChannel.Start();
                                }
                            }

                            if (subscriptionSettings.TryGetValue("inputMeasurementKeys", out setting))
                            {
                                // Send updated signal index cache to client with validated rights of the selected input measurement keys
                                connection.RequestedInputFilter = setting;
                                inputMeasurementKeys = LegacyAdapterBase.ParseInputMeasurementKeys(DataSource, setting);
                                UpdateSignalIndexCache(clientID, connection.SignalIndexCache, inputMeasurementKeys);
                                byte[] serializedSignalIndexCache = SerializeSignalIndexCache(clientID, connection.SignalIndexCache);
                                SendClientResponse(clientID, ServerResponse.UpdateSignalIndexCache, ServerCommand.Subscribe, serializedSignalIndexCache);
                            }

                            // TODO: Add a flag to the database to allow payload encryption to be subsciber specific instead of global to publisher...
                            // Send new or updated cipher keys
                            if (connection.Authenticated && EncryptPayload)
                                connection.RotateCipherKeys();

                            try
                            {
                                uint subscriptionID;
                                List<ClientConnection> subscriberConnections;

                                lock (m_subscriberConnectionsLock)
                                {
                                    if (m_subscriberConnections.TryGetValue(connection.SubscriberID, out subscriberConnections))
                                    {
                                        // Send signal to UCS to remove unsubscribed measurements
                                        if (ConfigGraph.GetProviderFor(connection.SubscriberID, out subscriptionID))
                                            Receiver.SendNotification(subscriptionID, "sendThese", new HashSet<Guid>(subscriberConnections.SelectMany(c => c.SignalIndexCache.Reference.Values).Select(value => value.Item1)));
                                    }
                                }

                                // Notify any direct publisher consumers about the new client connection
                                OnClientConnected(connection.SubscriberID, connection.ConnectionID, connection.SubscriberInfo);
                            }
                            catch (Exception ex)
                            {
                                LogDispatcher.E(StrLogSource,
                                                "ProcessException", new InvalidOperationException(String.Format(
                                                   "ClientConnected event handler exception: {0}", ex.Message),
                                                    ex)
                                               );
                            }

                            if ((object)inputMeasurementKeys != null)
                                message = string.Format("Client subscribed as {0}compact {1}synchronized with {2} signals.", useCompactMeasurementFormat ? "" : "non-", useSynchronizedSubscription ? "" : "un", inputMeasurementKeys.Length);
                            else
                                message = string.Format("Client subscribed as {0}compact {1}synchronized, but no signals were specified. Make sure \"inputMeasurementKeys\" setting is properly defined.", useCompactMeasurementFormat ? "" : "non-", useSynchronizedSubscription ? "" : "un");

                            connection.IsSubscribed = true;
                            SendClientResponse(clientID, ServerResponse.Succeeded, ServerCommand.Subscribe, message);
                            LogDispatcher.I(StrLogSource, "StatusMessage", message);
                        }
                        else
                        {
                            if (byteLength > 0)
                                message = "Not enough buffer was provided to parse client data subscription.";
                            else
                                message = "Cannot initialize client data subscription without a connection string.";

                            SendClientResponse(clientID, ServerResponse.Failed, ServerCommand.Subscribe, message);
                            LogDispatcher.E(StrLogSource, "ProcessException", message);
                        }
                    }
                }
                else
                {
                    message = "Not enough buffer was provided to parse client data subscription.";
                    SendClientResponse(clientID, ServerResponse.Failed, ServerCommand.Subscribe, message);
                    LogDispatcher.E(StrLogSource, "ProcessException", message);
                }
            }
            catch (Exception ex)
            {
                message = "Failed to process client data subscription due to exception: " + ex.Message;
                SendClientResponse(clientID, ServerResponse.Failed, ServerCommand.Subscribe, message);
                LogDispatcher.E(StrLogSource,
                                "ProcessException",
                                new InvalidOperationException(message, ex));
                               
            }
        }

        // Handles unsubscribe request
        private void HandleUnsubscribeRequest(ClientConnection connection)
        {
            List<ClientConnection> subscriberConnections;
            Guid clientID = connection.ClientID;
            uint subscriptionID;

            if (connection.IsSubscribed)
            {
                lock (m_subscriberConnectionsLock)
                {
                    // Clear out the signal index cache
                    connection.SignalIndexCache.Reference = new ConcurrentDictionary<ushort, Tuple<Guid, string, uint>>();

                    if (m_subscriberConnections.TryGetValue(connection.SubscriberID, out subscriberConnections))
                    {
                        // Send signal to UCS to remove unsubscribed measurements
                        if (ConfigGraph.GetProviderFor(connection.SubscriberID, out subscriptionID))
                            Receiver.SendNotification(subscriptionID, "sendThese", new HashSet<Guid>(subscriberConnections.SelectMany(c => c.SignalIndexCache.Reference.Values).Select(value => value.Item1)));
                    }
                }

                connection.IsSubscribed = false;
            }

            SendClientResponse(clientID, ServerResponse.Succeeded, ServerCommand.Unsubscribe, "Client unsubscribed.");
            LogDispatcher.I(StrLogSource, "StatusMessage", connection.ConnectionID + " unsubscribed.");
        }

        // Handles meta-data refresh request
        private void HandleMetadataRefresh(ClientConnection connection)
        {
            Guid clientID = connection.ClientID;
            Guid nodeID;
            string message;

            try
            {
                ConfigurationFile configFile = ConfigurationFile.Current;
                CategorizedSettingsElementCollection systemSettings = configFile.Settings["systemSettings"];

                // Initialize active node ID
                nodeID = Guid.Parse(systemSettings["NodeID"].Value);

                using (AdoDataConnection adoDatabase = new AdoDataConnection("systemSettings"))
                {
                    IDbConnection dbConnection = adoDatabase.Connection;
                    DataSet metadata = new DataSet();
                    DataTable table;

                    byte[] serializedMetadata;

                    // Copy key meta-data tables
                    foreach (string tableExpression in m_metadataTables.Split(';'))
                    {
                        if (!string.IsNullOrWhiteSpace(tableExpression))
                        {
                            // Query the table or view information from the database
                            table = dbConnection.RetrieveData(adoDatabase.AdapterType, tableExpression);

                            // Remove any expression from table name
                            Match regexMatch = Regex.Match(tableExpression, @"FROM \w+");
                            table.TableName = regexMatch.Value.Split(' ')[1];

                            // If table has a NodeID column, filter table data for just this node.
                            // Also, determine whether we need to check subscriber for rights to the data
                            bool applyNodeIDFilter = table.Columns.Contains("NodeID");
                            bool checkSubscriberRights = table.Columns.Contains("SignalID");

                            if (SharedDatabase || (!applyNodeIDFilter && !checkSubscriberRights))
                            {
                                // Add a copy of the results to the dataset for meta-data exchange
                                metadata.Tables.Add(table.Copy());
                            }
                            else
                            {
                                IEnumerable<DataRow> filteredRows;
                                List<DataRow> filteredRowList;

                                // Make a copy of the table structure
                                metadata.Tables.Add(table.Clone());

                                // Reduce data to only this node
                                if (applyNodeIDFilter)
                                    filteredRows = table.Select(string.Format("NodeID = '{0}'", nodeID));
                                else
                                    filteredRows = table.Rows.Cast<DataRow>();

                                // Reduce data to only what the subscriber has rights to
                                if (checkSubscriberRights)
                                    filteredRows = filteredRows.Where(row => SubscriberHasRights(connection.SubscriberID, adoDatabase.Guid(row, "SignalID")));

                                filteredRowList = filteredRows.ToList();

                                if (filteredRowList.Count > 0)
                                {
                                    DataTable metadataTable = metadata.Tables[table.TableName];

                                    // Manually copy-in each row into table
                                    foreach (DataRow row in filteredRowList)
                                    {
                                        DataRow newRow = metadataTable.NewRow();

                                        // Copy each column of data in the current row
                                        for (int x = 0; x < table.Columns.Count; x++)
                                        {
                                            newRow[x] = row[x];
                                        }

                                        metadataTable.Rows.Add(newRow);
                                    }
                                }
                            }
                        }
                    }

                    serializedMetadata = SerializeMetadata(clientID, metadata);
                    SendClientResponse(clientID, ServerResponse.Succeeded, ServerCommand.MetaDataRefresh, serializedMetadata);
                }
            }
            catch (Exception ex)
            {
                message = "Failed to transfer meta-data due to exception: " + ex.Message;
                SendClientResponse(clientID, ServerResponse.Failed, ServerCommand.Subscribe, message);
                LogDispatcher.E(StrLogSource,
                                "ProcessException",
                                new InvalidOperationException(message, ex));
            }
        }

        // Handles request to update processing interval on client session
        private void HandleUpdateProcessingInterval(ClientConnection connection, byte[] buffer, int startIndex, int length)
        {
            Guid clientID = connection.ClientID;
            string message;

            // Make sure there is enough buffer for new processing interval value
            if (length >= 4)
            {
                // Next 4 bytes are an integer representing the new processing interval
                int processingInterval = EndianOrder.BigEndian.ToInt32(buffer, startIndex);

                connection.ProcessingInterval = processingInterval;
                SendClientResponse(clientID, ServerResponse.Succeeded, ServerCommand.UpdateProcessingInterval, "New processing interval of {0} assigned.", processingInterval);
                LogDispatcher.I(StrLogSource, "StatusMessage", "{0} was assigned a new processing interval of {1}.", connection.ConnectionID, processingInterval);
            }
            else
            {
                message = "Not enough buffer was provided to update client processing interval.";
                SendClientResponse(clientID, ServerResponse.Failed, ServerCommand.UpdateProcessingInterval, message);
                LogDispatcher.E(StrLogSource, "ProcessException", message);
            }
        }

        // Handle request to define operational modes for client connection
        private void HandleDefineOperationalModes(ClientConnection connection, byte[] buffer, int startIndex, int length)
        {
            uint operationalModes;

            if (length >= 4)
            {
                operationalModes = EndianOrder.BigEndian.ToUInt32(buffer, startIndex);

                if ((operationalModes & (uint)OperationalModes.VersionMask) != 0u)
                    LogDispatcher.W(StrLogSource, "StatusMessage", "WARNING: Protocol version not supported. Operational modes may not be set correctly for client {0}.", connection.ClientID);

                connection.OperationalModes = (OperationalModes)operationalModes;
            }
        }

        // Handle confirmation of receipt of notification
        private void HandleConfirmNotification(ClientConnection connection, byte[] buffer, int startIndex, int length)
        {
            int hash = EndianOrder.BigEndian.ToInt32(buffer, startIndex);
            Dictionary<int, string> notifications;
            string notification;

            if (length >= 4)
            {
                lock (m_clientNotificationsLock)
                {
                    if (m_clientNotifications.TryGetValue(connection.SubscriberID, out notifications))
                    {
                        if (notifications.TryGetValue(hash, out notification))
                        {
                            notifications.Remove(hash);
                            LogDispatcher.I(StrLogSource, "StatusMessage", "Subscriber {0} confirmed receipt of notification: {1}.", connection.ConnectionID, notification);
                            SerializeClientNotifications();
                        }
                        else
                        {
                            LogDispatcher.I(StrLogSource, "StatusMessage", "Confirmation for unknown notification received from client {0}.", connection.ConnectionID);
                        }
                    }
                    else
                    {
                        LogDispatcher.I(StrLogSource, "StatusMessage", "Unsolicited confirmation of notification received.");
                    }
                }
            }
            else
            {
                LogDispatcher.I(StrLogSource, "StatusMessage", "Malformed notification confirmation received.");
            }
        }

        private byte[] SerializeSignalIndexCache(Guid clientID, SignalIndexCache signalIndexCache)
        {
            byte[] serializedSignalIndexCache = null;
            ClientConnection connection;

            if (m_clientConnections.TryGetValue(clientID, out connection))
            {
                OperationalModes operationalModes = connection.OperationalModes;
                GatewayCompressionMode gatewayCompressionMode = (GatewayCompressionMode)(operationalModes & OperationalModes.CompressionModeMask);
                bool useCommonSerializationFormat = (operationalModes & OperationalModes.UseCommonSerializationFormat) > 0;
                bool compressSignalIndexCache = (operationalModes & OperationalModes.CompressSignalIndexCache) > 0;

                MemoryStream compressedData = null;
                GZipStream deflater = null;

                if (!useCommonSerializationFormat)
                {
                    // Use standard .NET BinaryFormatter
                    serializedSignalIndexCache = Serialization.Serialize(signalIndexCache, GSF.SerializationFormat.Binary);
                }
                else
                {
                    // Use ISupportBinaryImage implementation
                    signalIndexCache.Encoding = GetClientEncoding(clientID);
                    serializedSignalIndexCache = new byte[signalIndexCache.BinaryLength];
                    signalIndexCache.GenerateBinaryImage(serializedSignalIndexCache, 0);
                }

                if (compressSignalIndexCache && gatewayCompressionMode == GatewayCompressionMode.GZip)
                {
                    try
                    {
                        // Compress serialized signal index cache into compressed data buffer
                        compressedData = new MemoryStream();
                        deflater = new GZipStream(compressedData, CompressionMode.Compress);
                        deflater.Write(serializedSignalIndexCache, 0, serializedSignalIndexCache.Length);
                        deflater.Close();
                        deflater = null;

                        serializedSignalIndexCache = compressedData.ToArray();
                    }
                    finally
                    {
                        if ((object)deflater != null)
                            deflater.Close();

                        if ((object)compressedData != null)
                            compressedData.Close();
                    }
                }
            }

            return serializedSignalIndexCache;
        }

        private byte[] SerializeMetadata(Guid clientID, DataSet metadata)
        {
            byte[] serializedMetadata = null;
            ClientConnection connection;

            if (m_clientConnections.TryGetValue(clientID, out connection))
            {
                OperationalModes operationalModes = connection.OperationalModes;
                GatewayCompressionMode gatewayCompressionMode = (GatewayCompressionMode)(operationalModes & OperationalModes.CompressionModeMask);
                bool useCommonSerializationFormat = (operationalModes & OperationalModes.UseCommonSerializationFormat) > 0;
                bool compressMetadata = (operationalModes & OperationalModes.CompressMetadata) > 0;

                MemoryStream encodedData = null;
                XmlTextWriter unicodeWriter = null;

                MemoryStream compressedData = null;
                GZipStream deflater = null;

                if (!useCommonSerializationFormat)
                {
                    serializedMetadata = Serialization.Serialize(metadata, GSF.SerializationFormat.Binary);
                }
                else
                {
                    try
                    {
                        // Encode XML into encoded data buffer
                        encodedData = new MemoryStream();
                        unicodeWriter = new XmlTextWriter(encodedData, GetClientEncoding(clientID));
                        metadata.WriteXml(unicodeWriter, XmlWriteMode.WriteSchema);
                        unicodeWriter.Close();
                        unicodeWriter = null;

                        // Return result of compression
                        serializedMetadata = encodedData.ToArray();
                    }
                    finally
                    {
                        if ((object)unicodeWriter != null)
                            unicodeWriter.Close();

                        if ((object)encodedData != null)
                            encodedData.Close();
                    }
                }

                if (compressMetadata && gatewayCompressionMode == GatewayCompressionMode.GZip)
                {
                    try
                    {
                        // Compress serialized metadata into compressed data buffer
                        compressedData = new MemoryStream();
                        deflater = new GZipStream(compressedData, CompressionMode.Compress);
                        deflater.Write(serializedMetadata, 0, serializedMetadata.Length);
                        deflater.Close();
                        deflater = null;

                        serializedMetadata = compressedData.ToArray();
                    }
                    finally
                    {
                        if ((object)deflater != null)
                            deflater.Close();

                        if ((object)compressedData != null)
                            compressedData.Close();
                    }
                }

            }

            return serializedMetadata;
        }

        #endregion

        #region [ Command Channel Handlers ]

        private void m_commandChannel_ReceiveClientDataComplete(object sender, EventArgs<Guid, byte[], int> e)
        {
            try
            {
                Guid clientID = e.Argument1;
                byte[] buffer = e.Argument2;
                int length = e.Argument3;
                int index = 0;

                if (length > 0 && buffer != null)
                {
                    ClientConnection connection;
                    ServerCommand command;
                    string message;
                    byte commandByte = buffer[index];
                    index++;

                    // Attempt to parse solicited server command
                    bool validServerCommand = Enum.TryParse<ServerCommand>(commandByte.ToString(), out command);

                    // Look up this client connection
                    if (!m_clientConnections.TryGetValue(clientID, out connection))
                    {
                        // Received a request from an unknown client, this request is denied
                        LogDispatcher.W(StrLogSource, "StatusMessage", "WARNING: Ignored {0} byte {1} command request received from an unrecognized client: {2}", length, validServerCommand ? command.ToString() : "unidentified", clientID);
                    }
                    else if (validServerCommand)
                    {
                        if (command != ServerCommand.DefineOperationalModes)
                        {
                            if (command == ServerCommand.Authenticate)
                            {
                                // Handle authenticate
                                HandleAuthenticationRequest(connection);
                                return;
                            }

                            if (!connection.Authenticated)
                            {
                                message = string.Format("Subscriber not authenticated - {0} request denied.", command);
                                SendClientResponse(clientID, ServerResponse.Failed, command, message);
                                LogDispatcher.W(StrLogSource, "StatusMessage", "WARNING: Client {0} {1} command request denied - subscriber not authenticated.", connection.ConnectionID, command);
                                return;
                            }
                        }

                        switch (command)
                        {
                            case ServerCommand.Subscribe:
                                // Handle subscribe
                                HandleSubscribeRequest(connection, buffer, index, length);
                                break;
                            case ServerCommand.Unsubscribe:
                                // Handle unsubscribe
                                HandleUnsubscribeRequest(connection);
                                break;
                            case ServerCommand.MetaDataRefresh:
                                // Handle meta data refresh (per subscriber request)
                                HandleMetadataRefresh(connection);
                                break;
                            case ServerCommand.RotateCipherKeys:
                                // Handle rotation of cipher keys (per subscriber request)
                                connection.RotateCipherKeys();
                                break;
                            case ServerCommand.UpdateProcessingInterval:
                                // Handle request to update processing interval
                                HandleUpdateProcessingInterval(connection, buffer, index, length);
                                break;
                            case ServerCommand.DefineOperationalModes:
                                // Handle request to define oeprational modes
                                HandleDefineOperationalModes(connection, buffer, index, length);
                                break;
                            case ServerCommand.ConfirmNotification:
                                // Handle confirmation of receipt of notification
                                HandleConfirmNotification(connection, buffer, index, length);
                                break;
                        }
                    }
                    else
                    {
                        // Handle unrecognized commands
                        message = " sent an unrecognized server command: 0x" + commandByte.ToString("X").PadLeft(2, '0');
                        SendClientResponse(clientID, (byte)ServerResponse.Failed, commandByte, GetClientEncoding(clientID).GetBytes("Client" + message));
                        LogDispatcher.W(StrLogSource, "ProcessException", "WARNING: " + connection.ConnectionID + message);
                    }
                }
            }
            catch (Exception ex)
            {
                LogDispatcher.E(StrLogSource,
                                "ProcessException",
                                new InvalidOperationException("Encountered an exception while processing received client data: " + ex.Message, ex));
            }
        }

        private void m_commandChannel_ClientConnected(object sender, EventArgs<Guid> e)
        {
            Guid clientID = e.Argument;
            ClientConnection connection = new ClientConnection(this, clientID, m_commandChannel);
            List<ClientConnection> connections;

            TryFindClientDetails(connection);
            m_clientConnections[clientID] = connection;
            connection.Authenticated = connection.ValidIPAddresses.Contains(connection.IPAddress);

            if (connection.Authenticated)
            {
                lock (m_subscriberConnectionsLock)
                {
                    if (!m_subscriberConnections.TryGetValue(connection.SubscriberID, out connections))
                    {
                        connections = new List<ClientConnection>();
                        m_subscriberConnections.Add(connection.SubscriberID, connections);
                        MessageBusProxy.Subscribe(new HashSet<Guid>() { connection.SubscriberID });
                    }

                    connections.Add(connection);
                }

                lock (m_clientNotificationsLock)
                {
                    // Send any queued notifications to authenticated client
                    SendNotifications(connection);
                }

                LogDispatcher.I(StrLogSource, "StatusMessage", "Client connected to command channel.");
            }
            else
            {
                LogDispatcher.E(StrLogSource, "ProcessException",
                                "Unable to authenticate client as \"{0}\". IP address [{1}] not found among the list of valid IP addresses.",
                                connection.SubscriberName, connection.IPAddress);
            }
        }

        private void m_commandChannel_ClientDisconnected(object sender, EventArgs<Guid> e)
        {
            DisconnectClient(e.Argument);
        }

        private void m_commandChannel_ClientConnectingException(object sender, EventArgs<Exception> e)
        {
            Exception ex = e.Argument;
            if (ex is SecurityException)
            {
                LogDispatcher.E(StrLogSource, "Security", ex.Message);
            }
            LogDispatcher.E(StrLogSource,
                            "ProcessException",
                            new InvalidOperationException("Data publisher encountered an exception while connecting client to command channel: " + ex.Message, ex));
        }

        private void m_commandChannel_ServerStarted(object sender, EventArgs e)
        {
            LogDispatcher.I(StrLogSource, "StatusMessage", "Data publisher command channel started.");
        }

        private void m_commandChannel_ServerStopped(object sender, EventArgs e)
        {
            // We must wait for command channel to completely shutdown before trying to restart...
            if (m_commandChannelRestartTimer != null)
                m_commandChannelRestartTimer.Start();
        }

        private void m_commandChannel_SendClientDataException(object sender, EventArgs<Guid, Exception> e)
        {
            Exception ex = e.Argument2;
            if (ex is SecurityException)
            {
                LogDispatcher.E(StrLogSource, "Security", ex.Message);
            }
            if (!HandleSocketException(e.Argument1, ex) && !(ex is NullReferenceException) &&
                !(ex is ObjectDisposedException))
                LogDispatcher.E(StrLogSource,
                                "ProcessException",
                                new InvalidOperationException("Data publisher encountered an exception while sending command channel data to client connection: " +
                                                              ex.Message, ex));
                                
        }

        private void m_commandChannel_ReceiveClientDataException(object sender, EventArgs<Guid, Exception> e)
        {
            Exception ex = e.Argument2;
            if (ex is SecurityException)
            {
                LogDispatcher.E(StrLogSource, "Security", ex.Message);
            }
            if (!HandleSocketException(e.Argument1, ex) && !(ex is NullReferenceException) &&
                !(ex is ObjectDisposedException))
            {
                string message = "Data publisher encountered an exception while receiving command channel data from client connection: {0}{1}{2}";
                message += ex.Message;
                LogDispatcher.E(StrLogSource,
                                "ProcessException",
                                new InvalidOperationException(message, ex));
            }
        }

        #endregion

        #endregion

        #region [ Static ]

        // Constant zero length integer byte array
        private readonly static byte[] s_zeroLengthBytes = new byte[] { 0, 0, 0, 0 };
        private const string StrLogSource = "DataPublisher";

        #endregion
    }
}
