/*****************************************************************************\
 *             University of Illinois/NCSA Open Source License
 *
 * 
 *      Copyright  2012,2013, The Board of Trustees of the University of Illinois. All rights reserved.
 *
 *      Developed by: Rakesh Bobba, Derek Dagit, Erich Heine, Denis Kholine,
 *                    Klara Nahrstedt, David Rodgers,   William H. Sanders, Tim Yardley
 *
 *      Information Trust Institute/Coordinated Science Laboratory
 * 
 *      University of Illinois
 *      
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      http://www.iti.illinois.edu / http://csl.illinois.edu
 * 
 *      Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
 *      associated documentation files (the Software), to deal with the Software without restriction, 
 *      including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *      and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
 *      subject to the following conditions:
 *
 *      Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *      following disclaimers.
 *      
 *      Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 *      the following disclaimers in the documentation and/or other materials provided with the distribution.
 *
 *      Neither the names of Information Trust institute, Coordinated Science Laboratory, University of Illinois,
 *      The Board of Trustees of the University of Illinois nor the names of its contributors may be used to 
 *      endorse or promote products derived from this Software without specific prior written permission.
 *      
 *      THE SOFTWARE IS PROVIDED AS IS, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 *      LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN
 *      NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *      WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 *      SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
 *      
 \*****************************************************************************/
using System;
using System.Diagnostics;
using System.Diagnostics.Eventing.Reader;

namespace SIEGate.Core.Logging
{
    public class EventViewerLogger : Logger
    {
        private string EventViewerSource = "SIEGate";
        private const string EventViewerLogType = "Application";

        public EventViewerLogger(string source = "SIEGate")
        {
            EventViewerSource = source;
        }

        public override void Log(LogLevel level, DateTime time, string source, string tag, string message)
        {
            CreateEventSourceIfNeeded();
            EventLog.WriteEntry(EventViewerSource,
                                string.Format("{0:G} {1:u}: {2}: {3}: {4}", level, time, source, tag, message),
                                GetEventLogEntryType(level));
        }

        public override void LogException(LogLevel level, DateTime time, string source, string tag, Exception ex)
        {
            CreateEventSourceIfNeeded();
            EventLog.WriteEntry(EventViewerSource,
                                string.Format("{0:G} {1:u}: {2}: {3}: {4}{5}{6}", level, time, source, tag, ex.Message, Environment.NewLine, ex.StackTrace),
                                GetEventLogEntryType(level));
        }

        private static EventLogEntryType GetEventLogEntryType(LogLevel level)
        {
            EventLogEntryType eventLogEntryType = EventLogEntryType.Information;
            switch (level)
            {
                case LogLevel.Debug:
                case LogLevel.Verbose:
                    eventLogEntryType = EventLogEntryType.Information;
                    break;
                case LogLevel.Warn:
                    eventLogEntryType = EventLogEntryType.Warning;
                    break;
                case LogLevel.Error:
                case LogLevel.Fatal:
                case LogLevel.Unexpected:
                    eventLogEntryType = EventLogEntryType.Error;
                    break;
            }
            return eventLogEntryType;
        }

        private void CreateEventSourceIfNeeded()
        {
            if (EventLog.SourceExists(EventViewerSource))
                return;
            // The first time this is called, the user must be an Administrator.
            // Since we generally won't run as Administrator, we'll need to do this
            // as part of the install process.
            // http://msdn.microsoft.com/en-us/library/k1wask0c.aspx
            try
            {
                EventLog.CreateEventSource(EventViewerSource, EventViewerLogType);
            }
            catch
            {
                throw new EventLogNotFoundException(String.Format("{0} {1} does not exist in EventLog, please create as an Administrator.", EventViewerSource, EventViewerLogType));
            }
        }
    }
}
