/*****************************************************************************\
 *             University of Illinois/NCSA Open Source License
 *
 * 
 *      Copyright  2012,2013, The Board of Trustees of the University of Illinois. All rights reserved.
 *
 *      Developed by: Rakesh Bobba, Derek Dagit, Erich Heine, Denis Kholine,
 *                    Klara Nahrstedt, David Rodgers, William H. Sanders, Tim Yardley
 *
 *      Information Trust Institute/Coordinated Science Laboratory
 * 
 *      University of Illinois
 *      
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      http://www.iti.illinois.edu / http://csl.illinois.edu
 * 
 *      Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
 *      associated documentation files (the Software), to deal with the Software without restriction, 
 *      including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *      and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
 *      subject to the following conditions:
 *
 *      Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *      following disclaimers.
 *      
 *      Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 *      the following disclaimers in the documentation and/or other materials provided with the distribution.
 *
 *      Neither the names of Information Trust institute, Coordinated Science Laboratory, University of Illinois,
 *      The Board of Trustees of the University of Illinois nor the names of its contributors may be used to 
 *      endorse or promote products derived from this Software without specific prior written permission.
 *      
 *      THE SOFTWARE IS PROVIDED AS IS, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 *      LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN
 *      NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *      WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 *      SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
 *      
 \*****************************************************************************/

#region [ Using ]
using System.Globalization;
using System.Reflection;
using GSF.Diagnostics;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using SIEGate.Core.Logging;
#endregion

namespace SIEGate.Core.Stats
{
    // Consider moving this to a separate namespace, particularly since StatInst, and StatsLogger are public classes.
    public class Stats
    {
        // Map name of stat to list of IDs of adapters
        // TODO: Aggregates
        //private readonly IDictionary<string, IDictionary<string, StatInst.Aggregate>> m_statTimers = new Dictionary<string, IDictionary<string, StatInst.Aggregate>>();

        // Map type (Counter or Duration) to stat name to adapter
        // Type and typeof(StatInst) must match.
        // e.g., { Counter => { "Bytes In" => { "Adapter A" => WeakReference(265) } }
        // Thinking this through, I don't think we want to use WeakReferences, 
        // because if we do, then we get in a state where it's possible for
        // monotonically increasing values (such as bytes out) to decrease.
        // That shouldn't happen.
        // Instead, we end up keeping stats around forever, which is a memory 
        // leak.
        // TODO: Create a way to aggregate expired stats, then remove them from this list
        private readonly IDictionary<Type, IDictionary<string, ISet<StatInst>>> m_statInstances = new Dictionary<Type, IDictionary<string, ISet<StatInst>>>();

        // HACK: We need a source ID for stats; using 0 for the moment as a "from this system," need a better ID in the long run.
        public const uint SystemStatSourceId = uint.MaxValue;
        public static Stats StatDispatcher
        {
            get
            {
                if (_mStats == null)
                {
                    _mStats = new Stats();
                    _mStats.Initialize();
                }
                return _mStats;
            }
        }
        
        public event Logger.LogDelegate OnLog;

        private string _mLogSource = "Stats";
        private string _mLogTag = "Stats";
        private LogLevel _mLogLevel = LogLevel.Info;
        private readonly IDictionary<uint, Cluster> _mClusters = new Dictionary<uint, Cluster>();
        private static Stats _mStats;

        public const string StrStatCpuUsage = "CPU Usage";
        public const string StrStatProcessCpuUsage = "Process CPU Usage";
        public const string StrStatProcessMemoryUsage = "Memory Usage";

        // TODO: Make sure all these stats are reflected.
        //"CPU Usage"; // Percentage of CPU currently used by this process.
        //"Average CPU Usage"; // Average percentage of CPU used by this process.
        //"Memory Usage"; // Amount of memory currently used by this process in megabytes.
        //"Average Memory Usage"; // Average amount of memory used by this process in megabytes.
        //"Thread Count"; // Number of threads currently used by this process.
        //"Average Thread Count"; // Average number of threads used by this process.
        //"Threading Contention Rate"; // Current thread lock contention rate in attempts per second.
        //"Average Threading Contention Rate"; // Average thread lock contention rate in attempts per second.
        //"IO Usage"; // Amount of IO currently used by this process in kilobytes per second.
        //"Average IO Usage"; // Average amount of IO used by this process in kilobytes per second.
        //"Datagram Send Rate"; // Number of IPv4 datagrams currently sent by this process per second.
        //"Average Datagram Send Rate"; // Average number of IPv4 datagrams sent by this process per second.
        //"Datagram Receive Rate"; // Number of IPv4 datagrams currently received by this process per second.
        //"Average Datagram Receive Rate"; // Average number of IPv4 datagrams received by this process per second.

        protected uint Guid
        {
            get { return SystemStatSourceId; }
        }

        public Cluster SystemStats
        {
            get { return _mClusters[SystemStatSourceId]; }
        }

        public LogLevel LogLevel
        {
            get { return _mLogLevel; }
            set { _mLogLevel = value; }
        }

        public string LogSource
        {
            get { return _mLogSource; }
            set { _mLogSource = value; }
        }

        public string LogTag
        {
            get { return _mLogTag; }
            set { _mLogTag = value; }
        }

        // Private since we want the Stats object to be a singleton.
        // Weirdness in testing indicated this constructor had to do something or it wouldn't be called when running tests, so call base()
        private Stats()
        {
           
        }
        private void Initialize()
        {
            // TODO:Add category exists verification.
            if (System.Diagnostics.PerformanceCounterCategory.Exists("Processor"))
            {
             // This can't be included in the constructor because the Register functions make reference to StatDispatcher which won't exist
            // at the time the constructor is being invoked.
            Cluster systemCluster = RegisterCluster(SystemStatSourceId);

                System.Diagnostics.PerformanceCounter sdpc = new System.Diagnostics.PerformanceCounter("Processor","% Processor Time","_Total");

            // Two questions:
            // 1) How do we use the GSF.Diagnostics.PerformanceCounter to get the performance data?
            // Pre-1)  Do we even want to use the GSF.Diagnostics.PerformanceCounter?
            // 2) Why is memory throwing an exception?  How do I get the memory of the current process?
            PerformanceCounter cpuCounter = new PerformanceCounter("Processor", "% Processor Time", "_Total");
            // RawValue isn't thread-safe, but the only cost is lost accuracy 
            // which we accept in stats in order to gain better performance
                systemCluster.RegisterDelegateDoubleGenerator("System", StrStatCpuUsage, () => (double)cpuCounter.LastValue);

            PerformanceCounter processCounter = new PerformanceCounter("Process", "% Processor Time", System.Diagnostics.Process.GetCurrentProcess().ProcessName);
            systemCluster.RegisterDelegateDoubleGenerator("System", StrStatProcessCpuUsage, () => (double)processCounter.LastValue);

            //PerformanceCounter memoryCounter = new PerformanceCounter("Memory", "Memory Used", System.Diagnostics.Process.GetCurrentProcess().ProcessName);
            //systemCluster.RegisterDelegateFloat("System", StrStatProcessMemoryUsage, () => { return memoryCounter.LastValue; });
        }
        }

        public Cluster RegisterCluster(uint id)
        {
            if (_mClusters.ContainsKey(id))
                return _mClusters[id];
            _mClusters.Add(id, new Cluster(id));
            return _mClusters[id];
        }

        public Cluster GetCluster(uint id)
        {
            if (_mClusters.ContainsKey(id))
            {
                return _mClusters[id];
            }
            return null;
        }

        /// <summary>
        /// Gets a StatInst for a given ID and statName.
        /// </summary>
        /// <param name="ID">The ID of the object from which to get the StatInst.</param>
        /// <param name="statName">The name of the statistic.</param>
        /// <returns>A StatInst for the statistic statName from the object identified by ID.</returns>
        public StatInst GetStatistic(uint id, string statName)
        {
            // TODO: Consider changing this to DuplcicateCurrentValue() instead.
            if (_mClusters.ContainsKey(id))
            {
                return _mClusters[id].GetStatInst(statName);
            }
            return null;
        }

        /// <summary>
        /// Since everything is accessed through a proxy called Cluster, we can be a bit looser about how this works. Always keep internal, we don't want
        /// people screwing around with this if we can avoid it.
        /// </summary>
        /// <param name="getOwnerNameFunction"></param>
        /// <param name="statName"></param>
        /// <param name="statType"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        internal StatInst RegisterStat(Func<string> getOwnerNameFunction, string statName, Type statType, params object[] args)
        {
            if (!typeof(StatInst).IsAssignableFrom(statType))
            {
                throw new Exception("Must register Stat Types!");
            }
            List<object> newargs = new List<object>() {getOwnerNameFunction, statName};
           newargs.AddRange(args);
            BindingFlags flags = BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.Public;
            CultureInfo culture = null;
            StatInst result = (StatInst)Activator.CreateInstance(statType, flags, null, newargs.ToArray(), culture);
            AddStat(statName, result);
            return result;
        }


        internal StatInst InstantiateStat(Func<string> getOwnerNameFunction, string statName, Type statType, params object[] args)
        {
            if (!typeof(StatInst).IsAssignableFrom(statType))
            {
                throw new Exception("Must instantiate Stat Types!");
            }
            List<object> newargs = new List<object>() { getOwnerNameFunction, statName };
            newargs.AddRange(args);
            BindingFlags flags = BindingFlags.NonPublic | BindingFlags.Instance | BindingFlags.Public;
            CultureInfo culture = null;
            StatInst result = (StatInst)Activator.CreateInstance(statType, flags, null, newargs.ToArray(), culture);
            // We do not register at this point since this is for UnregisteredClusters.
            return result;
        }

        private void AddStat(string statName, StatInst singleStatInstance)
        {
            Type statInstType = singleStatInstance.GetType();
            if (!m_statInstances.ContainsKey(statInstType))
            {
                m_statInstances.Add(statInstType, new Dictionary<string, ISet<StatInst>>());
            }
            if (!m_statInstances[statInstType].ContainsKey(statName))
            {
                m_statInstances[statInstType].Add(statName, new HashSet<StatInst>());
            }
            m_statInstances[statInstType][statName].Add(singleStatInstance);
        }

        /// <summary>
        /// Gets the total for a given Stat Name
        /// </summary>
        /// <param name="statName">The name of the statistic to be queried</param>
        public long GetCounterTotal(string statName)
        {
            if (m_statInstances.ContainsKey(typeof (Counter))
                && m_statInstances[typeof (Counter)].ContainsKey(statName))
            {
                return m_statInstances
                           [typeof (Counter)]
                           [statName]
                       .AsEnumerable()
                       .Sum(c => (long) c.CurrentAggregateValue);
            }
            return 0;
        }

        // TODO: Should Counter objects be able to be directly accessed, or should it be forced to visit through a Stats layer?
        // TODO: Access controls.  Right now anything that can reference 
        // Stats.Instance can get any stats at all.  That doesn't sound
        // right.

        public ISet<StatInst> GetCounters(string statName)
        {
            if (m_statInstances.ContainsKey(typeof(Counter))
                && m_statInstances[typeof(Counter)].ContainsKey(statName))
            {
                return m_statInstances[typeof(Counter)][statName];
            }
            return null;
        }

        public ISet<StatInst> GetDelegateCounters(string statName)
        {
            if (m_statInstances.ContainsKey(typeof(DelegateCounter))
                && m_statInstances[typeof(DelegateCounter)].ContainsKey(statName))
            {
                return m_statInstances[typeof(DelegateCounter)][statName];
            }
            return null;
        }

        public ISet<StatInst> GetDelegateFloats(string statName)
        {
            if (m_statInstances.ContainsKey(typeof(DelegateDoubleGenerator))
                && m_statInstances[typeof(DelegateDoubleGenerator)].ContainsKey(statName))
            {
                return m_statInstances[typeof(DelegateDoubleGenerator)][statName];
            }
            return null;
        }

        /// <summary>
        /// Get all StatInsts stored in this Stat as an IEnumerable.
        /// </summary>
        /// <returns>All StatInsts stored in this Stat.</returns>
        public IEnumerable<StatInst> GetStatInsts()
        {
            IEnumerable<StatInst> toReturn = null;
            foreach (Type type in m_statInstances.Keys)
            {
                foreach (ISet<StatInst> set in m_statInstances[type].Values)
                {
                    toReturn = ((toReturn == null) ? set : toReturn.Union(set));
                }
            }
            return toReturn;
        }

        // TODO: Should Duration objects be able to be directly accessed, or should it be forced to visit through a Stats layer?
        public ISet<StatInst> GetDurations(string statName)
        {
            if (m_statInstances.ContainsKey(typeof(Duration))
                && m_statInstances[typeof(Duration)].ContainsKey(statName))
            {
                return m_statInstances[typeof(Duration)][statName];
            }
            return null;
        }

        /// <summary>
        /// OnLog must be set for Log to do anything.  Typically this will be a (Logger subclass).Log action
        /// Start logging every 10 minutes like this: Interval.Instance.AddTimer(Log, 60000);
        /// </summary>
        public void Log()
        {
            if (OnLog != null)
            {
                OnLog(LogLevel, DateTime.Now, LogSource, LogTag, StatDispatcher.ToString());
            }
        }

        // TODO: Refactor this to use LINQ
        public override string ToString()
        {
            StringBuilder sb = new StringBuilder();
            sb.Append("{ ");
            // Empty returns "{ }"
            // Stat names but no items in them returns "{ foo => { } }"
            // Stat names, stat instances, but no DataPassingStructs returns "{ foo => { bar => [ ] } }"
            // Typically returns "{ foo => { bar => [ { baz => (66, 22) }, { foobar => (99, 44) } [ } }"
            // It would be nice if this used built-ins.

            // For testing, we need these in a deterministic order.
            // It's woefully inefficient, but ToString() shouldn't be used 
            // often (if ever) in live code.
            List<Type> keyTypes = m_statInstances.Keys.ToList();
            keyTypes.Sort();

            foreach (Type keyType in keyTypes)
            {
                IDictionary<string, ISet<StatInst>> statNameToStatInstanceDictionary =
                    m_statInstances[keyType];
                // foo in example above
                sb.Append(keyType);
                sb.Append(" => { ");

                // For testing, we need these in a deterministic order.
                // It's woefully inefficient, but ToString() shouldn't be used 
                // often (if ever) in live code.
                List<string> statNames = statNameToStatInstanceDictionary.Keys.ToList();
                statNames.Sort();

                foreach (string statName in statNames)
                {
                    // bar in example above
                    sb.Append(statName);
                    sb.Append(" => [ ");

                    // It doesn't make sense to sort these, since the values change frequently.
                    // For testing, 
                    foreach (StatInst statInst in statNameToStatInstanceDictionary[statName])
                    {
                        // "{ baz => (66, 22) }"  or "{ foobar => (99, 44) }" in example above
                        sb.Append(statInst);
                        sb.Append(", ");
                    }
                    sb.Append("], ");
                }
                sb.Append("}, ");
            }
            sb.Append("}");
            // Remove trailing commas
            sb.Replace(", ]", " ]");
            sb.Replace(", }", " }");
            return sb.ToString();
        }

#if DEBUG
        /// <summary>
        /// Clear the m_statInstances.  Used for testing.  Should never be 
        /// used in live application.
        /// </summary>
        public void Reset()
        {
            m_statInstances.Clear();
            OnLog = null;
            _mStats = null;
        }
#endif

        public void IntervalElapsed()
        {
            foreach (Type t in m_statInstances.Keys)
            {
                foreach (ISet<StatInst> set in m_statInstances[t].Values)
                {
                    foreach (StatInst statInst in set)
                    {
                        statInst.IntervalComplete();
                    }
                }
            }
        }
    }
}
