/*****************************************************************************\
 *             University of Illinois/NCSA Open Source License
 *
 * 
 *      Copyright  2012,2013, The Board of Trustees of the University of Illinois. All rights reserved.
 *
 *      Developed by: Rakesh Bobba, Derek Dagit, Erich Heine, Denis Kholine,
 *                    Klara Nahrstedt, David Rodgers, William H. Sanders, Tim Yardley
 *
 *      Information Trust Institute/Coordinated Science Laboratory
 * 
 *      University of Illinois
 *      
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      http://www.iti.illinois.edu / http://csl.illinois.edu
 * 
 *      Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
 *      associated documentation files (the Software), to deal with the Software without restriction, 
 *      including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *      and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so,
 *      subject to the following conditions:
 *
 *      Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *      following disclaimers.
 *      
 *      Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 *      the following disclaimers in the documentation and/or other materials provided with the distribution.
 *
 *      Neither the names of Information Trust institute, Coordinated Science Laboratory, University of Illinois,
 *      The Board of Trustees of the University of Illinois nor the names of its contributors may be used to 
 *      endorse or promote products derived from this Software without specific prior written permission.
 *      
 *      THE SOFTWARE IS PROVIDED AS IS, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 *      LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN
 *      NO EVENT SHALL THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *      WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
 *      SOFTWARE OR THE USE OR OTHER DEALINGS WITH THE SOFTWARE.
 *      
 \*****************************************************************************/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text.RegularExpressions;
using System.Threading;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using SIEGate.Core;
using SIEGate.Core.Measurements;
using SIEGate.Core.Stats;

namespace SIEGateCoreTest.StatTests
{
    [TestClass]
    public class StatTests
    {
        [TestInitialize]
        public void TestReset()
        {
            Stats.StatDispatcher.Reset();
        }

        [TestMethod]
        public void TestEmptyStatsToString()
        {
            string s = Stats.StatDispatcher.ToString();
            Assert.AreEqual("{ }", s);
        }

        [TestMethod]
        public void TestNameOnlyStatsToString()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            cluster.RegisterCounter("testNameOnlyStats", "foo");
            string s = Stats.StatDispatcher.ToString();
            Assert.AreEqual("{ SIEGate.Core.Stats.Counter => { foo => [ Current aggregate value: 0, last interval value: 0, since last interval value: 0 ] } }", s);
        }

        [TestMethod]
        public void TestRegisterCounterGetCounters()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter("foo", "bar");
            ISet<StatInst> counters = Stats.StatDispatcher.GetCounters("bar");
            Assert.AreEqual(1, counters.Count);
            foreach (Counter c in counters)
            {
                Assert.AreEqual(c, counter);
            }
        }

        [TestMethod]
        public void TestCounterDuplicateCurrentStatus()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter("foo", "bar");
            counter.Add(1);
            Stats.StatDispatcher.IntervalElapsed();
            counter.Add(2);

            // Counter and newCounter are identical, with CurrentAggregateValue
            // of 3, LastInterValValue of 1, and SinceLastInterval of 2
            Counter newCounter = counter.DuplicateCurrentStatus();

            // Since newCounter isn't registered in StatDispatcher (and
            // shouldn't be), calling IntervalElapsed shouldn't affect it
            Stats.StatDispatcher.IntervalElapsed();
            counter.Add(4);

            // Changing counter should not have changed newCounter in any way
            Assert.AreEqual(3, newCounter.CurrentAggregateValue);
            Assert.AreEqual(1, newCounter.LastIntervalValue);
            Assert.AreEqual(2, newCounter.SinceLastIntervalValue);

            Assert.AreNotEqual(counter.CurrentAggregateValue, newCounter.CurrentAggregateValue);
            Assert.AreNotEqual(counter.LastIntervalValue, newCounter.LastIntervalValue);
            Assert.AreNotEqual(counter.SinceLastIntervalValue, newCounter.SinceLastIntervalValue);
        }

        [TestMethod]
        public void TestGetCountersEmpty()
        {
            Assert.AreEqual(null, Stats.StatDispatcher.GetCounters("bar"));
        }

        [TestMethod]
        public void TestCounterGetOwnerNameAutomaticallyCreatedDelegate()
        {
            string name = "foo";
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter(name, "bar");
            Assert.AreEqual(name, counter.GetOwnerName());
        }

        [TestMethod]
        public void TestCounterGetValue()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter("foo", "bar");
            counter.Add(1);
            Assert.AreEqual(1, counter.CurrentAggregateValue);
        }

        [TestMethod]
        public void TestCounterGetValueSinceLastInterval()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter("foo", "bar");
            counter.Add(2);
            Assert.AreEqual(2, counter.SinceLastIntervalValue);
        }

        [TestMethod]
        public void TestCounterTotalerOwnerNameGet()
        {
            string name = "counterTotaler";
            CounterTotaler counterTotaler = new CounterTotaler(name, "stat");
            Assert.AreEqual(name, counterTotaler.GetOwnerName());
        }

        [TestMethod]
        public void TestCounterTotalerUseCachedValue()
        {
            CounterTotaler counterTotaler = new CounterTotaler("counterTotaler", "stat");
            counterTotaler.UseCachedValue = true;
            Assert.IsTrue(counterTotaler.UseCachedValue);
            counterTotaler.UseCachedValue = false;
            Assert.IsFalse(counterTotaler.UseCachedValue);
        }

        [TestMethod]
        public void TestCounterTotalerTrackChanges()
        {
            CounterTotaler counterTotaler = new CounterTotaler("counterTotaler", "stat");

            counterTotaler.UseCachedValue = false;

            const uint statSourceId = 0;

            Cluster cluster1 = new Cluster(1);
            Counter counter1 = cluster1.RegisterCounter("counter1", "stat");
            counterTotaler.AddCounter(counter1);

            Cluster cluster2 = new Cluster(2);
            Counter counter2 = cluster2.RegisterCounter("counter2", "stat");
            counterTotaler.AddCounter(counter2);

            counter1.Add(1);
            counter2.Add(2);
            Stats.StatDispatcher.IntervalElapsed();
            counter1.Add(4);
            counter2.Add(8);

            // Get the value here to make sure that the cached value isn't used
            long throwAway = counterTotaler.CurrentAggregateValue;
            throwAway = counterTotaler.SinceLastIntervalValue;

            Stats.StatDispatcher.IntervalElapsed();
            counter1.Add(16);
            counter2.Add(32);

            Assert.AreEqual(63, counterTotaler.CurrentAggregateValue);
            Assert.AreEqual(48, counterTotaler.SinceLastIntervalValue);
        }

        [TestMethod]
        public void TestCounterTotalerCache()
        {
            CounterTotaler counterTotaler = new CounterTotaler("counterTotaler", "stat");

            counterTotaler.UseCachedValue = true;

            const uint statSourceId = 0;

            Cluster cluster1 = new Cluster(1);
            Counter counter1 = cluster1.RegisterCounter("counter1", "stat");
            counterTotaler.AddCounter(counter1);

            Cluster cluster2 = new Cluster(2);
            Counter counter2 = cluster2.RegisterCounter("counter2", "stat");
            counterTotaler.AddCounter(counter2);

            counter1.Add(1);
            counter2.Add(2);
            Stats.StatDispatcher.IntervalElapsed();
            counter1.Add(4);
            counter2.Add(8);

            // Get the value here to make sure that the cached value is used
            long throwAway = counterTotaler.CurrentAggregateValue;
            throwAway = counterTotaler.SinceLastIntervalValue;

            counter1.Add(16);
            counter2.Add(32);

            Assert.AreEqual(15, counterTotaler.CurrentAggregateValue);
            Assert.AreEqual(3, counterTotaler.LastIntervalValue);
            Assert.AreEqual(12, counterTotaler.SinceLastIntervalValue);
        }

        [TestMethod]
        public void TestCounterTotalerDuplicateCurrentStatus()
        {
            CounterTotaler counterTotaler = new CounterTotaler("counterTotaler", "stat");

            counterTotaler.UseCachedValue = true;

            const uint statSourceId = 0;

            Cluster cluster1 = new Cluster(1);
            Counter counter1 = cluster1.RegisterCounter("counter1", "stat");
            counterTotaler.AddCounter(counter1);

            Cluster cluster2 = new Cluster(2);
            Counter counter2 = cluster2.RegisterCounter("counter2", "stat");
            counterTotaler.AddCounter(counter2);

            counter1.Add(1);
            counter2.Add(2);
            Stats.StatDispatcher.IntervalElapsed();
            counter1.Add(4);
            counter2.Add(8);

            CounterTotaler duplicate = counterTotaler.DuplicateCurrentStatus();

            Stats.StatDispatcher.IntervalElapsed();
            counter1.Add(16);
            counter2.Add(32);

            Assert.AreEqual(15, duplicate.CurrentAggregateValue);
            Assert.AreEqual(12, duplicate.SinceLastIntervalValue);
        }

        [TestMethod]
        public void TestClusterTotalCounter()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter("foo", "bar");
            counter.Add(16);
            Assert.AreEqual(16, Stats.StatDispatcher.GetCounterTotal("bar"));
        }

        [TestMethod]
        public void TestClusterAddToCounter()
        {
            string statName = "foo";
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter("foo", statName);
            cluster.AddToCounter(statName, 1);
            Assert.AreEqual(1, counter.CurrentAggregateValue);
        }

        [TestMethod]
        public void TestClusterGetCounter()
        {
            Cluster cluster = new Cluster(0);
            string statName = "bar";
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter("foo", statName);
            Assert.AreSame(counter, cluster.GetCounter(statName));
        }

        [TestMethod]
        public void TestClusterGetCounterExpectNull()
        {
            Cluster cluster = new Cluster(0);
            Assert.IsNull(cluster.GetCounter("does_not_exist"));
        }

        [TestMethod]
        public void TestClusterGetDuration()
        {
            Cluster cluster = new Cluster(0);
            string statName = "bar";
            const uint statSourceId = 0;
            Duration duration = cluster.RegisterDuration("foo", statName);
            Assert.AreSame(duration, cluster.GetDuration(statName));
        }

        [TestMethod]
        public void TestClusterGetDurationExpectNull()
        {
            Cluster cluster = new Cluster(0);
            Assert.IsNull(cluster.GetDuration("does_not_exist"));
        }

        [TestMethod]
        public void TestClusterCounterIntervalElapsed()
        {
            string statName = "foo";
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter("foo", statName);
            cluster.AddToCounter(statName, 1);
            Stats.StatDispatcher.IntervalElapsed();
            cluster.AddToCounter(statName, 2);
            Assert.AreEqual(2, counter.SinceLastIntervalValue);
            Assert.AreEqual(3, counter.CurrentAggregateValue);
        }

        [TestMethod]
        public void TestClusterDurationIntervalElapsed()
        {
            string statName = "bar";
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Duration duration = cluster.RegisterDuration("foo", statName);
            // Guarantee that some duration has gone by
            System.Threading.Thread.Sleep(1);
            Stats.StatDispatcher.IntervalElapsed();
            // Guarantee that some duration has gone by
            System.Threading.Thread.Sleep(1);
            TimeSpan resetValue = duration.SinceLastIntervalValue;
            // Guarantee that some duration has gone by
            System.Threading.Thread.Sleep(1);
            // Have to GetValue after get_SinceLastIntervalValue so that we
            // can be sure more time has gone by.
            TimeSpan currentValue = duration.CurrentAggregateValue;
            Assert.IsTrue(resetValue < currentValue);
        }

        [TestMethod]
        public void TestClusterToString()
        {
            Cluster cluster = new Cluster(0);
            string name = "foo";
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter(name, "bar");
            Duration duration = cluster.RegisterDuration(name, "baz");
            counter.Add(32);
            Stats.StatDispatcher.IntervalElapsed();
            counter.Add(64);
            Stats.StatDispatcher.IntervalElapsed();
            Assert.IsTrue(
                Regex.IsMatch(cluster.ToString(),
                              @"{ SIEGate\.Core\.Stats\.Counter => { bar => \( Current aggregate value: 96, last interval value: 64, since last interval value: 0 \) }, SIEGate\.Core\.Stats\.Duration => { baz => \( Current aggregate value: [0-9]{2}:[0-5][0-9]:[0-5][0-9](\.[0-9]*)?, last interval value: [0-9]{2}:[0-5][0-9]:[0-5][0-9](\.[0-9]*)?, since last interval value: [0-9]{2}:[0-5][0-9]:[0-5][0-9](\.[0-9]*)? \) } }"));
        }

        [TestMethod]
        public void TestClusterTotalCounterNonexistingStat()
        {
            Assert.AreEqual(0, Stats.StatDispatcher.GetCounterTotal("bar"));
        }

        [TestMethod]
        public void TestClusterRegisterDurationGetDurations()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Duration duration = cluster.RegisterDuration("foo", "bar");
            ISet<StatInst> durations = Stats.StatDispatcher.GetDurations("bar");
            Assert.AreEqual(1, durations.Count);
            foreach (Duration d in durations)
            {
                Assert.AreEqual(d, duration);
            }
        }

        [TestMethod]
        public void TestDurationDuplicateCurrentStats()
        {
            Cluster cluster = new Cluster(0);
            const string statName = "bar";
            const uint statSourceId = 0;
            Duration duration = cluster.RegisterDuration("foo", statName);
            System.Threading.Thread.Sleep(10);
            Stats.StatDispatcher.IntervalElapsed();
            Duration duplicate = duration.DuplicateCurrentStatus();
            Assert.AreEqual(1, Stats.StatDispatcher.GetDurations(statName).Count);
            Assert.AreEqual(duration.Name, duplicate.Name);
            TimeSpan currentDuration = duration.CurrentAggregateValue;
            // Make sure that there's a difference before the next statement executes.
            System.Threading.Thread.Sleep(1);
            TimeSpan currentDuplicate = duplicate.CurrentAggregateValue;
            Assert.IsTrue(currentDuration < currentDuplicate);
            TimeSpan currentDurationSinceLastIntervalElapsed = duration.SinceLastIntervalValue;
            // Make sure that there's a difference before the next statement executes.
            System.Threading.Thread.Sleep(1);
            TimeSpan currentDuplicateSinceLastIntervalElapsed = duplicate.SinceLastIntervalValue;
            Assert.IsTrue(currentDurationSinceLastIntervalElapsed < currentDuplicateSinceLastIntervalElapsed);
        }

        [TestMethod]
        public void TestDurationGetOwnerNameAutomaticallyCreatedDelegate()
        {
            string name = "foo";
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Duration duration = cluster.RegisterDuration(name, "bar");
            Assert.AreEqual(name, duration.GetOwnerName());
        }

        [TestMethod]
        public void TestDurationGetValue()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Duration duration = cluster.RegisterDuration("foo", "bar");
            // Guarantee that some duration has gone byStatMap
            System.Threading.Thread.Sleep(1);
            TimeSpan value = duration.CurrentAggregateValue;
            Assert.IsTrue(value > TimeSpan.Zero, value.ToString() + " was not greater than 0");
        }

        [TestMethod]
        public void TestDurationGetValueSinceLastIntervalElapsed()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Duration duration = cluster.RegisterDuration("foo", "bar");
            // Guarantee that some duration has gone by
            System.Threading.Thread.Sleep(1);
            TimeSpan resetValue = duration.SinceLastIntervalValue;
            Assert.IsTrue(resetValue > TimeSpan.Zero, resetValue.ToString() + " was not greater than 0");
        }

        [TestMethod]
        public void TestDurationGetValueSinceLastIntervalValue()
        {
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Duration duration = cluster.RegisterDuration("foo", "bar");
            // Guarantee that some duration has gone by
            System.Threading.Thread.Sleep(1);
            Stats.StatDispatcher.IntervalElapsed();
            // Guarantee that some duration has gone by
            System.Threading.Thread.Sleep(1);
            TimeSpan resetValue = duration.SinceLastIntervalValue;
            // Guarantee that some duration has gone by
            System.Threading.Thread.Sleep(1);
            // Have to GetValue after SinceLastIntervalValue so that we can be
            // sure more time has gone by.
            TimeSpan currentValue = duration.CurrentAggregateValue;
            Assert.IsTrue(resetValue < currentValue);
        }

        [TestMethod]
        public void TestGetDurationsEmpty()
        {
            Assert.AreEqual(null, Stats.StatDispatcher.GetDurations("bar"));
        }

        [TestMethod]
        public void TestStatInstName_get()
        {
            string name = "foo";
            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter = cluster.RegisterCounter("foo", name);
            Assert.AreEqual(name, counter.Name);
        }

        // Add is gone, replaced with AttemptAddMeasurementCreatorForGuid, which uses the Info object to create counters
        //[TestMethod]
        //public void TestStatMapperStatsReadyEvent()
        //{
        //    StatMapper statMapper = new StatMapper(new StatMapper.AdapterInfo());
        //    Cluster cluster = new Cluster(0);
        //    const uint statSourceId = 0;
        //    Counter counter = cluster.RegisterCounter(statSourceId, "foo", "counter");

        //    uint sourceId = 0;

        //    Guid currentCounterGuid = statMapper.Add(sourceId, counter, TypeOfStat.Current);
        //    Guid lastIntervalCounterGuid = statMapper.Add(sourceId, counter, TypeOfStat.LastInterval);
        //    Guid sinceLastIntervalCounterGuid = statMapper.Add(sourceId, counter, TypeOfStat.SinceLastInterval);

        //    Duration duration = cluster.RegisterDuration(statSourceId, "foo", "duration");
        //    // Adding sinceLastInterval first so that it's more likely to execute first
        //    Guid sinceLastIntervalDurationGuid = statMapper.Add(sourceId, duration, TypeOfStat.SinceLastInterval);
        //    Guid lastIntervalDurationGuid = statMapper.Add(sourceId, duration, TypeOfStat.LastInterval);
        //    Guid currentDurationGuid = statMapper.Add(sourceId, duration, TypeOfStat.Current);

        //    // Need a moderately high sleep time in order to ensure that
        //    // enough time passes before and after reset for the test
        //    // to pass
        //    const int SLEEP_TIME = 50;
        //    System.Threading.Thread.Sleep(SLEEP_TIME);
        //    counter.Add(1);
        //    Stats.StatDispatcher.IntervalElapsed();
        //    System.Threading.Thread.Sleep(SLEEP_TIME);
        //    counter.Add(2);

        //    long resultCounterCurrent = 0;
        //    long resultCounterLastInterval = 0;
        //    long resultCounterSinceLastInterval = 0;
        //    TimeSpan resultTimeSpanSinceEver = TimeSpan.Zero;
        //    TimeSpan resultTimeSpanLastInterval = TimeSpan.Zero;
        //    TimeSpan resultTimeSpanSinceInterval = TimeSpan.Zero;

        //    statMapper.Emit += measurements =>
        //        {
        //            foreach (IMeasurement measurement in measurements)
        //            {
        //                if (measurement.StreamID.Equals(currentCounterGuid))
        //                {
        //                    resultCounterCurrent = measurement.Value.GetRaw();
        //                }
        //                else if (measurement.StreamID.Equals(lastIntervalCounterGuid))
        //                {
        //                    resultCounterLastInterval = measurement.Value.GetRaw();
        //                }
        //                else if (measurement.StreamID.Equals(sinceLastIntervalCounterGuid))
        //                {
        //                    resultCounterSinceLastInterval = measurement.Value.GetRaw();
        //                }
        //                else if (measurement.StreamID.Equals(currentDurationGuid))
        //                {
        //                    resultTimeSpanSinceEver = measurement.Value.GetRaw();
        //                }
        //                else if (measurement.StreamID.Equals(lastIntervalDurationGuid))
        //                {
        //                    resultTimeSpanLastInterval = measurement.Value.GetRaw();
        //                }
        //                else if (measurement.StreamID.Equals(sinceLastIntervalDurationGuid))
        //                {
        //                    resultTimeSpanSinceInterval = measurement.Value.GetRaw();
        //                }
        //            }
        //        };

        //    statMapper.StatsReadyEvent();

        //    Assert.AreEqual(3, resultCounterCurrent);
        //    Assert.AreEqual(1, resultCounterLastInterval);
        //    Assert.AreEqual(2, resultCounterSinceLastInterval);
        //    Assert.IsTrue(resultTimeSpanSinceEver >= TimeSpan.FromMilliseconds(SLEEP_TIME * 2), "Time span since ever was: " + resultTimeSpanSinceEver + ", expected at least " + TimeSpan.FromMilliseconds(SLEEP_TIME * 2));
        //    Assert.IsTrue(resultTimeSpanSinceInterval >= TimeSpan.FromMilliseconds(SLEEP_TIME), "Time span since last interval was: " + resultTimeSpanSinceInterval + ", expected at least " + TimeSpan.FromMilliseconds(SLEEP_TIME));
        //    Assert.IsTrue(resultTimeSpanLastInterval >= TimeSpan.FromMilliseconds(SLEEP_TIME), "Time span since ever was: " + resultTimeSpanLastInterval + ", expected at least " + TimeSpan.FromMilliseconds(SLEEP_TIME));
        //    Assert.IsTrue(resultTimeSpanSinceEver > resultTimeSpanSinceInterval, "Time span since ever was: " + resultTimeSpanSinceEver + ", expected at least " + resultTimeSpanSinceInterval);
        //}

        //[TestMethod]
        //public void TestStatMapperCounterMeasurementCopy()
        //{
        //    StatMapper statMapper = new StatMapper(new StatMapper.AdapterInfo());
        //    Cluster cluster = new Cluster(0);
        //    const uint statSourceId = 0;
        //    Counter counter = cluster.RegisterCounter(statSourceId, "foo", "counter");

        //    uint sourceId = 0;
        //    Guid currentCounterGuid = statMapper.Add(sourceId, counter, TypeOfStat.Current);

        //    counter.Add(1);
        //    Stats.StatDispatcher.IntervalElapsed();
        //    counter.Add(2);

        //    Measurement resultCounterCurrent = null;

        //    statMapper.Emit += measurements =>
        //    {
        //        foreach (IMeasurement measurement in measurements)
        //        {
        //            if (measurement.StreamID.Equals(currentCounterGuid))
        //            {
        //                resultCounterCurrent = (Measurement) measurement;
        //            }
        //        }
        //    };

        //    statMapper.StatsReadyEvent();

        //    IMeasurement copy = resultCounterCurrent.Copy();
        //    Assert.AreEqual(resultCounterCurrent.StateFlags, copy.StateFlags);
        //    Assert.AreEqual(resultCounterCurrent.StreamID, copy.StreamID);
        //    Assert.AreEqual(resultCounterCurrent.Timestamp, copy.Timestamp);
        //    Assert.AreEqual(resultCounterCurrent.Value, copy.Value);
        //}

        //[TestMethod]
        //public void TestStatMapperDurationMeasurementCopy()
        //{
        //    StatMapper statMapper = new StatMapper(new StatMapper.AdapterInfo());
        //    Cluster cluster = new Cluster(0);
        //    const uint statSourceId = 0;
        //    Duration duration = cluster.RegisterDuration(statSourceId, "foo", "duration");

        //    uint sourceId = 0;
        //    Guid currentCounterGuid = statMapper.Add(sourceId, duration, TypeOfStat.Current);

        //    // Need a moderately high sleep time in order to ensure that
        //    // enough time passes before and after reset for the test
        //    // to pass
        //    const int SLEEP_TIME = 100;
        //    System.Threading.Thread.Sleep(SLEEP_TIME);
        //    Stats.StatDispatcher.IntervalElapsed();
        //    System.Threading.Thread.Sleep(SLEEP_TIME);

        //    Measurement resultDurationCurrent = null;

        //    statMapper.Emit += measurements =>
        //    {
        //        foreach (IMeasurement measurement in measurements)
        //        {
        //            resultDurationCurrent = (Measurement) measurement;
        //        }
        //    };

        //    statMapper.StatsReadyEvent();

        //    IMeasurement copy = resultDurationCurrent.Copy();

        //    Assert.AreEqual(resultDurationCurrent.StateFlags, copy.StateFlags);
        //    Assert.AreEqual(resultDurationCurrent.StreamID, copy.StreamID);
        //    Assert.AreEqual(resultDurationCurrent.Timestamp, copy.Timestamp);
        //    Assert.AreEqual(resultDurationCurrent.Value, copy.Value);
        //}
        
        [TestMethod]
        public void TestGetInterval()
        {
            Interval.Instance.ClearStatOutputtersAndUpdaters();

            Interval.Instance.TimerInterval = 128;
            Assert.AreEqual(128, Interval.Instance.TimerInterval);
            // Change again just in case the timer happened to be 128 the first time.
            Interval.Instance.TimerInterval = 256;
            Assert.AreEqual(256, Interval.Instance.TimerInterval);
        }

        [TestMethod]
        public void TestSetInterval()
        {
            Interval.Instance.ClearStatOutputtersAndUpdaters();

            Cluster cluster = new Cluster(0);
            const uint statSourceId = 0;
            Counter counter1 = cluster.RegisterCounter("foo", "counter1");
            Counter counter2 = cluster.RegisterCounter("foo", "counter2");
            bool calledUpdater = false, calledOutputter = false;
            Interval.Instance.StatUpdaters += (sender, args) => calledUpdater = true;
            Interval.Instance.StatOutputters += (sender, args) => calledOutputter = true;
            Interval.Instance.TimerInterval = 50;
            Thread.Sleep(200);
            Assert.IsTrue(calledUpdater);
            Assert.IsTrue(calledOutputter);
        }

        [TestMethod]
        public void TestLog()
        {
            LogLevel inputLevel = LogLevel.Verbose;
            string inputSource = "TestingSource";
            string inputTag = "TestingTag";

            Stats.StatDispatcher.LogLevel = inputLevel;
            Stats.StatDispatcher.LogSource = inputSource;
            Stats.StatDispatcher.LogTag = inputTag;

            LogLevel localLevel = LogLevel.Fatal;
            string localSource = "";
            string localTag = "";
            Stats.StatDispatcher.OnLog += (level, time, source, tag, message) =>
                {
                    localLevel = level;
                    localSource = source;
                    localTag = tag;
                };
            Stats.StatDispatcher.Log();

            Assert.AreEqual(inputLevel, localLevel);
            Assert.AreEqual(inputSource, localSource);
            Assert.AreEqual(inputTag, localTag);
        }

        [TestMethod]
        public void TestReregisterCounter()
        {
            uint id = 256;
            Cluster c1 = Stats.StatDispatcher.RegisterCluster(id);
            const uint statSourceId = 0;
            string ownerName = "foo";
            string statName = "bar";
            c1.RegisterCounter(ownerName, statName);
            c1.AddToCounter(statName, 1);

            Cluster c2 = Stats.StatDispatcher.RegisterCluster(id);
            Stats.StatDispatcher.IntervalElapsed();
            c2.AddToCounter(statName, 2);
            Assert.AreEqual(3, c2.GetCounter(statName).CurrentAggregateValue, "Stat value did not refresh across multiple registrations.");
            Assert.AreEqual(1, c2.GetCounter(statName).LastIntervalValue, "Stat value from last interval did not refresh across multiple registrations.");
            Assert.AreEqual(2, c2.GetCounter(statName).SinceLastIntervalValue, "Stat value since last interval did not refresh across multiple registrations.");
        }

        [TestMethod]
        public void TestSystemStats()
        {
            float cpuUsage = Stats.StatDispatcher.SystemStats.GetDelegateDoubleGeneratorValue(Stats.StrStatCpuUsage);
            float processCpuUsage = Stats.StatDispatcher.SystemStats.GetDelegateDoubleGeneratorValue(Stats.StrStatProcessCpuUsage);

            Assert.IsTrue(cpuUsage > 0, "We should be using more than 0% CPU, but we're using " + cpuUsage);
            Assert.IsTrue(processCpuUsage > 0, "The process should be using more than 0% CPU, but we're using " + processCpuUsage);

//            float processMemoryUsage = Stats.StatDispatcher.SystemStats.GetCounter(Stats.StrStatCpuUsage).CurrentAggregateValue;
//            Assert.IsTrue(processMemoryUsage > 0, "Not using any memory; using " + processMemoryUsage);
        }

        [TestMethod]
        public void ExploreGuidToString()
        {
            Guid g = Guid.NewGuid();
            string s0 = g.ToString();
            string s1 = string.Format("Guid: {0}", g);
            Assert.IsNotNull(s1);
        }

        [TestMethod]
        public void ExploreProcessCounter()
        {
            string[] interestingCategories = new string[] { "ICMP", "ICMPv6", "IPHTTPS Global", "IPsec AuthIP IPv4",
                                                            "IPsec AuthIP IPv6", "IPsec Driver", "IPsec IKEv1 IPv4", 
                                                            "IPsec IKEv1 IPv6", "IPsec IKEv2 IPv4", "IPsec IKEv2 IPv6",
                                                            "IPv4", "IPv6", "Memory", "MSSQL$SQLEXPRESS:Access Methods", 
                                                            "MSSQL$SQLEXPRESS:Availability Replica", 
                                                            "MSSQL$SQLEXPRESS:Backup Device", 
                                                            "MSSQL$SQLEXPRESS:Batch Resp Statistics", 
                                                            "MSSQL$SQLEXPRESS:Broker Activation", 
                                                            "MSSQL$SQLEXPRESS:Broker Statistics", 
                                                            "MSSQL$SQLEXPRESS:Broker TO Statistics", 
                                                            "MSSQL$SQLEXPRESS:Broker/DBM Transport", 
                                                            "MSSQL$SQLEXPRESS:Buffer Manager", 
                                                            "MSSQL$SQLEXPRESS:Buffer Node", 
                                                            "MSSQL$SQLEXPRESS:Catalog Metadata", 
                                                            "MSSQL$SQLEXPRESS:CLR", 
                                                            "MSSQL$SQLEXPRESS:Cursor Manager by Type", 
                                                            "MSSQL$SQLEXPRESS:Cursor Manager Total", 
                                                            "MSSQL$SQLEXPRESS:Database Mirroring", 
                                                            "MSSQL$SQLEXPRESS:Database Replica", 
                                                            "MSSQL$SQLEXPRESS:Databases", 
                                                            "MSSQL$SQLEXPRESS:Deprecated Features", 
                                                            "MSSQL$SQLEXPRESS:Exec Statistics", 
                                                            "MSSQL$SQLEXPRESS:FileTable", 
                                                            "MSSQL$SQLEXPRESS:General Statistics", 
                                                            "MSSQL$SQLEXPRESS:Latches", 
                                                            "MSSQL$SQLEXPRESS:Locks", 
                                                            "MSSQL$SQLEXPRESS:Memory Broker Clerks", 
                                                            "MSSQL$SQLEXPRESS:Memory Manager", 
                                                            "MSSQL$SQLEXPRESS:Memory Node", 
                                                            "MSSQL$SQLEXPRESS:Plan Cache", 
                                                            "MSSQL$SQLEXPRESS:Query Execution", 
                                                            "MSSQL$SQLEXPRESS:Replication Agents", 
                                                            "MSSQL$SQLEXPRESS:Replication Dist.", 
                                                            "MSSQL$SQLEXPRESS:Replication Logreader", 
                                                            "MSSQL$SQLEXPRESS:Replication Merge", 
                                                            "MSSQL$SQLEXPRESS:Replication Snapshot", 
                                                            "MSSQL$SQLEXPRESS:Resource Pool Stats", 
                                                            "MSSQL$SQLEXPRESS:SQL Errors", 
                                                            "MSSQL$SQLEXPRESS:SQL Statistics", 
                                                            "MSSQL$SQLEXPRESS:Transactions", 
                                                            "MSSQL$SQLEXPRESS:User Settable", 
                                                            "MSSQL$SQLEXPRESS:Wait Statistics", 
                                                            "MSSQL$SQLEXPRESS:Workload Group Stats", 
                                                            "Paging File", 
                                                            "Per Processor Network Activity Cycles", 
                                                            "Per Processor Network Interface Card Activity", 
                                                            "PhysicalDisk", "Power Meter", "Process",
                                                            "Processor", "System", "TCPv4", "TCPv6",
                                                            "Thread", "UDPv4", "UDPv6"
                                                            };
            foreach (PerformanceCounterCategory category in PerformanceCounterCategory.GetCategories())
            {
                foreach (string interestingCategory in interestingCategories)
                {
                    if (interestingCategory.Equals(category.CategoryName))
                    {
                        System.Diagnostics.Debug.WriteLine("====================");
                        System.Diagnostics.Debug.WriteLine("For category: " + interestingCategory);
                        try
                        {
                            foreach (string instanceName in category.GetInstanceNames())
                            {
                                System.Diagnostics.Debug.WriteLine("I    " + instanceName);
                                try
                                {
                                    foreach (PerformanceCounter counter in category.GetCounters(instanceName))
                                    {
                                        try
                                        {
                                            System.Diagnostics.Debug.WriteLine("IC   " + counter.CounterName);
                                        }
                                        catch (Exception e)
                                        {
                                            System.Diagnostics.Debug.WriteLine(
                                                "Exception while trying to get write instanced counter");
                                        }
                                    }
                                }
                                catch (Exception e)
                                {
                                    System.Diagnostics.Debug.WriteLine("Exception while trying to get counters");
                                }
                            }
                        }
                        catch (Exception e)
                        {
                            System.Diagnostics.Debug.WriteLine("Exception while trying to get instance names.");
                        }
                        System.Diagnostics.Debug.WriteLine("====================");
                    }
                }
            }
        }
    }
}
