﻿//******************************************************************************************************
//  App.xaml.cs - Gbtc
//
//  Copyright © 2011, Grid Protection Alliance.  All Rights Reserved.
//
//  Licensed to the Grid Protection Alliance (GPA) under one or more contributor license agreements. See
//  the NOTICE file distributed with this work for additional information regarding copyright ownership.
//  The GPA licenses this file to you under the Eclipse Public License -v 1.0 (the "License"); you may
//  not use this file except in compliance with the License. You may obtain a copy of the License at:
//
//      http://www.opensource.org/licenses/eclipse-1.0.php
//
//  Unless agreed to in writing, the subject software distributed under the License is distributed on an
//  "AS-IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. Refer to the
//  License for the specific language governing permissions and limitations.
//
//  Code Modification History:
//  ----------------------------------------------------------------------------------------------------
//  09/07/2010 - Stephen C. Wills
//       Generated original version of source code.
//  03/02/2011 - J. Ritchie Carroll
//       Added unhandled exception logger with dialog for better end user problem diagnosis.
//
//******************************************************************************************************

using System;
using System.Diagnostics;
using System.IO;
using System.Security.Principal;
using System.Windows;
using System.Xml;
using GSF.ErrorManagement;
using GSF.IO;
using GSF.Security.Cryptography;

namespace ConfigurationSetupUtility
{
    /// <summary>
    /// Interaction logic for App.xaml
    /// </summary>
    public partial class App : Application
    {
        #region [ Members ]

        // Constants
        public const CipherStrength CryptoStrength = CipherStrength.Aes256;
        public const string CipherLookupKey = "0679d9ae-aca5-4702-a3f5-604415096987";
        public const string ApplicationExe = "SIEGate.exe";
        public const string ApplicationConfig = "SIEGate.exe.config";
        public const string Manager = "SIEGateManager";
        public const string ManagerExe = "SIEGateManager.exe";
        public const string ManagerConfig = "SIEGateManager.exe.config";
        public const string BaseSqliteConfig = "SIEGate.db";
        public const string SqliteConfigv2 = "SIEGatev2.db";
        public const string SqliteSampleData = "SIEGate-SampleDataSet.db";
        public const string SqliteInitialData = "SIEGate-InitialDataSet.db";
        private ErrorLogger m_errorLogger;
        private Func<string> m_defaultErrorText;

        #endregion

        #region [ Constructors ]

        public App()
        {
            AppDomain.CurrentDomain.SetPrincipalPolicy(PrincipalPolicy.WindowsPrincipal);

            m_errorLogger = new ErrorLogger();
            m_defaultErrorText = m_errorLogger.ErrorTextMethod;
            m_errorLogger.ErrorTextMethod = ErrorText;
            m_errorLogger.ExitOnUnhandledException = false;
            m_errorLogger.HandleUnhandledException = true;
            m_errorLogger.LogToEmail = false;
            m_errorLogger.LogToEventLog = true;
            m_errorLogger.LogToFile = true;
            m_errorLogger.LogToScreenshot = true;
            m_errorLogger.LogToUI = true;
            m_errorLogger.Initialize();

            // When run from the installer the current directory may not be the directory where this application is running
            Directory.SetCurrentDirectory(FilePath.GetAbsolutePath(""));

            // Attempt to create an event log source for SIEGate applications that need to do authentication logging.
            // This needs to be done here since the CSU runs with administrative privileges and the SIEGate applications
            // normally do not; also there is a short system delay that exists before you can write to a new event log
            // source after it is first created.
            TryCreateEventSource("SIEGate.exe.config", "SIEGate");
            TryCreateEventSource("SIEGateManager.exe.config", "SIEGate Manager");
        }

        #endregion

        #region [ Methods ]

        private void TryCreateEventSource(string configFileName, string defaultApplicationName)
        {
            try
            {
                // Attempt to load the application name defined for the application in its configuration file
                string configFilePath = FilePath.GetAbsolutePath(configFileName);

                if (File.Exists(configFilePath))
                {
                    XmlDocument configFile = new XmlDocument();
                    configFile.Load(configFilePath);

                    XmlNode securitySettings = configFile.SelectSingleNode("configuration/categorizedSettings/securityProvider");

                    if ((object)securitySettings != null)
                    {
                        string applicationName = null;

                        foreach (XmlNode child in securitySettings.ChildNodes)
                        {
                            if ((object)child.Attributes != null && (object)child.Attributes["name"] != null)
                            {
                                if (child.Attributes["name"].Value == "ApplicationName")
                                {
                                    applicationName = child.Attributes["value"].Value;
                                    break;
                                }
                            }
                        }

                        if (string.IsNullOrWhiteSpace(applicationName))
                            applicationName = defaultApplicationName;

                        // Create the event log source based on defined application name for application if it does not already exist
                        if (!EventLog.SourceExists(applicationName))
                            EventLog.CreateEventSource(applicationName, "Application");
                    }
                }
            }
            catch (Exception ex)
            {
                m_errorLogger.Log(new InvalidOperationException(string.Format("Warning: failed to create or validate the event log source for {0}: {1}", defaultApplicationName, ex.Message), ex), false);
            }
        }

        private string ErrorText()
        {
            string errorMessage = m_defaultErrorText();
            Exception ex = m_errorLogger.LastException;

            if (ex != null)
            {
                if (string.Compare(ex.Message, "UnhandledException", true) == 0 && ex.InnerException != null)
                    ex = ex.InnerException;

                errorMessage = string.Format("{0}\r\n\r\nError details: {1}", errorMessage, ex.Message);
            }

            return errorMessage;
        }

        #endregion
    }
}
